// ======================================================================
// _cfgBase.hpp - defines the base CFG class: INTERNAL USE ONLY.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120602: Benjamin Han <benhdj@cs.cmu.edu> Created (total rewrite).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _CFGBASE_HPP
#define _CFGBASE_HPP

#include "tbconsts.hpp"

#include <list>

namespace Toolbox {

// undefine this to compile a version with smaller memory footprint, but you
// won't be able to remove rules
//#define WITH_REMOVE_RULE

#define CFG_SIZETYPE unsigned int

template <class RLiteral> struct RHS: public std::list<RLiteral> {};

// ======================================================================
// INTERNAL USE: CFG base class - deal with generic interactions between 
//               Index and Rules
// ======================================================================

template <class LLiteral, class RLiteral, class Body,
	  class Key, class _Index, class _Rules>
class _CFGBase {
protected:
  typedef _Index Index;
  typedef _Rules Rules;

public:
  typedef CFG_SIZETYPE size_type;
  typedef RHS<RLiteral> RHS;

  struct RuleIterator: public Rules::iterator {
    typedef typename Rules::iterator _Parent;
    typename Index::iterator ii;

    RuleIterator () {}
    RuleIterator (_Parent i) { _Parent::operator=(i); }
  };

protected:
  Index index;    // an index into the rules
  Rules rules;    // storage of the rule contents

public:
  
  // return false if k doesn't exist; [riFirst,riLast] indicates the matching rules
  // being agnostic about the implementation of index.find() - it could be non-const
  bool findRule (const Key &k, RuleIterator &riFirst, RuleIterator &riLast) {
    typename Index::iterator ii;
    bool result=index.find(k,ii,riFirst,riLast);

    riFirst.ii=riLast.ii=ii;
    return result;
  }

  // add a rule and return the body of the new rule (so you can modify it)
  Body &addRule (const LLiteral &lhs, const RHS &rhs) {
    typename Index::iterator ii;
    typename Rules::iterator ri,riFirst,riLast;    

    // if the new rule is already in index, update the corresponding index entry
    // otherwise add a new entry to index
    if (index.find(lhs,rhs,ii,riFirst,riLast))
      index.add(ii,ri=rules.add(lhs,rhs,riFirst,riLast));  // add by updating
    else
      index.add(lhs,rhs,ri=rules.add(lhs,rhs));
    return ri->body;
  }

#ifdef WITH_REMOVE_RULE
  // remove a rule based on the given rule iterator
  void removeRule (RuleIterator ri) {
    index.remove(ri.ii,ri);
    rules.remove(ri);
  }
#endif

  void clear () {
    index.clear();
    rules.clear();
  }

  // return # of rules
  typename Rules::size_type size () const { return rules.size(); }
};


// ======================================================================
// class _IndexEntry (internal use ONLY): generic entries in an index
// ======================================================================

template <class Rules>
struct _IndexEntry { 
  typename Rules::iterator riFirst,riLast; // an *inclusive* range of rules

  _IndexEntry () {}   // dummy to satisfy Trie
  _IndexEntry (typename Rules::iterator ri) { riFirst=riLast=ri;  }
};

};

#endif
