// ======================================================================
// _treeNode.hpp - Nodes of a tree (both unordered and ordered): INTERNAL
//                 USE ONLY.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120602: Benjamin Han <benhdj@cs.cmu.edu> Changed name from treeNode.hpp
//         to _treeNode.hpp.
// 102402: Benjamin Han <benhdj@cs.cmu.edu> Pulled out all embedded structure/
//         class definitions as the first step in making the code compile
//         under the M$ VC6 (boy does it suck); added the missing #include
//         assert.h>
// 111401: Benjamin Han <benhdj@cs.cmu.edu> Created (moved from treeIter.hpp).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _TREENODE_HPP
#define _TREENODE_HPP

#include "tbconsts.hpp"

#include <list>
#include <set>
#include <map>

#include <assert.h>

namespace Toolbox {

template <class TNode> struct _TNPtrList: public std::list<TNode*> {};
template <class TNode> struct _TNPtrSet: public std::set<TNode*> {};

// INTERNAL USE
// this is to avoid using T directly as the keys of DMap;
// this reduces the memory usage since we won't store T twice (otherwise
// the keys in DMap would be T)
template <class T, class Cmp>
struct _PtrCmp {
  Cmp cmp;

  bool operator () (T *ptr1, T *ptr2) const {
    return cmp(*ptr1,*ptr2);
  }
};

// INTERNAL USE
template <class T, class Cmp, class TNode>
struct _DMap: public std::map<T*,_TNPtrSet<TNode>,_PtrCmp<T,Cmp> > {
  typedef std::map<T*,_TNPtrSet<TNode>,_PtrCmp<T,Cmp> > _Parent;
  typedef typename _Parent::iterator iterator;

  iterator find (const T *dataPtr) {
    return _Parent::find(const_cast<T*>(dataPtr));
  }

  void erase (TNode *tnPtr) {
    assert(tnPtr);
    iterator dmi=find(&(tnPtr->data));
    assert(dmi!=end());
    if (dmi->second.size()==1) _Parent::erase(dmi);
    else {
      // a strange bug prevented me from doing this:
      // dmi->second.erase(tnPtr->di);
      // it'll remove more than one elements from the _TNPtrSet<TNode>!
      typename _TNPtrSet<TNode>::iterator tmpIter=dmi->second.find(tnPtr);
      assert(tmpIter!=dmi->second.end());
      dmi->second.erase(tmpIter);
      if (dmi->first==&(tnPtr->data))
	*(const_cast<T**>(&(dmi->first)))=&((*(dmi->second.begin()))->data);
    }
  }
};

// INTERNAL USE
template <class T, class Cmp, class TNode>
struct _TreeNodeIndex {
  typename _DMap<T,Cmp,TNode>::iterator dmi;
  typename _TNPtrSet<TNode>::const_iterator si;
    
  TNode *readPtr () const { return *si; }  
  bool next () { return ((++si)!=dmi->second.end()); }
};

// Both _UTreeNode and _TreeNode provides the same interface so they can be
// plugged in to produce unordered/ordered tree/tree iterators.

// ===========================================================================
// INTERNAL USE
// class _UTreeNode (internal use ONLY): an unordered tree node
// 'T' is the type of the payload, and 'Cmp' is a binary predicate "less-than"
// to compare two pieces of data.
// ===========================================================================

template <class T, class Cmp>
class _UTreeNode {    
  // map of daughters for quick lookup
  _DMap<T,Cmp,_UTreeNode> dMap;

  // the position in mother->dMap
  typename _DMap<T,Cmp,_UTreeNode>::iterator dmi;
  typename _TNPtrSet<_UTreeNode>::iterator si;

public:
  typedef size_t size_type;

  _UTreeNode *mother;  
  size_type depth;
  T data;

  _UTreeNode (const T &data, _UTreeNode *mother=NULL)
    :mother(mother),data(data) {
    if (mother) depth=mother->depth+1;
    else depth=0;
  }

  size_type size() const { 
    typename _DMap<T,Cmp,_UTreeNode>::const_iterator dmi;
    size_type s=0;
    for (dmi=dMap.begin();dmi!=dMap.end();++dmi) s+=dmi->second.size();
    return s; 
  }

  // WARNING: make sure the element exists
  const T& front () const { assert(dMap.size()); return *(dMap.begin()->first); }
  const T& back () const { assert(dMap.size()); return *((--dMap.end())->first); }
  
  _UTreeNode *previous () const {  // return the previous sibling
    assert(mother);
    assert(!(dmi==mother->dMap.begin() && si==dmi->second.begin()));
    typename _TNPtrSet<_UTreeNode>::iterator prevSI;

    if (si!=dmi->second.begin()) --(prevSI=si);
    else {
      typename _DMap<T,Cmp,_UTreeNode>::iterator prevDMI;
      --(prevDMI=dmi);
      --(prevSI=prevDMI->second.end());
    }
    return *prevSI;
  }
  _UTreeNode *next () const {     // return the next sibling
    assert(mother);
    typename _TNPtrSet<_UTreeNode>::iterator nextSI;
    
    if (++(nextSI=si)!=dmi->second.end()) return *nextSI;
    else {
      typename _DMap<T,Cmp,_UTreeNode>::iterator nextDMI;
      if (++(nextDMI=dmi)!=mother->dMap.end()) return *(nextDMI->second.begin());
      else return NULL;
    }
  }

  _UTreeNode *frontPtr () const { 
    assert(dMap.size()); 
    return *(dMap.begin()->second.begin()); 
  }
  _UTreeNode *backPtr () const { 
    assert(dMap.size()); 
    return *(--((--dMap.end())->second.end())); 
  }

  bool find (const T *dataPtr, _TreeNodeIndex<T,Cmp,_UTreeNode> &foundIdx) {
    if ((foundIdx.dmi=dMap.find(const_cast<T*>(dataPtr)))!=dMap.end()) {
      foundIdx.si=foundIdx.dmi->second.begin();
      return true;
    }
    else return false;
  }

  bool isFirstD () const {
    if (mother) return (mother->frontPtr()==this);
    else return true;
  }

  bool isLastD () const {
    if (mother) return (mother->backPtr()==this);
    else return true;
  }

  // update the payload and ensure the correct indices
  void update (const T &d) {
    if (mother) {
      mother->dMap.erase(this);
      mother->dMap[&(data=d)].insert(this);
    }
    else data=d;
  }

  // remove a daughter tnPtr
  void erase (_UTreeNode *tnPtr) { dMap.erase(tnPtr); }

  // clear all daughters
  void clear () { dMap.clear(); }

  // insert a daughter 'newPtr' before 'posPtr'
  // posPtr doesn't matter really for unordered tree nodes - it gets included
  // because we want a consistent interface across both unordered and ordered tree nodes
  void insert (_UTreeNode *posPtr, _UTreeNode *newPtr) {
    if ((newPtr->dmi=dMap.find(&(newPtr->data)))==dMap.end())
      newPtr->dmi=(dMap.insert(make_pair(&(newPtr->data),_TNPtrSet<_UTreeNode>()))).first;
    newPtr->si=(newPtr->dmi->second.insert(newPtr)).first;
  }

  // insert a daughter 'newPtr' at the beginning
  // for unordered tree nodes it's the same as insert()
  void push_front (_UTreeNode *newPtr) { insert(NULL,newPtr); }

  // replace the old daughter "oldPtr" with the new daughter "newPtr"
  void replaceDaughter (_UTreeNode *oldPtr, _UTreeNode *newPtr) {
    dMap.erase(oldPtr);
    insert(NULL,newPtr);
  }
};
  
// ===========================================================================
// INTERNAL USE
// class _TreeNode (internal use ONLY): an ordered tree node
// 'T' is the type of the payload, and 'Cmp' is a binary predicate "less-than"
// to compare two pieces of data.
// ===========================================================================

template <class T, class Cmp>
class _TreeNode {
  // preserving the sibling ordering
  _TNPtrList<_TreeNode> daughters;  

  // map of daughters for quick lookup
  _DMap<T,Cmp,_TreeNode> dMap;

  // the position in mother->daughters
  typename _TNPtrList<_TreeNode>::iterator di;

public:
  typedef size_t size_type;

  _TreeNode *mother;  
  size_type depth;
  T data;

  _TreeNode (const T &data, _TreeNode *mother=NULL)
    :mother(mother),data(data) {
    if (mother) depth=mother->depth+1;
    else depth=0;
  }

  size_type size() const { return daughters.size(); }

  // WARNING: make sure the element exists
  const T& front () const { assert(daughters.size()); return daughters.front()->data; }
  const T& back () const { assert(daughters.size()); return daughters.back()->data; }
  
  _TreeNode *previous () const {  // return the previous sibling
    assert(mother);
    assert(di!=mother->daughters.begin());
    typename _TNPtrList<_TreeNode>::iterator prevDI;

    --(prevDI=di);
    return *prevDI;
  }
  _TreeNode *next () const {     // return the next sibling
    assert(mother);
    typename _TNPtrList<_TreeNode>::iterator nextDI;
    
    ++(nextDI=di);
    return (nextDI!=mother->daughters.end()?*nextDI:NULL);
  }

  _TreeNode *frontPtr () const { assert(daughters.size()); return daughters.front(); }
  _TreeNode *backPtr () const { assert(daughters.size()); return daughters.back(); }

  bool find (const T *dataPtr, _TreeNodeIndex<T,Cmp,_TreeNode> &foundIdx) {
    if ((foundIdx.dmi=dMap.find(const_cast<T*>(dataPtr)))!=dMap.end()) {
      foundIdx.si=foundIdx.dmi->second.begin();
      return true;
    }
    else return false;
  }

  bool isFirstD () const {
    if (mother) return (mother->daughters.front()==this);
    else return true;
  }

  bool isLastD () const {
    if (mother) return (mother->daughters.back()==this);
    else return true;
  }

  // update the payload and ensure the correct indices
  void update (const T &d) {
    if (mother) {
      mother->dMap.erase(this);
      mother->dMap[&(data=d)].insert(this);
    }
    else data=d;
  }

  // remove a daughter tnPtr
  void erase (_TreeNode *tnPtr) {
    daughters.erase(tnPtr->di);
    dMap.erase(tnPtr);
  }

  // clear all daughters
  void clear () {
    daughters.clear();
    dMap.clear();
  }

  // insert a daughter 'newPtr' before 'posPtr'
  void insert (_TreeNode *posPtr, _TreeNode *newPtr) {
    newPtr->di=daughters.insert(posPtr?posPtr->di:daughters.end(),newPtr);
    dMap[&(newPtr->data)].insert(newPtr);
  }

  // insert a daughter 'newPtr' at the beginning
  void push_front (_TreeNode *newPtr) {
    daughters.push_front(newPtr);
    newPtr->di=daughters.begin();
    dMap[&(newPtr->data)].insert(newPtr);
  }

  // replace the old daughter "oldPtr" with the new daughter "newPtr"
  void replaceDaughter (_TreeNode *oldPtr, _TreeNode *newPtr) {
    newPtr->di=oldPtr->di;   // new daughter occupies the same position
    *(newPtr->di)=newPtr;

    dMap.erase(oldPtr);
    dMap[&(newPtr->data)].insert(newPtr);
  }
};

};

#endif
