// ======================================================================
// _vfHelper.hpp - helpers for VFile (for portability on brain-dead M$ VC6):
//                 INTERNAL USE ONLY.
// 061604: Benjamin Han <benhdj@cs.cmu.edu> Added a virtual destructor to 
//         VFObject.
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _VFHELPER_HPP
#define _VFHELPER_HPP

#include <vector>
#include <list>
#include <set>
#include <map>

namespace Toolbox {

#define VFILE_SIZETYPE int

class VFile;

// For all user-defined types they must inherit from VFObject and provide
// implementations for writeToVFile(VFile&) and readFromVFile(VFile&)!
struct VFObject {
  virtual void writeToVFile (VFile &vf) const=0;
  virtual void readFromVFile (VFile &vf)=0;

  virtual ~VFObject () {};  // must have a virtual destructor if we have virtual functions
};

template <class T>
struct _VFHelper {

  typedef VFILE_SIZETYPE size_type;

  template <class U>
  static void _writeSequence (const U &seq, VFile &vf) {
    typename U::size_type i,s=seq.size();
    typename U::const_iterator si;
    
    vf.write(s);
    for (i=0,si=seq.begin();i<s;++i,++si) vf.write(*si);
  }
  
  template <class U>
  static void _readSequence (U &seq, VFile &vf) {
    typename U::size_type i,s;
    typename U::iterator si;
    
    vf.read(s);
    seq.resize(s);
    for (i=0,si=seq.begin();i<s;++i,++si) vf.read(*si);
  }
  
  template <class U>
  static void _writeMap (const U &m, VFile &vf) {
    typename U::size_type i,s=m.size();
    typename U::const_iterator mi;
    
    vf.write(s);
    for (i=0,mi=m.begin();i<s;++i,++mi) {
      vf.write(mi->first);
      vf.write(mi->second);
    }
  }

  // ====== Writers for non-basic datatypes ======
  
  // TO-DO: for basic datatypes (e.g., int[]) this is inefficient since we
  //        can "block-write" - but this is necessary for user-defined types
  //        maybe implement more specialized _VFHelper<>?
  template <class U>
  static void write (const U ptr[], size_type size, VFile &vf) {
    vf.write(size);
    for (size_type i=0;i<size;++i) vf.write(ptr[i]);
  }
  
  template <class U, class V>
  static void write (const std::vector<U,V> &v, VFile &vf) { _writeSequence(v,vf); }
  
  template <class U, class V>
  static void write (const std::list<U,V> &l, VFile &vf) { _writeSequence(l,vf); }
  
  template <class U, class V, class W>
  static void write (const std::set<U,V,W> &s, VFile &vf) { _writeSequence(s,vf); }
  
  template <class U, class V, class W>
  static void write (const std::multiset<U,V,W> &s, VFile &vf) { _writeSequence(s,vf); }
  
  template <class U, class V, class W, class X>
  static void write (const std::map<U,V,W,X> &m, VFile &vf) { _writeMap(m,vf); }
  
  template <class U, class V, class W, class X>
  static void write (const std::multimap<U,V,W,X> &m, VFile &vf) { _writeMap(m,vf); }
  
  static void write (const VFObject &obj, VFile &vf) { obj.writeToVFile(vf); }
  
  // ====== Readers for non-basic datatypes ======
  
  // TO-DO: for basic datatypes (e.g., int[]) this is inefficient since we
  //        can "block-read" - but this is necessary for user-defined types
  //        maybe implement more specialized _VFHelper<>?
  template <class U>
  static void read (U *&ptr, size_type &size, VFile &vf) {
    vf.read(size);
    ptr=new U[size];
    for (size_type i=0;i<size;++i) vf.read(ptr[i]);
  }
  
  template <class U, class V>
  static void read (std::vector<U,V> &v, VFile &vf) { _readSequence(v,vf); }
  
  template <class U, class V>
  static void read (std::list<U,V> &l, VFile &vf) { _readSequence(l,vf); }
  
  template <class U, class V, class W>
  static void read (std::set<U,V,W> &s, VFile &vf) {
    U u;
    typename std::set<U,V,W>::size_type i,size;
    
    vf.read(size);
    for (i=0;i<size;++i) {
      vf.read(u);
      s.insert(u);
    }
  }
  
  template <class U, class V, class W>
  static void read (std::multiset<U,V,W> &s, VFile &vf) {
    U u;
    typename std::multiset<U,V,W>::size_type i,size;
    
    vf.read(size);
    for (i=0;i<size;++i) {
      vf.read(u);
      s.insert(u);
    }
  }
  
  template <class U, class V, class W, class X>
  static void read (std::map<U,V,W,X> &m, VFile &vf) {
    typename std::map<U,V>::size_type i,s;
    U u;
    V v;
    
    vf.read(s);
    for (i=0;i<s;++i) {
      vf.read(u);
      vf.read(v);
      m.insert(make_pair(u,v));
    }
  }
  
  template <class U, class V, class W, class X>
  static void read (std::multimap<U,V,W,X> &m, VFile &vf) {
    typename std::multimap<U,V>::size_type i,s;
    U u;
    V v;
    
    vf.read(s);
    for (i=0;i<s;++i) {
      vf.read(u);
      vf.read(v);
      m.insert(make_pair(u,v));
    }
  }
  
  static void read (VFObject &obj, VFile &vf) { obj.readFromVFile(vf); }
};
  
};

#endif
