// ======================================================================
// cfgAnalysis.hpp - an anlysis context-free grammar.
// 
// 061504: Benjamin Han <benhdj@cs.cmu.edu> bug fixed in _GenRule::readRHS():
//         the root of a trie has nothing - corrected the loop.
// 060704: Benjamin Han <benhdj@cs.cmu.edu> tmpI in CFGAnalysis::Iterator::readRHS()
//         should be an Index::const_iterator, not Index::ConstIterator.
// 122502: Benjamin Han <benhdj@cs.cmu.edu> No MS VC6 support for this
//         module.
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120602: Benjamin Han <benhdj@cs.cmu.edu> Created (total rewrite).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// So far VC6 has problems with too long symbol names (warning C4503), which
// truncated names and made them non-unique. The only solution now is NOT
// PROVIDING ANALYSIS GRAMMAR! (duh)
#if _MSC_VER!=1200

#ifndef CFGANALYSIS_HPP
#define CFGANALYSIS_HPP

#include "_cfgBase.hpp"
#include "trie.hpp"

#include <set>
#include <map>

namespace Toolbox {


// ===========================================================================
// class _AIndexEntry (internal use ONLY): an entry in _AIndex
// 1. wrap around the pointer - because we want to initialize it to NULL
// 2. it's the implicit payload of the trie so storage is saved compared
//    to saving _IndexEntry<Rules> in each Trie node.
// ===========================================================================

template <class Rules>
class _AIndexEntry {
  _IndexEntry<Rules> *ptr;

public:

  _AIndexEntry ():ptr(NULL) {}
  _AIndexEntry (_IndexEntry<Rules> *ptr):ptr(ptr) {}

  // only delete if it's set before
  void clear () {
    if (ptr) {
      delete ptr;
      ptr=NULL;
    }
  }

  bool isSet () const { return ptr!=NULL; }

  _IndexEntry<Rules>* operator -> () { assert(ptr); return ptr; }
  const _IndexEntry<Rules> * operator -> () const { assert(ptr); return ptr; }
};


// ===========================================================================
// class _AIndex (internal use ONLY): index for analysis grammar (key: RHS)
// ===========================================================================

template <class LLiteral, class RLiteral, class Rules>
class _AIndex: protected Trie<RLiteral,_AIndexEntry<Rules> > {
public:

  typedef typename Trie<RLiteral,_AIndexEntry<Rules> >::Iterator iterator;
  typedef typename Trie<RLiteral,_AIndexEntry<Rules> >::ConstIterator const_iterator;
  typedef typename Trie<RLiteral,_AIndexEntry<Rules> >::PreIterator PreIterator;
  typedef CFG_SIZETYPE size_type;

  // TO-DO: this is a slow implementation; I could store all the pointers inside
  //        a list so I don't need to traverse the entire trie - but that takes memory
  void clear () {
    PreIterator pi;

    for (pi=begin();pi!=pi.nullTI();++pi) pi.get().data.clear();
    Trie<RLiteral,_AIndexEntry<Rules> >::clear();
  }
  ~_AIndex () { clear(); }

  bool find (const RHS<RLiteral> &k, iterator &ii,
	     typename Rules::iterator &riFirst, typename Rules::iterator &riLast) {
    ii=Trie<RLiteral,_AIndexEntry<Rules> >::findPrefix(k);

    if (ii!=ii.nullTI()) {
      riFirst=ii->data->riFirst;
      riLast=ii->data->riLast;
      return true;
    }
    return false;
  }

  // key-neutral search; ii points to the matching entry, if any
  bool find (const LLiteral &lhs, const RHS<RLiteral> &rhs, iterator &ii,
	     typename Rules::iterator &riFirst, typename Rules::iterator &riLast) {
    return find(rhs,ii,riFirst,riLast);
  }

  void add (const LLiteral &lhs, const RHS<RLiteral> &rhs, typename Rules::iterator ri) {
    Trie<RLiteral,_AIndexEntry<Rules> >::addPrefix(rhs).get().data=
      _AIndexEntry<Rules>(new _IndexEntry<Rules>(ri));
  }

  void add (iterator ii, typename Rules::iterator ri) {
    ii.get().data->riLast=ri;
  };

#ifdef WITH_REMOVE_RULE
  void remove (iterator ii, typename Rules::iterator ri) {
    assert(ii->data.isSet());
    typename Rules::iterator &riFirst=ii.get().data->riFirst;
    typename Rules::iterator &riLast=ii.get().data->riLast;

    if (ri!=riFirst && ri!=riLast) return;
    else if (riFirst==riLast) {
      iterator tmpI;

      ii.get().data.clear();
      // note the root of the trie is a pseudo node
      while (!ii.isRoot() && ii.size()==0 && !ii->data.isSet()) {
	tmpI=ii;
	ii.up();
	prune(tmpI);
      }
    }
    else if (ri==riFirst) ++riFirst;
    else --riLast;
  }
#endif

  size_type trieSize () const { return Trie<RLiteral,_AIndexEntry<Rules> >::size(); }

  iterator begin () { return Trie<RLiteral,_AIndexEntry<Rules> >::begin(); }
  iterator end () { return iterator::nullTI(); }
};

// ===========================================================================
// class _ARule (internal use ONLY): an analysis rule
// ===========================================================================

template <class LLiteral, class Body>
class _ARule {

#if _MSC_VER==1200
public:
#else
  template <class LLiteral_, class RLiteral_, class Body_> friend class _ARules;
#endif

#ifdef WITH_REMOVE_RULE
  typedef std::map<LLiteral,CFG_SIZETYPE> LHSTable;
#else
  typedef std::set<LLiteral> LHSTable;
#endif

  typename LHSTable::iterator ti;

#ifdef WITH_REMOVE_RULE
  // remove the LHS in the lhsTable
  void remove (LHSTable &lhsTable) {
    assert(ti->second>0);
    if (ti->second>1) --(ti->second);
    else lhsTable.erase(ti);
  }
#endif

public:

  Body body;

  _ARule (const typename LHSTable::iterator ti):ti(ti) {}  

  const LLiteral &readLHS () const { 
#ifdef WITH_REMOVE_RULE
    return ti->first;
#else
    return *ti;
#endif
  }
};


// ===========================================================================
// class _ARules (internal use ONLY): a list of analysis rules
// ===========================================================================

template <class LLiteral, class RLiteral, class Body>
class _ARules: public std::list<_ARule<LLiteral,Body> > {
  typedef std::list<_ARule<LLiteral,Body> > _Parent;
  typedef typename _ARule<LLiteral,Body>::LHSTable LHSTable;

  LHSTable lhsTable;

public:
  typedef _ARule<LLiteral,Body> Rule;
  typedef typename _Parent::iterator iterator;
  typedef CFG_SIZETYPE size_type;

  // adding rules with the hinted range
  // ASSUMPTION: the RHS of the new rule was already added before this point
  iterator add (const LLiteral &lhs, const RHS<RLiteral> &rhs,
		iterator riFirst, iterator riLast) {    
#ifdef WITH_REMOVE_RULE
    typename LHSTable::iterator ti=lhsTable.insert(make_pair(lhs,0)).first;
    ++(ti->second);
    return _Parent::insert(++riLast,Rule(ti));
#else
    return _Parent::insert(++riLast,Rule(lhsTable.insert(lhs).first));
#endif
  }

  // ASSUMPTION: the RHS of the new rule was *not* added before this point
  iterator add (const LLiteral &lhs, const RHS<RLiteral> &rhs) {
#ifdef WITH_REMOVE_RULE
    typename LHSTable::iterator ti=lhsTable.insert(make_pair(lhs,0)).first;
    ++(ti->second);
    _Parent::push_back(Rule(ti));
#else
    _Parent::push_back(Rule(lhsTable.insert(lhs).first));
#endif
    return --_Parent::end();
  }

#ifdef WITH_REMOVE_RULE
  void remove (iterator ri) {
    ri->remove(lhsTable);
    _Parent::erase(ri);
  }
#endif

  size_type lhsTableSize () const { return lhsTable.size(); }

  void clear () {
    lhsTable.clear();
    _Parent::clear();
  }
};


// ===========================================================================
// class CFGAnalysis: Context-free Grammar for Analysis
// ===========================================================================

template <class LLiteral, class RLiteral, class Body>
class CFGAnalysis: public _CFGBase<LLiteral,RLiteral,Body,RHS<RLiteral>,
				   _AIndex<LLiteral,RLiteral,_ARules<LLiteral,RLiteral,Body> >,
				   _ARules<LLiteral,RLiteral,Body> > {
public:

  typedef CFG_SIZETYPE size_type;
  typedef _AIndex<LLiteral,RLiteral,_ARules<LLiteral,RLiteral,Body> > Index;
  typedef typename _CFGBase<LLiteral,RLiteral,Body,RHS<RLiteral>,
			    _AIndex<LLiteral,RLiteral,_ARules<LLiteral,RLiteral,Body> >,
			    _ARules<LLiteral,RLiteral,Body> >::RuleIterator RuleIterator;

  // a non-const iterator for sifting rules grouped by keys
  class Iterator {
    typename Index::PreIterator i;
  public:
    Iterator () {}
    Iterator (typename Index::PreIterator i):i(i) {}

    Iterator &operator ++ () {
      while (++i!=i.nullTI() && !i->data.isSet());
      return *this;
    }
    Iterator operator ++ (int) {
      Iterator tmpI(i);
      while (++i!=i.nullTI() && !i->data.isSet());
      return tmpI;
    }

    bool operator == (const Iterator other) const { return i==other.i; }
    bool operator != (const Iterator other) const { return i!=other.i; }

    RHS<RLiteral> readRHS () const {	
      RHS<RLiteral> rhs;
      typename Index::const_iterator tmpI;
      
      // the root node of a trie has nothing there
      for (tmpI=i;!tmpI.isRoot();tmpI.up())
	rhs.push_front(tmpI->label);
      
      return rhs;
    }

    RuleIterator first () {
      assert(i->data.isSet());
      RuleIterator ri=i->data->riFirst;
      ri.ii=i;
      return ri;
    }
    RuleIterator last () {
      assert(i->data.isSet());
      RuleIterator ri=i->data->riLast;
      ri.ii=i;
      return ri;
    }
  };

  Iterator begin () {
    Iterator i(index.begin());  // the root node of the trie is a pseudo node
    return ++i;
  }
  Iterator end () { return Iterator(index.end()); }

  // return # of unique LHS literals
  size_type lhsSize() const { return rules.lhsTableSize(); }

  // return # of trie nodes used to store all RHSs
  size_type trieSize() const { return index.trieSize(); }
};

};

#endif

#endif
