// ======================================================================
// cfgGen.hpp - a generation context-free grammar.
// 
// 061504: Benjamin Han <benhdj@cs.cmu.edu> bug fixed in _GenRule::readRHS():
//         the root of a trie has nothing - corrected the loop.
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120602: Benjamin Han <benhdj@cs.cmu.edu> Created (total rewrite).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef CFGGEN_HPP
#define CFGGEN_HPP

#include "_cfgBase.hpp"
#include "trie.hpp"

#ifndef WITH_REMOVE_RULE
#include "labelTrie.hpp"
#endif

#include <map>

namespace Toolbox {

// ===========================================================================
// class _GenIndex (internal use ONLY): index for generation grammar (key: LHS)
// ===========================================================================

template <class LLiteral, class RLiteral, class Rules>
class _GenIndex: protected std::map<LLiteral,_IndexEntry<Rules> > {
  typedef std::map<LLiteral,_IndexEntry<Rules> > _Parent;

public:

  typedef typename _Parent::iterator iterator;
  typedef typename _Parent::const_iterator const_iterator;
  typedef CFG_SIZETYPE size_type;

  bool find (const LLiteral &k, iterator &ii,
	     typename Rules::iterator &riFirst, typename Rules::iterator &riLast) {
    ii=_Parent::find(k);

    if (ii!=_Parent::end()) {
      riFirst=ii->second.riFirst;
      riLast=ii->second.riLast;
      return true;
    }
    return false;
  }

  // key-neutral search (hence direction-neutral); ii points to the matching entry (if any)
  bool find (const LLiteral &lhs, const RHS<RLiteral> &rhs, iterator &ii,
	     typename Rules::iterator &riFirst, typename Rules::iterator &riLast) {
    return find(lhs,ii,riFirst,riLast);
  }

  void add (const LLiteral &lhs, const RHS<RLiteral> &rhs, typename Rules::iterator ri) {
#ifndef NDEBUG
    std::pair<iterator,bool> result=
#endif
      _Parent::insert(make_pair(lhs,_IndexEntry<Rules>(ri)));

    assert(result.second);
  }

  void add (iterator ii, typename Rules::iterator ri) {
    ii->second.riLast=ri;
  };

#ifdef WITH_REMOVE_RULE
  void remove (iterator ii, typename Rules::iterator ri) {
    typename Rules::iterator &riFirst=ii->second.riFirst;
    typename Rules::iterator &riLast=ii->second.riLast;

    if (ri!=riFirst && ri!=riLast) return;
    else if (riFirst==riLast) _Parent::erase(ii);
    else if (ri==riFirst) ++riFirst;
    else --riLast;
  }
#endif

  size_type size () const { return _Parent::size(); }

  iterator begin () { return _Parent::begin(); }
  iterator end () { return _Parent::end(); }

  void clear () { _Parent::clear(); }
};


// ===========================================================================
// class _GenRule (internal use ONLY): a generation rule
// ===========================================================================

template <class RLiteral, class Body>
class _GenRule {

#if _MSC_VER==1200
public:
#else
  template <class LLiteral_, class RLiteral_, class Body_> friend class _GenRules;
#endif

#ifdef WITH_REMOVE_RULE
  typename Trie<RLiteral,CFG_SIZETYPE>::Iterator rti;
#else
  typename LabelTrie<RLiteral>::Iterator rti;  // points to the end of an RHS
#endif

#ifdef WITH_REMOVE_RULE
  // remove the RHS in rhsTrie
  void remove (Trie<RLiteral,CFG_SIZETYPE> &rhsTrie) {
    assert(rti->data>0);
    typename Trie<RLiteral,CFG_SIZETYPE>::Iterator tmpRTI;

    --(rti.get().data);
    // note the root of the trie is a pseudo node
    while (!rti.isRoot() && rti.size()==0 && rti->data==0) {
      tmpRTI=rti;
      rti.up();
      rhsTrie.prune(tmpRTI);
    }
  }
#endif

public:

  Body body;

#ifdef WITH_REMOVE_RULE
  _GenRule (typename Trie<RLiteral,CFG_SIZETYPE>::Iterator rti):rti(rti) {}  
#else
  _GenRule (typename LabelTrie<RLiteral>::Iterator rti):rti(rti) {}  
#endif

  RHS<RLiteral> readRHS () const {
    RHS<RLiteral> rhs;

#ifdef WITH_REMOVE_RULE
    typename Trie<RLiteral,CFG_SIZETYPE>::ConstIterator tmpRTI;
#else
    typename LabelTrie<RLiteral>::ConstIterator tmpRTI;
#endif
    
    // the root node of a trie has nothing there
    for (tmpRTI=rti;!tmpRTI.isRoot();tmpRTI.up()) {
#ifdef WITH_REMOVE_RULE
      rhs.push_front(tmpRTI->label);
#else
      rhs.push_front(*tmpRTI);
#endif
    }

    return rhs;
  }
};


// ===========================================================================
// class _GenRules (internal use ONLY): a list of generation rules
// ===========================================================================

template <class LLiteral, class RLiteral, class Body>
class _GenRules: public std::list<_GenRule<RLiteral,Body> > {
  typedef std::list<_GenRule<RLiteral,Body> > _Parent;

#ifdef WITH_REMOVE_RULE
  Trie<RLiteral,CFG_SIZETYPE> rhsTrie;
#else
  LabelTrie<RLiteral> rhsTrie;
#endif

public:
  typedef _GenRule<RLiteral,Body> Rule;
  typedef typename _Parent::iterator iterator;
  typedef CFG_SIZETYPE size_type;

  // adding rules with the hinted range
  // ASSUMPTION: the LHS of the new rule was already added before this point
  typename _Parent::iterator add (const LLiteral &lhs, const RHS<RLiteral> &rhs,
				  iterator riFirst, iterator riLast) {    
#ifdef WITH_REMOVE_RULE
    typename Trie<RLiteral,CFG_SIZETYPE>::Iterator ti=rhsTrie.addPrefix(rhs,0);
    ++(ti.get().data);
    return _Parent::insert(++riLast,Rule(ti));
#else
    return _Parent::insert(++riLast,Rule(rhsTrie.addPrefix(rhs)));
#endif
  }

  // ASSUMPTION: the LHS of the new rule was *not* added before this point
  typename _Parent::iterator add (const LLiteral &lhs, const RHS<RLiteral> &rhs) {
#ifdef WITH_REMOVE_RULE
    typename Trie<RLiteral,CFG_SIZETYPE>::Iterator ti=rhsTrie.addPrefix(rhs,0);
    ++(ti.get().data);
    _Parent::push_back(Rule(ti));
#else
    _Parent::push_back(Rule(rhsTrie.addPrefix(rhs)));
#endif
    return --_Parent::end();
  }

#ifdef WITH_REMOVE_RULE
  void remove (iterator ri) {
    ri->remove(rhsTrie);
    _Parent::erase(ri);
  }
#endif

  size_type trieSize () const { return rhsTrie.size(); }

  void clear () {
    rhsTrie.clear();
    _Parent::clear();
  }
};


// ===========================================================================
// class CFGGen: Context-free Grammar for Generation
// ===========================================================================

template <class LLiteral, class RLiteral, class Body>
class CFGGen: public _CFGBase<LLiteral,RLiteral,Body,LLiteral,
			      _GenIndex<LLiteral,RLiteral,_GenRules<LLiteral,RLiteral,Body> >,
			      _GenRules<LLiteral,RLiteral,Body> > {
public:

  typedef CFG_SIZETYPE size_type;
  typedef _GenIndex<LLiteral,RLiteral,_GenRules<LLiteral,RLiteral,Body> > Index;
  typedef typename _CFGBase<LLiteral,RLiteral,Body,LLiteral,
			    _GenIndex<LLiteral,RLiteral,_GenRules<LLiteral,RLiteral,Body> >,
			    _GenRules<LLiteral,RLiteral,Body> >::RuleIterator RuleIterator;

  // a non-const iterator for sifting rules grouped by keys
  class Iterator {
    typename Index::iterator i;
  public:
    Iterator () {}
    Iterator (typename Index::iterator i):i(i) {}

    Iterator &operator ++ () { ++i; return *this; }
    Iterator operator ++ (int) { Iterator tmpI(i++); return tmpI; }

    bool operator == (const Iterator other) const { return i==other.i; }
    bool operator != (const Iterator other) const { return i!=other.i; }

    const LLiteral &readLHS () const { return i->first; }

    RuleIterator first () {
      RuleIterator ri=i->second.riFirst;
      ri.ii=i;
      return ri;
    }
    RuleIterator last () {
      RuleIterator ri=i->second.riLast;
      ri.ii=i;
      return ri;
    }
  };

  Iterator begin () { return Iterator(index.begin()); }
  Iterator end () { return Iterator(index.end()); }

  // return # of unique LHS/RHS literals
  size_type lhsSize() const { return index.size(); }

  // return # of trie nodes used to store all RHSs
  size_type trieSize() const { return rules.trieSize(); }
};

};

#endif
