// ======================================================================
// graph.hpp - A generic vertex-indexed, directed and single-edge graph.
// 
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Made Graph<>::copy() public; minor 
//         revisions to supress warnings under -pedantic -Wall (mostly adding
//         'typename').
// 082501: Benjamin Han <benhdj@cs.cmu.edu> Revised
//         eraseOne(_GIterator<VT,ET,VT&,VT*,Agenda_,Cmp>) and prune() in 
//         Graph<> to work with the revised PreorderAgenda and PostorderAgenda;
//         made GIsomorphism<> in graphAlgorithm a friend of Graph<>.
// 082101: Benjamin Han <benhdj@cs.cmu.edu> Added new iterators - data iterators
//         DIterator and ConstDIterator, and vertex iterators VIterator and
//         ConstVIterator; added dBegin(), dEnd(), dFind() and dSize() in 
//         Graph<>.
// 081301: Benjamin Han <benhdj@cs.cmu.edu> Abandoned the idea of using stamp
//         for traversal (keeping stamps in each vertex makes multiple
//         traversals at the same time infeasible) - removed stamp from Graph<>.
// 080901: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef GRAPH_HPP
#define GRAPH_HPP

#include "graphIter.hpp"

#include <functional>

// ===========================================================================
// class Graph: a generic vertex-indexed, directed and single-edge graph 
// ===========================================================================

namespace Toolbox {

template <class VT, class ET, class Cmp=std::less<VT> >
class Graph {
public:

  typedef typename _GVMap<VT,ET,Cmp>::size_type size_type;

  // all of the iterator types for easy access
  // NOTE: only DIterator/ConstDIterator and NIterator/ConstNIterator are
  //       bidirectional iterators (i.e., can be decremented).
  typedef GIteratorBase(VT,ET,Cmp) Iterator;
  typedef GConstIteratorBase(VT,ET,Cmp) ConstIterator;
  typedef GDIterator(VT,ET,Cmp) DIterator;
  typedef GConstDIterator(VT,ET,Cmp) ConstDIterator;
  typedef GVIterator(VT,ET,Cmp) VIterator;
  typedef GConstVIterator(VT,ET,Cmp) ConstVIterator;
  typedef GNIterator(VT,ET,Cmp) NIterator;
  typedef GConstNIterator(VT,ET,Cmp) ConstNIterator;
  typedef GIterator(VT,ET,PreorderAgenda,Cmp) PreIterator;
  typedef GConstIterator(VT,ET,PreorderAgenda,Cmp) ConstPreIterator;
  typedef GIterator(VT,ET,PostorderAgenda,Cmp) PostIterator;
  typedef GConstIterator(VT,ET,PostorderAgenda,Cmp) ConstPostIterator;
  typedef GIterator(VT,ET,BFSAgenda,Cmp) BFSIterator;
  typedef GConstIterator(VT,ET,BFSAgenda,Cmp) ConstBFSIterator;
  typedef GIterator(VT,ET,FrontAgenda,Cmp) FrontIterator;
  typedef GConstIterator(VT,ET,FrontAgenda,Cmp) ConstFrontIterator;
  typedef GPathIterator(VT,ET,Cmp) PathIterator;
  typedef GConstPathIterator(VT,ET,Cmp) ConstPathIterator;

private:

  typedef _Vertex<VT,ET,Cmp> Vertex;
  typedef Vertex* VPtr;

  _GVMap<VT,ET,Cmp> vMap;

  std::set<VPtr> emptyVSet;

  // for remembering for findNext()
  typename _GVMap<VT,ET,Cmp>::iterator vmi;
  typename std::set<VPtr>::iterator vsi;

#if _MSC_VER!=1200

  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_>
  friend class _GIteratorBase;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_>
  friend class _GDIterator;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Agenda_, class Cmp_>
  friend class _GIterator;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_>
  friend class Agenda;

  // allow access for some algorithm classes for efficiency
  template <class VT_, class ET_, class Cmp_, class EOp_>
  friend class _GIsoBase;

#endif

public:

  void clear ();

  Graph () {}
  // ASSUMPTION: VT has its copy constructor defined
  Graph (const Graph &g) { copy(g); }
  ~Graph () { clear(); }

  // NOTE begin() returns an Iterator so you can't expect
  // begin()++ works.
  Iterator begin () {
    typename _GVMap<VT,ET,Cmp>::iterator vmi;  // shadowing Graph::vmi but that's ok

    if ((vmi=vMap.begin())!=vMap.end()) return Iterator(this,*(vmi->second.begin()));
    else return Iterator(this,NULL);   // effectively set gi to nullGI
  }
  ConstIterator begin () const {
    typename _GVMap<VT,ET,Cmp>::const_iterator vmi;  // shadowing Graph::vmi but that's ok

    if ((vmi=vMap.begin())!=vMap.end())
      // TO-DO: do we need to pass a non-const Graph* here?
      return Iterator(const_cast<Graph*>(this),*(vmi->second.begin()));
    else return ConstIterator(const_cast<Graph*>(this),NULL);   // effectively set gi to nullGI
  }
  DIterator dBegin () {
    return DIterator(this,vMap.begin());
  }
  ConstDIterator dBegin () const {
    return DIterator(const_cast<Graph*>(this),const_cast<_GVMap<VT,ET,Cmp>*>(&vMap)->begin());
  }
  DIterator dEnd () {
    return DIterator(this,vMap.end());
  }
  ConstDIterator dEnd () const {
    return DIterator(const_cast<Graph*>(this),const_cast<_GVMap<VT,ET,Cmp>*>(&vMap)->end());
  }

  DIterator dFind (const VT &data) {
    DIterator di(this,vMap.find(&data));
    return di;
  }
  ConstDIterator dFind (const VT &data) const {
    ConstDIterator di(this,vMap.find(&data));
    return di;
  }

  size_type size () const { return vMap.size(); }    // # of vertices
  size_type eSize () const;                          // # of edges
  size_type dSize () const { return vMap.dSize(); }  // # of different vertex data

  // add(const VT&):     add a single vertex.
  // add(ConstIterator): add all reachable vertices and the edges from pos; return the
  //                     iterator corresponding to pos in the newly created subgraph.
  Iterator add (const VT &data);
  Iterator add (ConstIterator pos);

  // eraseOne(): erase a single vertex and its edges, and update pos to point to the
  //             'next' vertex, depending on the particular vertex ordering
  NIterator eraseOne (NIterator pos);
  template <class Agenda_>
  _GIterator<VT,ET,VT&,VT*,Agenda_,Cmp> eraseOne (_GIterator<VT,ET,VT&,VT*,Agenda_,Cmp> pos);
  
  // pruneOne(): prune a single vertex and its edges
  // prune():    prune all reachable vertices and the edges from pos
  void pruneOne (Iterator pos);
  void prune (Iterator pos);

  Iterator find (const VT &data) {  // find vertex 
    if ((vmi=vMap.find(&data))!=vMap.end()) {
      vsi=vmi->second.begin();
      return Iterator(this,*vsi);
    }
    else return Iterator(this,NULL);
  }  
  Iterator findNext () {            // WARNING: bomb if find() is not called first
    if ((++vsi)!=vmi->second.end()) return Iterator(this,*vsi);
    else return Iterator(this,NULL);
  }

  // ASSUMPTION: VT has its copy constructor defined
  // WARNING: won't remember where to findNext()
  void copy (const Graph &g);

  // ASSUMPTION: VT has its copy constructor defined
  // WARNING: won't remember where to findNext()!
  void operator = (const Graph &g) { clear(); copy(g); }
};

template <class VT, class ET, class VRef, class VPtr, class Cmp>
void _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::update (const VT &data) const {
  assert(gPtr && vPtr);

  typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;

  // deleting the old information
  gPtr->vMap.erase(vPtr);
  for (vmi=vPtr->inMap.begin();vmi!=vPtr->inMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi)
      emi->first->outMap.erase(vPtr);
  for (vmi=vPtr->outMap.begin();vmi!=vPtr->outMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi)
      emi->first->inMap.erase(vPtr);

  // assign the new data
  vPtr->data=data;

  // updating _Graph<>::vMap, and optionally _Vertex<>::vmi and _Vertex<>::vsi
#ifdef GRAPH_FAST_REMOVE
  std::pair<typename std::set<_Vertex<VT,ET,Cmp>*>::iterator, bool> result;

  if ((vPtr->vmi=gPtr->vMap.find(&(vPtr->data)))==gPtr->vMap.end()) {
    std::pair<typename _GVMap<VT,ET,Cmp>::iterator, bool> result=
      gPtr->vMap.insert(make_pair(&(vPtr->data),std::set<_Vertex<VT,ET,Cmp>*>()));
    vPtr->vmi=result.first;
  }
  result=vPtr->vmi->second.insert(vPtr);
  vPtr->vsi=result.first;
#else
  gPtr->vMap[&(vPtr->data)].insert(vPtr);
#endif

  // updating the inward and the outward vertices
  for (vmi=vPtr->inMap.begin();vmi!=vPtr->inMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi)
      emi->first->outMap[&(vPtr->data)].insert(make_pair(vPtr,emi->second));
  for (vmi=vPtr->outMap.begin();vmi!=vPtr->outMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi)
      emi->first->inMap[&(vPtr->data)].insert(make_pair(vPtr,emi->second));
}

template <class VT, class ET, class Cmp>
struct _TMapEntry {
  _Vertex<VT,ET,Cmp> *newVPtr;
  bool init;
  
  _TMapEntry () {}
  _TMapEntry (_Vertex<VT,ET,Cmp> *newVPtr, bool init):newVPtr(newVPtr),init(init) {}
};

template <class VT, class ET, class Cmp>
void Graph<VT,ET,Cmp>::clear () {
  // shadowing Graph::vmi and Graph::vsi but that's ok
  typename _GVMap<VT,ET,Cmp>::iterator vmi;
  typename std::set<VPtr>::iterator vsi;
  typename Vertex::_VMap::iterator vvmi;
  typename Vertex::_EMap::iterator emi;

  // clear each vertex
  for (vmi=vMap.begin();vmi!=vMap.end();++vmi)
    for (vsi=vmi->second.begin();vsi!=vmi->second.end();++vsi) {
      // clear inward edges
      for (vvmi=(*vsi)->inMap.begin();vvmi!=(*vsi)->inMap.end();++vvmi)
	for (emi=vvmi->second.begin();emi!=vvmi->second.end();++emi) {
	  // only free the edge data ONCE!
	  (*vsi)->inMap.erase(emi->first);
	  emi->first->outMap.erase(*vsi,false);
	}
      
      // clear outward edges
      for (vvmi=(*vsi)->outMap.begin();vvmi!=(*vsi)->outMap.end();++vvmi)
	for (emi=vvmi->second.begin();emi!=vvmi->second.end();++emi)
	  // but do *nothing* for any self-looping edge - it's deleted already
	  if (emi->first!=*vsi) {
	    // only free the edge data ONCE!
	    emi->first->inMap.erase(*vsi);
	    (*vsi)->outMap.erase(emi->first,false);
	  }
      
      delete *vsi;
    }
  vMap.clear();
}

template <class VT, class ET, class Cmp>
typename Graph<VT,ET,Cmp>::size_type Graph<VT,ET,Cmp>::eSize () const {
  size_type s;

  // shadowing Graph::vmi and Graph::vsi but that's ok
  typename _GVMap<VT,ET,Cmp>::const_iterator vmi;
  typename std::set<VPtr>::const_iterator vsi;

  for (vmi=vMap.begin(),s=0;vmi!=vMap.end();++vmi)
    for (vsi=vmi->second.begin();vsi!=vmi->second.end();++vsi)
      s+=(*vsi)->outDegree();

  return s;  
}

template <class VT, class ET, class Cmp>
typename Graph<VT,ET,Cmp>::Iterator Graph<VT,ET,Cmp>::add (const VT &data) {
  VPtr vPtr=new Vertex(data);

#ifdef GRAPH_FAST_REMOVE
  std::pair<typename std::set<VPtr>::iterator, bool> result;

  if ((vPtr->vmi=vMap.find(&(vPtr->data)))==vMap.end()) {
    std::pair<typename _GVMap<VT,ET,Cmp>::iterator, bool> result=
      vMap.insert(make_pair(&(vPtr->data),emptyVSet));
    vPtr->vmi=result.first;
  }
  result=vPtr->vmi->second.insert(vPtr);
  vPtr->vsi=result.first;
#else
  vMap[&(vPtr->data)].insert(vPtr);
#endif

  return Iterator(this,vPtr);
}

template <class VT, class ET, class Cmp>
typename Graph<VT,ET,Cmp>::Iterator Graph<VT,ET,Cmp>::add (Graph<VT,ET,Cmp>::ConstIterator pos) {
  typedef std::map<VPtr,VPtr> TMap;  // pointer translation map

  ConstPreIterator gpi;
  Iterator result;
  TMap tMap;
  typename Vertex::_VMap::const_iterator vmi;
  typename Vertex::_EMap::const_iterator emi;
  VPtr newSrcPtr,newDestPtr;
  ET *newEPtr;

  // add all reachable vertices first
  for (gpi=pos;gpi!=gpi.nullGI;++gpi)
    if (gpi==pos)
      tMap.insert(make_pair(gpi.vPtr,(result=add(*gpi)).vPtr));
    else tMap.insert(make_pair(gpi.vPtr,add(*gpi).vPtr));

  // add all edges among the duplicated vertices
  for (gpi=pos;gpi!=gpi.nullGI;++gpi) {
    newSrcPtr=tMap[gpi.vPtr];
    const typename Vertex::_VMap &oldMap=gpi.vPtr->outMap;
    typename Vertex::_VMap &newSrcMap=newSrcPtr->outMap;
    
    for (vmi=oldMap.begin();vmi!=oldMap.end();++vmi)
      for (emi=vmi->second.begin();emi!=vmi->second.end();++emi) {
	newDestPtr=tMap[emi->first];
	newSrcMap[&(newDestPtr->data)].insert(make_pair(newDestPtr,newEPtr=new ET(*(emi->second))));
	newDestPtr->inMap[&(newSrcPtr->data)].insert(make_pair(newSrcPtr,newEPtr));
      }
  }
  
  return result;
}

template <class VT, class ET, class Cmp>
typename Graph<VT,ET,Cmp>::NIterator Graph<VT,ET,Cmp>::eraseOne (Graph<VT,ET,Cmp>::NIterator pos) {
  // it's safe to increment pos after pruning it since the iterator keeps the next vertex
  pruneOne(pos);
  return ++pos;
}

template <class VT, class ET, class Cmp>
template <class Agenda_>
_GIterator<VT,ET,VT&,VT*,Agenda_,Cmp> Graph<VT,ET,Cmp>::eraseOne (_GIterator<VT,ET,VT&,VT*,Agenda_,Cmp> pos) {
  _GIterator<VT,ET,VT&,VT*,Agenda_,Cmp> p;

  ++(p=pos);
  pruneOne(pos);
  return p;
}

template <class VT, class ET, class Cmp>
void Graph<VT,ET,Cmp>::pruneOne (Iterator pos) {
  assert(pos.gPtr==this && pos.vPtr);

  Vertex &v=*(pos.vPtr);
  typename Vertex::_VMap::iterator vmi;
  typename Vertex::_EMap::iterator emi;

  // removing inward edges
  for (vmi=v.inMap.begin();vmi!=v.inMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi) {
      // only free the edge data ONCE!
      v.inMap.erase(emi->first);
      emi->first->outMap.erase(&v,false);
    }
  
  // removing outward edges
  for (vmi=v.outMap.begin();vmi!=v.outMap.end();++vmi)
    for (emi=vmi->second.begin();emi!=vmi->second.end();++emi)
      // but do *nothing* for any self-looping edge - it's deleted already
      if (emi->first!=&v) {
	// only free the edge data ONCE!
	emi->first->inMap.erase(&v);
	v.outMap.erase(emi->first,false);
      }

#ifdef GRAPH_FAST_REMOVE
  vMap.erase(v.vmi,v.vsi);
#else
  vMap.erase(&v);
#endif
  delete &v;  
}

template <class VT, class ET, class Cmp>
inline void Graph<VT,ET,Cmp>::prune (Graph<VT,ET,Cmp>::Iterator pos) {
  for (PostIterator gpi=pos;gpi!=gpi.nullGI;gpi=eraseOne(gpi));
}

template <class VT, class ET, class Cmp>
void Graph<VT,ET,Cmp>::copy (const Graph &g) {
  typedef std::map<VPtr,_TMapEntry<VT,ET,Cmp> > TMap;  // pointer translation map

  // for the src graph g (the corresponding iterators for the target graph have
  // suffix '1')
  typename _GVMap<VT,ET,Cmp>::const_iterator vmi2;
  typename std::set<VPtr>::const_iterator vsi2;
  typename Vertex::_VMap::const_iterator vvmi2;
  typename Vertex::_EMap::const_iterator emi2;

  typename Vertex::_EMap::iterator emi1;

  typename Vertex::_EMap emptyEMap;
  std::pair<typename _GVMap<VT,ET,Cmp>::iterator,bool> vResult;
#ifdef GRAPH_FAST_REMOVE
  std::pair<typename std::set<VPtr>::iterator,bool> sResult;
#endif
  std::pair<typename Vertex::_VMap::iterator,bool> eResult;
  bool fResult;
  TMap tMap;
  typename TMap::iterator tmi;
  VPtr newVPtr1,newVPtr2;

  // for each different datum
  for (vmi2=g.vMap.begin();vmi2!=g.vMap.end();++vmi2) {
    vResult=vMap.insert(make_pair(vmi2->first,emptyVSet));
    typename _GVMap<VT,ET,Cmp>::iterator &vmi1=vResult.first;
    
    // for each vertex of the same data
    for (vsi2=vmi2->second.begin();vsi2!=vmi2->second.end();++vsi2) {
      if ((tmi=tMap.find(*vsi2))==tMap.end())
	// a new vertex needs to be created
	tMap[*vsi2]=_TMapEntry<VT,ET,Cmp>(newVPtr1=new Vertex((*vsi2)->data),true);
      else {
	newVPtr1=tmi->second.newVPtr;
	tmi->second.init=true;
      }
#ifdef GRAPH_FAST_REMOVE
      sResult=vmi1->second.insert(newVPtr1);
      newVPtr1->vmi=vmi1;
      newVPtr1->vsi=sResult.first;
#else
      vmi1->second.insert(newVPtr1);
#endif
      
      // for all inward edges
      for (vvmi2=(*vsi2)->inMap.begin();vvmi2!=(*vsi2)->inMap.end();++vvmi2) {
	eResult=newVPtr1->inMap.insert(make_pair(vvmi2->first,emptyEMap));
	typename Vertex::_VMap::iterator &vvmi1=eResult.first;
	
	// for all edges connecting from the vertices with the same VT
	for (emi2=vvmi2->second.begin();emi2!=vvmi2->second.end();++emi2)
	  if ((tmi=tMap.find(emi2->first))==tMap.end()) {
	    // a new src vertex and edge needs to be created
	    tMap[emi2->first]=_TMapEntry<VT,ET,Cmp>(newVPtr2=new Vertex(emi2->first->data),false);
	    vvmi1->second.insert(make_pair(newVPtr2,new ET(*(emi2->second))));
	  }
	  else if (tmi->second.init && tmi->second.newVPtr!=newVPtr1) { // no self-looping edges!
	    fResult=tmi->second.newVPtr->outMap.find(newVPtr1,emi1);
	    assert(fResult);
	    vvmi1->second.insert(make_pair(tmi->second.newVPtr,emi1->second));
	  }
	  else
	    vvmi1->second.insert(make_pair(tmi->second.newVPtr,new ET(*(emi2->second))));
	
	// the keys of Vertex::_VMap are VT* - need to update them
	*(const_cast<VT**>(&(vvmi1->first)))=&(vvmi1->second.begin()->first->data);
      }
      
      // for all outward edges
      for (vvmi2=(*vsi2)->outMap.begin();vvmi2!=(*vsi2)->outMap.end();++vvmi2) {
	emi2=vvmi2->second.begin();
	eResult=newVPtr1->outMap.insert(make_pair(&(emi2->first->data),emptyEMap));
	typename Vertex::_VMap::iterator &vvmi1=eResult.first;
	
	// for all edges connecting to the vertices with the same VT
	for (;emi2!=vvmi2->second.end();++emi2)
	  if ((tmi=tMap.find(emi2->first))==tMap.end()) {
	    // a new dest vertex and edge needs to be created
	    tMap[emi2->first]=_TMapEntry<VT,ET,Cmp>(newVPtr2=new Vertex(emi2->first->data),false);
	    vvmi1->second.insert(make_pair(newVPtr2,new ET(*(emi2->second))));
	  }
	  else if (tmi->second.init) {
	    fResult=tmi->second.newVPtr->inMap.find(newVPtr1,emi1);
	    assert(fResult);
	    vvmi1->second.insert(make_pair(tmi->second.newVPtr,emi1->second));
	  }
	  else
	    vvmi1->second.insert(make_pair(tmi->second.newVPtr,new ET(*(emi2->second))));
	
	// the keys of Vertex::_VMap are VT* - need to update them
	*(const_cast<VT**>(&(vvmi1->first)))=&(vvmi1->second.begin()->first->data);
      }
    }
    
    // the keys of _GVMap<> are VT* - need to update them
    *(const_cast<VT**>(&(vmi1->first)))=&((*(vmi1->second.begin()))->data);
  }
}

};

#endif
