// ======================================================================
// graphAlgorithm.hpp - A bunch of algorithm implementations for a graph.
// 
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Fixed the self-looping edges
//         bug in GIsomorphism<>: now the unary constraints get checked in 
//         the beginning; added SubGIsomorphism<> and restructured the code
//         so both GIsomorphism<> and SubGIsomorphism<> share a good deal of
//         code; minor revisions to supress warnings under -pedantic -Wall
//         (mostly adding 'typename').
// 082901: Benjamin Han <benhdj@cs.cmu.edu> Finished GIsomorphism<>; changed
//         the way a graph is printed in operator<<().
// 081001: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef GRAPHALGORITHM_HPP
#define GRAPHALGORITHM_HPP

#include "graph.hpp"

#include <list>
#include <set>
#include <map>

#include <limits.h>

namespace Toolbox {

// ===========================================================================
// struct _DefaultIsomorphicEOp (internal use ONLY): default edge equivalence
//   operator for the (subgraph) isomorphism test.
// ASSUMPTION: operator==() is defined for ET
// ===========================================================================

template <class ET>
struct _DefaultIsomorphicEOp {
  bool operator () (const ET &e1, const ET &e2) const {
    return (e1==e2);
  }
};

// ===========================================================================
// struct GIsoMap: Graph/subgraph isomorphism mappings
// ===========================================================================

template <class T>
struct GIsoMap: public std::map<typename T::Iterator,typename T::Iterator> {};

// ===========================================================================
// struct _CMapEntry (internal use): A constraint map entry for recording both
//   local and global constraints; used in _GIsoBase<>.
// ===========================================================================

template <class VT, class ET, class Cmp>
struct _CMapEntry {
  typedef _Vertex<VT,ET,Cmp>* VPtr;
  typedef std::set<VPtr> CSet;

  struct LCLEntry {
    bool notAssigned;
    VPtr vPtr;
    CSet lc;                  // local constraints
                              // invariant: lc and gc do not intersect
  };

  typedef std::list<LCLEntry> LCList;

  CSet gc;                    // global constraints
  typename _GVMap<VT,ET,Cmp>::const_iterator vmi;  // points into the input graph

  LCList lcl;                 // local constraints list
};

// ===========================================================================
// struct _GIStackEntry (internal use): A stack entry for backtracking; used in
//   _GIsoBase<>.
// ===========================================================================

template <class VT, class ET, class Cmp>
struct _GIStackEntry {
  typedef typename std::map<VT*,_CMapEntry<VT,ET,Cmp>,_PtrCmp<VT,Cmp> >::iterator CMIterator;
  typedef typename _CMapEntry<VT,ET,Cmp>::LCList::iterator LCLIterator;
  typedef _Vertex<VT,ET,Cmp>* VPtr;
  typedef std::list<VPtr> VList;

  // used to remove the elements in vList from *lcPtr (local constraints)
  struct ULEntry {
    bool remove;   // true iff the elemenets in vList need to be *removed* from *lcPtr
    typename _CMapEntry<VT,ET,Cmp>::CSet *lcPtr;
    VList vList;
  };

  typedef std::list<ULEntry> UndoList;

  CMIterator i1;
  LCLIterator i2;

  // vsi points to the matched vertex in the input graph
  typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsi;

  // for recording the operations on the other local constraints
  UndoList uList;
};


// ===========================================================================
// struct _GIsoBase (internal use): Base class for both GIsomorphism and
//   SubGIsomorphism (graph and subgraph isomorphism tests).
// ===========================================================================

template <class VT, class ET, class Cmp=std::less<VT>,
	  class EOp=_DefaultIsomorphicEOp<ET> >
class _GIsoBase {
protected:

  typedef std::map<VT*,_CMapEntry<VT,ET,Cmp>,_PtrCmp<VT,Cmp> > CMap;
  typedef std::list<_GIStackEntry<VT,ET,Cmp> > GIStack;

  const Graph<VT,ET,Cmp> *gPtr1,*gPtr2;
  EOp eOp;

  typename _CMapEntry<VT,ET,Cmp>::LCLEntry emptyLE;
  typename _GIStackEntry<VT,ET,Cmp>::ULEntry emptyUE;
  CMap cMap;
  typename Graph<VT,ET,Cmp>::size_type numV;
  GIStack giStack;

#ifndef NDEBUG
  bool initFlag;
#endif

  // return true iff there exists a valid next candidate
  bool nextCandidate (typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator &vsi,
		      typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsiEnd,
		      const typename _CMapEntry<VT,ET,Cmp>::CSet &gc,
		      const typename _CMapEntry<VT,ET,Cmp>::CSet &lc) const {
    while (vsi!=vsiEnd &&
	   (gc.find(*vsi)!=gc.end() || lc.find(*vsi)!=lc.end())) ++vsi;

    return (vsi!=vsiEnd);
  }

  const _GVMap<VT,ET,Cmp> &readGVMap () const {
    assert(gPtr2);
    return gPtr2->vMap;
  }

  // the actual graph matching code
  // ASSUMPTIONS:
  // 1. cMap is properly initialized w.r.t the input graph
  // 2. giStack is properly initialized
  bool match ();

public:

  _GIsoBase ():gPtr1(NULL),gPtr2(NULL) {
#ifndef NDEBUG
    initFlag=false;
#endif

    emptyLE.notAssigned=emptyUE.remove=true;
    emptyUE.lcPtr=NULL;
  }

  // must call init() to set up the model graph before calling operator()!
  void init (const Graph<VT,ET,Cmp> &modelG);

  void readMapping (GIsoMap<Graph<VT,ET,Cmp> > &giMap) const;
};

template <class VT, class ET, class Cmp, class EOp> 
void _GIsoBase<VT,ET,Cmp,EOp>::init (const Graph<VT,ET,Cmp> &modelG) {
  typename _GVMap<VT,ET,Cmp>::const_iterator vmi;
  typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsi;
  typename Graph<VT,ET,Cmp>::size_type i,j;

#ifndef NDEBUG
  initFlag=true;
#endif

  // initialize cMap
  cMap.clear();
  for (vmi=modelG.vMap.begin(),numV=0,i=modelG.dSize();i;++vmi,--i) {
    _CMapEntry<VT,ET,Cmp> &cme=cMap[vmi->first];

    numV+=(j=vmi->second.size());
    for (vsi=vmi->second.begin();j;++vsi,--j) {
      cme.lcl.push_back(emptyLE);
      cme.lcl.back().vPtr=*vsi;
#ifdef GA_DEBUG
      cout<<"push "<<(*vsi)->data<<"; "<<*vsi<<endl;
#endif
    }
  }
  
  gPtr1=&modelG;
}

template <class VT, class ET, class Cmp, class EOp>
void _GIsoBase<VT,ET,Cmp,EOp>::readMapping (GIsoMap<Graph<VT,ET,Cmp> > &giMap) const {
  typename GIStack::const_iterator gi;

  giMap.clear();
  for (gi=giStack.begin();gi!=giStack.end();++gi)
    giMap.insert(make_pair(Graph<VT,ET,Cmp>::Iterator(const_cast<Graph<VT,ET,Cmp>*>(gPtr1),gi->i2->vPtr),
			   Graph<VT,ET,Cmp>::Iterator(const_cast<Graph<VT,ET,Cmp>*>(gPtr2),*(gi->vsi))));
}

template <class VT, class ET, class Cmp, class EOp> 
bool _GIsoBase<VT,ET,Cmp,EOp>::match () {
  typename Graph<VT,ET,Cmp>::size_type i,j,k,min;
  typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsi;
  typename _CMapEntry<VT,ET,Cmp>::VPtr v0Ptr,w0Ptr,vPtr,wPtr;
  _GIStackEntry<VT,ET,Cmp> se,*topSEPtr;
  typename _GIStackEntry<VT,ET,Cmp>::VList::iterator vi;
  typename _GIStackEntry<VT,ET,Cmp>::UndoList::iterator ui;
  typename GIStack::iterator gi;
  typename CMap::iterator cmi;
  typename _CMapEntry<VT,ET,Cmp>::LCList::iterator li;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi1,emi2;
  bool hasEdge1,hasEdge2,addLC;

  // backtracking + forward-checking with the most-constrained-var heuristic
  do {
  stackLoop:
    
    _GIStackEntry<VT,ET,Cmp> &topSE=giStack.front();
    _CMapEntry<VT,ET,Cmp> &cme=topSE.i1->second;
    
    // update the global constraints, and make sure gc and all lc's do
    // not intersect
    cme.gc.insert(vPtr=*topSE.vsi);
    for (li=cme.lcl.begin();li!=cme.lcl.end();++li)
      if (li->lc.erase(vPtr)) {
	topSE.uList.push_back(emptyUE);
	typename _GIStackEntry<VT,ET,Cmp>::ULEntry &ue=topSE.uList.back();
	
	ue.remove=false;
	ue.lcPtr=&(li->lc);
	ue.vList.push_back(vPtr);	    
      }
    
#ifdef GA_DEBUG
      cout<<"*** entering stackLoop ***"<<endl;
      cout<<"vertex = "<<*topSE.i1->first<<"; "<<topSE.i2->vPtr<<endl;
      cout<<"candidate = "<<vPtr->data<<"; "<<vPtr<<endl;
#endif

      // forward-checking
      // for every vertex label
#ifdef GA_DEBUG
      cout<<"*** entering forward-checking ***"<<endl;
#endif
      for (cmi=cMap.begin(),min=INT_MAX;cmi!=cMap.end();++cmi) {
	_CMapEntry<VT,ET,Cmp> &cme=cmi->second;

#ifdef GA_DEBUG
	cout<<"vertex label = "<<*cmi->first<<endl;
#endif

	// for every non-assigned vertex of the same label
	for (li=cme.lcl.begin();li!=cme.lcl.end();++li)
	  if (li->notAssigned) {
	    v0Ptr=li->vPtr;
	    topSE.uList.push_back(emptyUE);
	    typename _GIStackEntry<VT,ET,Cmp>::ULEntry &ue=topSE.uList.back();
	    
#ifdef GA_DEBUG
	    cout<<"vertex = "<<v0Ptr->data<<"; "<<v0Ptr<<endl;
#endif

	    // for every non-conflicting candidate vertex
	    for (vsi=cme.vmi->second.begin();vsi!=cme.vmi->second.end();++vsi)
	      if (cme.gc.find(*vsi)==cme.gc.end() && 
		  li->lc.find(*vsi)==li->lc.end()) {
		w0Ptr=*vsi;
		
		// for every assignment check if v0 and w0 are compatible
		for (gi=giStack.begin(),addLC=false;gi!=giStack.end();++gi) {
		  vPtr=gi->i2->vPtr;
		  wPtr=*(gi->vsi);
		  
		  // check 1: (v0,v) <=> (w0,w)
		  hasEdge1=v0Ptr->outMap.find(vPtr,emi1);
		  hasEdge2=w0Ptr->outMap.find(wPtr,emi2);
		  if (!((!hasEdge1 && !hasEdge2) ||
			(hasEdge1 && hasEdge2 && 
			 eOp(*emi1->second,*emi2->second)))) {
		    addLC=true;
		    break;
		  }
		  
		  // check 2: (v,v0) <=> (w,w0)
		  hasEdge1=vPtr->outMap.find(v0Ptr,emi1);
		  hasEdge2=wPtr->outMap.find(w0Ptr,emi2);
		  if (!((!hasEdge1 && !hasEdge2) ||
			(hasEdge1 && hasEdge2 && 
			 eOp(*emi1->second,*emi2->second)))) {
		    addLC=true;
		    break;
		  }
		}
		
		// add w0Ptr into lc of v0Ptr, if w0 is not compatible with v0
		if (addLC) {
#ifdef GA_DEBUG
		  cout<<"lc = "<<w0Ptr->data<<"; "<<w0Ptr<<endl;
#endif
		  ue.vList.push_back(w0Ptr);  // for the undo operations
		}
	      }
	    
	    i=cme.vmi->second.size();          // total candidate set size
	    j=ue.vList.size();                 // new constraint size
	    k=j+li->lc.size()+cme.gc.size();   // total constraint size
	    
	    if (i==k) {
	      // no candidate is left: roll back and backtrack
#ifdef GA_DEBUG
	      cout<<"rollback!"<<endl;
#endif
	      
	      do {
		topSEPtr=&giStack.front();
		_CMapEntry<VT,ET,Cmp> &cme=topSEPtr->i1->second;
		
		// restore the global constraint
		cme.gc.erase(*(topSEPtr->vsi));
		
		// restore the local constraints
		// NOTE since the local constraints of the 'current'
		// vertex (recorded in the top stack element when first
		// entering this loop) haven't been updated, they
		// don't need to be restored (in this case ui->lcPtr=NULL)
		for (ui=topSEPtr->uList.begin();ui!=topSEPtr->uList.end();++ui)
		  if (ui->lcPtr)
		    if (ui->remove)
		      for (vi=ui->vList.begin();vi!=ui->vList.end();++vi)
			ui->lcPtr->erase(*vi);
		    else for (vi=ui->vList.begin();vi!=ui->vList.end();++vi)
		      ui->lcPtr->insert(*vi);
		
		// find the next non-conflicting candidate vertex
		if (nextCandidate(++(topSEPtr->vsi),cme.vmi->second.end(),
				  cme.gc,topSEPtr->i2->lc))
		  // we found a next candidate
		  // necessary evil? otherwise I need to break 3 loops
		  goto stackLoop;
		
		// retract the assignment
		topSEPtr->i2->notAssigned=true;
		giStack.pop_front();
		
	      } while (giStack.size());
	      
	      return false;  // exhausting all possibilities
	    }
	    
	    // find the vertex with min candidate set size
	    assert(i>k);
	    i-=k;
	    if (i<min) {
	      min=i;
	      se.i1=cmi;
	      se.i2=li;
	    }
	    
	    // update the local constraints if necessary
	    if (j) {
	      ue.lcPtr=&(li->lc);   // for the undo operations
	      
	      // now we really add the local constraints
	      for (vi=ue.vList.begin();vi!=ue.vList.end();++vi)
		li->lc.insert(*vi);
	    }
	    else topSE.uList.pop_back();
	    
	  }  // for every non-assigned vertex of the same label
      } // forward-checking loop
      
#ifdef GA_DEBUG
      cout<<"*** exiting forward-checking ***"<<endl;
#endif

      // prepare the next vertex
      _CMapEntry<VT,ET,Cmp> &cmeSE=se.i1->second;
#ifndef NDEBUG
      bool foundNext=
#endif
	nextCandidate(se.vsi=cmeSE.vmi->second.begin(),cmeSE.vmi->second.end(),
		      cmeSE.gc,se.i2->lc);
      assert(foundNext);

      se.i2->notAssigned=false;
      giStack.push_front(se);

    } while (giStack.size()!=numV);

  return true;
}

// ===========================================================================
// class GIsomorphism: Function object for graph isomorphism tests
//       
// NOTES: 1. EOp is a binary predicate testing if two edges are equivalent.
//        2. It's basically a constraint satisfaction problem solved by
//           backtracking + forward-checking, with the most-constrained-variable
//           heuristic. It's similar to J. R. Ullmann, An Algorithm for Subgraph 
//           Isomorphism, ACM, vol. 23, no. 1, 1976, pp 31-42, but it exploits
//           the heuristic at no cost of time complexity, and uses a map instead
//           of a matrix to record the associations of vertices (so less memory
//           is used at a minor expense of time).
// ===========================================================================

template <class VT, class ET, class Cmp=std::less<VT>,
	  class EOp=_DefaultIsomorphicEOp<ET> >
class GIsomorphism: public _GIsoBase<VT,ET,Cmp,EOp> {
  typedef typename _GIsoBase<VT,ET,Cmp,EOp>::CMap CMap;

public:

  bool operator () (const Graph<VT,ET,Cmp> &g);
};

template <class VT, class ET, class Cmp, class EOp> 
bool GIsomorphism<VT,ET,Cmp,EOp>::operator () (const Graph<VT,ET,Cmp> &g) {
  assert(initFlag);

  typename Graph<VT,ET,Cmp>::size_type i,j,k,min;

  if ((i=cMap.size())!=g.dSize()) return false;

  typename _GVMap<VT,ET,Cmp>::const_iterator vmi;
  typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsi;
  _GIStackEntry<VT,ET,Cmp> se;
  typename CMap::iterator cmi;
  typename _CMapEntry<VT,ET,Cmp>::LCList::iterator li;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi1,emi2;
  bool hasEdge1,hasEdge2;

  gPtr2=&g;

  // set up cMap based on the input graph g, and find the first vertex to assign
  for (cmi=cMap.begin(),vmi=readGVMap().begin(),min=INT_MAX;
       cmi!=cMap.end();
       ++cmi,++vmi) {
    _CMapEntry<VT,ET,Cmp> &cme=cmi->second;
    
    // # of vertex labels or vertex label doesn't match
    if (cme.lcl.size()!=(j=vmi->second.size()) || 
	*cmi->first!=*vmi->first) return false;
    
    // reset cMap and set g-related info
    // also initialize local constraints to reflect to the unary 
    // constraints (self-looping edges)
    cme.gc.clear();
    cme.vmi=vmi;
    for (li=cme.lcl.begin();li!=cme.lcl.end();++li) {
      li->notAssigned=true;
      li->lc.clear();
      
      hasEdge1=li->vPtr->outMap.find(li->vPtr,emi1);
      for (vsi=vmi->second.begin();vsi!=vmi->second.end();++vsi) {
	hasEdge2=(*vsi)->outMap.find(*vsi,emi2);
	if (!((hasEdge1 && hasEdge2 && eOp(*emi1->second,*emi2->second)) ||
	      (!hasEdge1 && !hasEdge2)))
	  li->lc.insert(*vsi);
      }
      
      // find the vertex with a minimal # of valid candidates
      // (most-constrained variable heuristic)
      if ((k=j-li->lc.size())<min)
	if (k) {
	  min=k;
	  se.i1=cmi;
	  se.i2=li;
	}
	else return false;  // empty candidate set!
    }
  }
  
  // initialize the stack with the 1st assignment
  _CMapEntry<VT,ET,Cmp> &cme=se.i1->second;
  se.i2->notAssigned=false;
#ifndef NDEBUG
  bool foundNext=
#endif
    nextCandidate(se.vsi=cme.vmi->second.begin(),cme.vmi->second.end(),
		  cme.gc,se.i2->lc);
  assert(foundNext);
  giStack.clear();
  giStack.push_front(se);

  return match();
}

// ===========================================================================
// class SubGIsomorphism: Function object for subgraph isomorphism tests
//       
// NOTES: 1. EOp is a binary predicate testing if two edges are equivalent.
//        2. It's basically a constraint satisfaction problem solved by
//           backtracking + forward-checking, with the most-constrained-variable
//           heuristic. It's similar to J. R. Ullmann, An Algorithm for Subgraph 
//           Isomorphism, ACM, vol. 23, no. 1, 1976, pp 31-42, but it exploits
//           the heuristic at no cost of time complexity, and uses a map instead
//           of a matrix to record the associations of vertices (so less memory
//           is used at a minor expense of time).
// ===========================================================================

template <class VT, class ET, class Cmp=std::less<VT>,
	  class EOp=_DefaultIsomorphicEOp<ET> >
class SubGIsomorphism: public _GIsoBase<VT,ET,Cmp,EOp> {
  typedef typename _GIsoBase<VT,ET,Cmp,EOp>::CMap CMap;

public:

  bool operator () (const Graph<VT,ET,Cmp> &g);
  bool operator () ();  // next match: must call the 1st form before calling this!
};

template <class VT, class ET, class Cmp, class EOp> 
bool SubGIsomorphism<VT,ET,Cmp,EOp>::operator () (const Graph<VT,ET,Cmp> &g) {
  assert(initFlag);

  typename Graph<VT,ET,Cmp>::size_type k,min;
  typename _GVMap<VT,ET,Cmp>::const_iterator vmi;
  typename _CMapEntry<VT,ET,Cmp>::CSet::const_iterator vsi;
  _GIStackEntry<VT,ET,Cmp> se;
  typename CMap::iterator cmi;
  typename _CMapEntry<VT,ET,Cmp>::LCList::iterator li;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi1,emi2;
  bool hasEdge1,hasEdge2;

  gPtr2=&g;

#ifdef GA_DEBUG
  cout<<"*** entering init section ***"<<endl;
#endif

  // set up cMap based on the input graph g, and find the first vertex to assign
  const _GVMap<VT,ET,Cmp> &gvMap=readGVMap();
  for (cmi=cMap.begin(),min=INT_MAX;
       cmi!=cMap.end() && (vmi=gvMap.find(cmi->first))!=gvMap.end();
       ++cmi) {
    _CMapEntry<VT,ET,Cmp> &cme=cmi->second;

    // reset cMap and set g-related info
    // also initialize local constraints to reflect to the unary 
    // constraints (self-looping edges)
    cme.gc.clear();
    cme.vmi=vmi;
    for (li=cme.lcl.begin();li!=cme.lcl.end();++li) {
      li->notAssigned=true;
      li->lc.clear();
      
#ifdef GA_DEBUG
      cout<<"model graph: "<<li->vPtr->data<<'@'<<li->vPtr<<endl;
#endif
      
      hasEdge1=li->vPtr->outMap.find(li->vPtr,emi1);
      for (vsi=vmi->second.begin();vsi!=vmi->second.end();++vsi) {
	hasEdge2=(*vsi)->outMap.find(*vsi,emi2);
	if (!((hasEdge1 && hasEdge2 && eOp(*emi1->second,*emi2->second)) ||
	      (!hasEdge1 && !hasEdge2))) {
#ifdef GA_DEBUG
	  cout<<"remove input graph: "<<(*vsi)->data<<'@'<<*vsi<<endl;
#endif
	  li->lc.insert(*vsi);
	}
      }
      
      // find the vertex with a minimal # of valid candidates
      // (most-constrained variable heuristic)
      if ((k=vmi->second.size()-li->lc.size())<min)
	if (k) {
	  min=k;
	  se.i1=cmi;
	  se.i2=li;
	}
	else return false;  // empty candidate set!
    }
  }
  
#ifdef GA_DEBUG
  cout<<"*** exiting init section ***"<<endl;
#endif

  if (cmi!=cMap.end()) return false;  // some vertex label is missing

  // initialize the stack with the 1st assignment
  _CMapEntry<VT,ET,Cmp> &cme=se.i1->second;
  se.i2->notAssigned=false;
#ifndef NDEBUG
  bool foundNext=
#endif
    nextCandidate(se.vsi=cme.vmi->second.begin(),cme.vmi->second.end(),
		  cme.gc,se.i2->lc);
  assert(foundNext);
  giStack.clear();
  giStack.push_front(se);
  return match();
}

template <class VT, class ET, class Cmp, class EOp> 
bool SubGIsomorphism<VT,ET,Cmp,EOp>::operator () () {
  assert(giStack.size());

  _GIStackEntry<VT,ET,Cmp> *topSEPtr=&giStack.front();
  _CMapEntry<VT,ET,Cmp> *cmePtr=&(topSEPtr->i1->second);
  typename _GIStackEntry<VT,ET,Cmp>::VList::iterator vi;
  typename _GIStackEntry<VT,ET,Cmp>::UndoList::iterator ui;

  do {
    // find the next non-conflicting candidate vertex...
    if (nextCandidate(++(topSEPtr->vsi),cmePtr->vmi->second.end(),
		      cmePtr->gc,topSEPtr->i2->lc))
      // ... then match
      return (giStack.size()==numV?true:match());
    
    // retract the assignment
    topSEPtr->i2->notAssigned=true;
    giStack.pop_front();
    
    if (giStack.size()) {
      topSEPtr=&giStack.front();
      cmePtr=&(topSEPtr->i1->second);
      
      // restore the global constraint
      cmePtr->gc.erase(*(topSEPtr->vsi));
      
      // restore the local constraints
      for (ui=topSEPtr->uList.begin();ui!=topSEPtr->uList.end();++ui) {
	assert(ui->lcPtr);
	if (ui->remove)
	  for (vi=ui->vList.begin();vi!=ui->vList.end();++vi)
	    ui->lcPtr->erase(*vi);
	else for (vi=ui->vList.begin();vi!=ui->vList.end();++vi)
	  ui->lcPtr->insert(*vi);
      }
    }
    else return false;      
    
  } while (true);
}

// ===========================================================================
// operator <<: Print out a graph vertex by vertex
// ===========================================================================

template <class VT, class ET, class Cmp>
std::ostream &operator << (std::ostream &os, const Graph<VT,ET,Cmp> &g) {
  typename Graph<VT,ET,Cmp>::ConstDIterator gdi;
  typename Graph<VT,ET,Cmp>::ConstVIterator gvi;
  typename Graph<VT,ET,Cmp>::ConstNIterator gni;

  for (gdi=g.dBegin();gdi!=g.dEnd();++gdi)
    for (gvi=gdi.begin();gvi!=gdi.end();++gvi) {
      os<<"+ Vertex "<<*gdi<<'@'<<gvi.vPtr<<" ["<<gvi.inDegree()<<','
	<<gvi.outDegree()<<"]:"<<endl;
      
      if (gvi.inDegree()) {
	os<<"  inward edges:";
	for (gni=gvi.inBegin();gni!=gvi.inEnd();++gni)
	  cout<<" ["<<*gni<<'|'<<gni.vPtr<<','<<gni.readEdge(gvi)<<']';
	os<<endl;
      }
      if (gvi.outDegree()) {
	os<<"  outward edges:";
	for (gni=gvi.outBegin();gni!=gvi.outEnd();++gni)
	  cout<<" ["<<*gni<<'|'<<gni.vPtr<<','<<gvi.readEdge(gni)<<']';
	os<<endl;
      }
    }
  
  return os;
}

};

#endif
