// ======================================================================
// graphIter.hpp - Iterators of a generic vertex-indexed, directed and 
//                 single-edge graph  - you can extend _GIterator<> by
//                 implementing your own Agenda.
//                 WARNING: this file is not meant to be used alone -
//                 include graph.hpp to use any iterator defined here!
// 
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Inserted an extra space before
//         '>' in all macros.
// 112002: Benjamin Han <benhdj@cs.cmu.edu> Make template friends go away
//         and use MS versions of iteration implementation if we're using
//         MS VC6.
// 103002: Benjamin Han <benhdj@cs.cmu.edu> Be explicit in publicly
//         subclassing various STL iterators.
// 102402: Benjamin Han <benhdj@cs.cmu.edu> Added the missing #include
//         <assert.h>.
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 090901: Benjamin Han <benhdj@cs.cmu.edu> Fixed bug in
//         _GPathIterator<>::operator--(): now pi is updated correctly;
//         added moveTo(), readPI() and pathDist() for _GPathIterator<>.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Made GIsomorphism<> in
//         graphAlgorithm a friend of _GIteratorBase<>; added a const version of
//         find() in _GVMap<>; added _GIteratorBase<>::printVPtr(); minor 
//         revisions to supress warnings under -pedantic -Wall (mostly adding
//         'typename').
// 082901: Benjamin Han <benhdj@cs.cmu.edu> Renamed disconnect() to 
//         removeEdgeTo() in _GIteratorBase<>; revised both addEdgeTo() in
//         _GIteratorBase<> so that if the edge exists it's updated and no
//         new edge is created; fixed the problem that _GDIterator<>::end() can't
//         be distinguished.
// 082501: Benjamin Han <benhdj@cs.cmu.edu> Revised PreorderAgenda and 
//         PostorderAgenda: now they keep only the vertices of the path from
//         the starting vertex to the current vertex, thus running faster and 
//         using less memory - note the behavior is also changed a little but is
//         now consistent with Cormen's version.
// 082101: Benjamin Han <benhdj@cs.cmu.edu> Added new iterators - data iterators
//         _GDIterator<> and vertex iterators _GVIterator<>.
// 081301: Benjamin Han <benhdj@cs.cmu.edu> Abandoned the idea of using stamp
//         for traversal (keeping stamps in each vertex makes multiple
//         traversals at the same time infeasible) - new _GIterator<> uses
//         History to keep track of the vertices discovered; changed conenctTo()
//         to addEdgeTo(), and isConnectedTo() to hasEdgeTo(); added
//         _GIteratorBase<>::isConnectedTo() - test if two vertices are connected
//         using a preorder traversal.
// 081101: Benjamin Han <benhdj@cs.cmu.edu> Most methods of Agenda<> now
//         have a const iterator argument; removed remove() from all agendas
//         - not necessary since we don't have something like Tree<>::erase();
//         revised some comments.
// 080901: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef GRAPHITER_HPP
#define GRAPHITER_HPP

#include "tbconsts.hpp"

#include <iostream>
#include <iterator>
#include <list>
#include <map>
#include <set>

#include <assert.h>

// WARNING: graphIter.hpp is NOT meant to be used alone - include graph.hpp
//          if you want to use any iterators defined here!

// default is enabling fast removal (consuming more memory)
#define GRAPH_FAST_REMOVE

namespace Toolbox {

// ===========================================================================
// struct _PtrCmp (internal use ONLY): this is to avoid using VT directly as 
//                                     the keys of a map.
// 'VT' is the type of the vertex payload and 'Cmp' is a binary predicate 
// "less-than" to compare two vertices.
// ===========================================================================

template <class VT, class Cmp>
struct _PtrCmp {
  Cmp cmp;
  
  bool operator () (VT *ptr1, VT *ptr2) const {
    return cmp(*ptr1,*ptr2);
  }
};

// ===========================================================================
// struct _GVMap (internal use ONLY): a global map for all vertices in a graph
// ===========================================================================

template <class VT, class ET, class Cmp> class _Vertex;

template <class VT, class ET, class Cmp>
struct _GVMap: public std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> > { // indexed by data
  typedef typename std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::iterator iterator;
  typedef typename std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::const_iterator const_iterator;
  typedef size_t size_type;
  
  iterator find (const VT *dataPtr) {
    return std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::find(const_cast<VT*>(dataPtr));
  }

  const_iterator find (const VT *dataPtr) const {
    return std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::find(const_cast<VT*>(dataPtr));
  }
  
  void erase (_Vertex<VT,ET,Cmp>* vPtr) {
    assert(vPtr);
    iterator vmi=find(&(vPtr->data));
    assert(vmi!=end());
    if (vmi->second.size()==1) std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::erase(vmi);
    else {
      typename std::set<_Vertex<VT,ET,Cmp>*>::iterator vsi;
      
      vsi=vmi->second.find(vPtr);
      assert(vsi!=vmi->second.end());
      
      vmi->second.erase(vsi);
      if (vmi->first==&(vPtr->data))
	*(const_cast<VT**>(&(vmi->first)))=&((*(vmi->second.begin()))->data);
    }
  }

#ifdef GRAPH_FAST_REMOVE
  void erase (iterator &vmi, typename std::set<_Vertex<VT,ET,Cmp>*>::iterator &vsi) {
    std::set<_Vertex<VT,ET,Cmp>*> &vSet=vmi->second;

    if (vSet.size()==1)
      std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::erase(vmi);
    else
      if (vmi->first==&((*vsi)->data)) {
	vSet.erase(vsi);
	
	// the keys of _GVMap<> are VT* - need to update them
	*(const_cast<VT**>(&(vmi->first)))=&((*(vmi->second.begin()))->data);
      }
      else vSet.erase(vsi);
  }
#endif
  
  size_type size () const {
    const_iterator vmi;
    size_type s;
    
    for (vmi=this->begin(),s=0;vmi!=this->end();++vmi) s+=vmi->second.size();
    return s;
  }

  size_type dSize () const {
    return std::map<VT*,std::set<_Vertex<VT,ET,Cmp>*>,_PtrCmp<VT,Cmp> >::size();
  }
};

// ===========================================================================
// struct _Vertex (internal use ONLY): a vertex
// 'VT' is the type of the vertex payload, 'ET' is the type of the edge payload,
// and 'Cmp' is a binary predicate "less-than" to compare two vertices.
// ===========================================================================

template <class VT, class ET, class Cmp>
struct _Vertex {
  typedef typename _GVMap<VT,ET,Cmp>::size_type size_type;

  // _EMap: for inward map _Vertex* points to the src vertex
  //        for outward map _Vertex* points to the dest vertex
  typedef std::map<_Vertex*,ET*> _EMap;        // indexed by ptrs to the vertices
  struct _VMap: public std::map<VT*,_EMap,_PtrCmp<VT,Cmp> > { // indexed by data
    typedef typename std::map<VT*,_EMap,_PtrCmp<VT,Cmp> >::iterator iterator;
    typedef typename std::map<VT*,_EMap,_PtrCmp<VT,Cmp> >::const_iterator const_iterator;

    iterator find (const VT *dataPtr) {
      return std::map<VT*,_EMap,_PtrCmp<VT,Cmp> >::find(const_cast<VT*>(dataPtr));
    }

    bool find (const _Vertex *vPtr, typename _EMap::iterator &iter) {
      assert(vPtr);
      iterator vmi=find(&(vPtr->data));
      if (vmi!=end()) {
	iter=vmi->second.find(const_cast<_Vertex*>(vPtr));
	return (iter!=vmi->second.end());
      }
      else return false;
    }

    // erase() also frees the memory of the edge data iff eraseEdgeData==true
    void erase (_Vertex *vPtr, bool eraseEdgeData=true) {
      assert(vPtr);
      iterator vmi=find(&(vPtr->data));
      assert(vmi!=end());
      if (vmi->second.size()==1) {
	// free the edge data
	if (eraseEdgeData) delete vmi->second.begin()->second;
	std::map<VT*,_EMap,_PtrCmp<VT,Cmp> >::erase(vmi);
      }
      else {
	typename _EMap::iterator emi;
	
	emi=vmi->second.find(vPtr);
	assert(emi!=vmi->second.end());
	
	// free the edge data
	if (eraseEdgeData) delete emi->second;
	vmi->second.erase(emi);
	if (vmi->first==&(vPtr->data))
	  *(const_cast<VT**>(&(vmi->first)))=&(vmi->second.begin()->first->data);
      }
    }
    
    _Vertex::size_type degree () const {
      const_iterator vmi;
      _Vertex::size_type d;
      
      for (vmi=begin(),d=0;vmi!=end();++vmi) d+=vmi->second.size();
      return d;
    }
  };

#ifdef GRAPH_FAST_REMOVE
  // vmi and vsi point back to the global map in Graph so that we can remove
  // a vertex real fast (see Graph::pruneOne(Iterator) in graph.hpp)
  typename _GVMap<VT,ET,Cmp>::iterator vmi;
  typename std::set<_Vertex<VT,ET,Cmp>*>::iterator vsi;
#endif

  _VMap inMap,outMap;
  VT data;

  size_type inDegree () const { return inMap.degree(); }
  size_type outDegree () const { return outMap.degree(); }

  _Vertex (const VT &data):data(data) {}
};

// ===========================================================================
// class _GIteratorBase (internal use ONLY): the graph iterator base class
// NOTE: use the macros GIteratorBase(VT,ET,Cmp) and
//       GConstIteratorBase(VT,ET,Cmp)
// ===========================================================================

#define GIteratorBase(VT,ET,Cmp) _GIteratorBase<VT,ET,VT&,VT*,Cmp>
#define GConstIteratorBase(VT,ET,Cmp) _GIteratorBase<VT,ET,const VT&,const VT*,Cmp>

template <class VT, class ET, class Cmp> class Graph;
template <class VT, class ET, class VRef, class VPtr, class Cmp> class _GNIterator;

template <class VT, class ET, class VRef, class VPtr, class Cmp>
class _GIteratorBase {

#if _MSC_VER==1200
public:

#else

  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_> friend class _GIteratorBase;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_> friend class Agenda;
  template <class VT_, class _ET, class Cmp_> friend class Graph;

  template <class VT_, class ET_, class Cmp_>
  friend std::ostream &operator<< (std::ostream &os, const Graph<VT_,ET_,Cmp_> &g);

  // allow access for some algorithm classes for efficiency and debugging
  template <class VT_, class ET_, class Cmp_, class EOp_>
  friend class _GIsoBase;

#endif

  // for remembering for findNext()
  typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;

protected:
  Graph<VT,ET,Cmp> *gPtr;
  _Vertex<VT,ET,Cmp> *vPtr;

  _GIteratorBase (Graph<VT,ET,Cmp> *gPtr, _Vertex<VT,ET,Cmp> *vPtr):
    gPtr(gPtr),vPtr(vPtr) {}
  _GIteratorBase (Graph<VT,ET,Cmp> *gPtr):gPtr(gPtr),vPtr(NULL) {}

public:

  typedef typename _Vertex<VT,ET,Cmp>::size_type size_type;

  static _GIteratorBase nullGI;

  _GIteratorBase ():gPtr(NULL),vPtr(NULL) {}

  // yielding neighbor iterators
  _GNIterator<VT,ET,VRef,VPtr,Cmp> inBegin () const;
  _GNIterator<VT,ET,VRef,VPtr,Cmp> inEnd () const;
  _GNIterator<VT,ET,VRef,VPtr,Cmp> outBegin () const;
  _GNIterator<VT,ET,VRef,VPtr,Cmp> outEnd () const;

  // WARNING: *FindNext() bombs if the corresponding *Find() is not called 
  //          first
  bool inFind (const VT &data, _GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter);
  bool inFindNext (_GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter);
  bool outFind (const VT &data, _GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter);
  bool outFindNext (_GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter);

  void addEdgeTo (_GIteratorBase &iter, const ET &data);
  ET &addEdgeTo (_GIteratorBase &iter);

  void removeEdgeTo (const GIteratorBase(VT,ET,Cmp) &iter) {
    assert(vPtr && iter.vPtr);
    assert(gPtr==iter.gPtr);

    // only free the edge data ONCE!
    iter.vPtr->inMap.erase(vPtr);
    vPtr->outMap.erase(iter.vPtr,false);
  }
  
  // ASSUMPTION: *this is connected to iter
  ET &getEdge (const GIteratorBase(VT,ET,Cmp) &iter) {
    assert(vPtr && iter.vPtr);
    assert(gPtr==iter.gPtr);

    // this shadows _GIteratorBase<>::vmi
    typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi=vPtr->outMap.find(&(*iter));
    assert(vmi!=vPtr->outMap.end());

    // this shadows _GIteratorBase<>::emi
    typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi=
      vmi->second.find(const_cast<_Vertex<VT,ET,Cmp>*>(iter.vPtr));
    assert(emi!=vmi->second.end());
    
    return *(emi->second);
  }
  const ET &readEdge (const GConstIteratorBase(VT,ET,Cmp) &iter) const {
    assert(vPtr && iter.vPtr);
    assert(gPtr==iter.gPtr);
    // this shadows _GIteratorBase<>::vmi
    typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi=vPtr->outMap.find(&(*iter));
    assert(vmi!=vPtr->outMap.end());

    // this shadows _GIteratorBase<>::emi
    typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi=
      vmi->second.find(const_cast<_Vertex<VT,ET,Cmp>*>(iter.vPtr));
    assert(emi!=vmi->second.end());
    
    return *(emi->second);
  }

  size_type inDegree () const { return (vPtr?vPtr->inDegree():0); }
  size_type outDegree () const { return (vPtr?vPtr->outDegree():0); }

  bool isSource () const { return (inDegree()==0); }
  bool isSink () const { return (outDegree()==0); }
  bool isIsolated () const { return (inDegree()==0 && outDegree()==0); }
  bool hasEdgeTo (const GConstIteratorBase(VT,ET,Cmp) &iter) const {
    assert(vPtr && iter.vPtr);
    assert(gPtr==iter.gPtr);

    typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi=vPtr->outMap.find(&(*iter));
    
    return (vmi!=vPtr->outMap.end()?
	    vmi->second.find(const_cast<_Vertex<VT,ET,Cmp>*>(iter.vPtr))!=vmi->second.end():
	    false);
  }
  bool isConnectedTo (const GConstIteratorBase(VT,ET,Cmp) &iter) const;

  // type conversion to a const iterator
  operator GConstIteratorBase(VT,ET,Cmp) () const {
    return GConstIteratorBase(VT,ET,Cmp)(gPtr,vPtr);
  }

  // note gPtr does not play a role here - but vPtr should be unique across
  // different graphs
  bool operator < (const _GIteratorBase &iter) const { return vPtr<iter.vPtr; }
  bool operator == (const _GIteratorBase &iter) const { return vPtr==iter.vPtr; }
  bool operator != (const _GIteratorBase &iter) const { return vPtr!=iter.vPtr; }

  const VT& operator * () const { assert(vPtr); return (*vPtr).data; }
  const VT* operator -> () const { return &(operator*()); }
  void update (const VT &data) const;  // will be defined in graph.hpp

  // WARNING: UNLESS YOU ARE SURE THE DATA MODIFICATION WON'T CHANGE THE DATA ORDERING,
  //          YOU SHOULD NOT USE THIS DANGEROUS METHOD!
  VT& get () const { assert(vPtr); return vPtr->data; }

  _GIteratorBase &operator = (const _GIteratorBase &iter) {
    gPtr=iter.gPtr;
    vPtr=iter.vPtr;
    return *this;
  }

  // print out the address of the embedded vertex pointer
  void printVPtr (std::ostream &os) const { os<<vPtr; }
};

// by default gPtr and vPtr are NULL
template <class VT, class ET, class VRef, class VPtr, class Cmp>
_GIteratorBase<VT,ET,VRef,VPtr,Cmp> _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::nullGI;

template <class VT, class ET, class VRef, class VPtr, class Cmp>
void _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::addEdgeTo (_GIteratorBase &iter, const ET &data) {
  assert(vPtr && iter.vPtr);
  assert(gPtr==iter.gPtr);

  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;
  
  // if the edge exists, update the edge data
  if (iter.vPtr->inMap.find(vPtr,emi))
    *(emi->second)=data;
  else {
    ET *ePtr=new ET(data);
    
    iter.vPtr->inMap[&(vPtr->data)].insert(make_pair(vPtr,ePtr));
    vPtr->outMap[&(iter.vPtr->data)].insert(make_pair(iter.vPtr,ePtr));
  }
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
ET &_GIteratorBase<VT,ET,VRef,VPtr,Cmp>::addEdgeTo (_GIteratorBase &iter) {
  assert(vPtr && iter.vPtr);
  assert(gPtr==iter.gPtr);

  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;

  // if the edge exists, return the edge data
  if (iter.vPtr->inMap.find(vPtr,emi))
    return *(emi->second);

  ET *ePtr=new ET;

  iter.vPtr->inMap[&(vPtr->data)].insert(make_pair(vPtr,ePtr));
  vPtr->outMap[&(iter.vPtr->data)].insert(make_pair(iter.vPtr,ePtr));
  return *ePtr;
}

// ===========================================================================
// class _GVIterator (internal use ONLY): the vertex iterator - iterating
//   through the vertices of the *same* data
// NOTE: Use the macros GVIterator(VT,ET,Cmp) and GConstDIterator(VT,ET,Cmp).
// ===========================================================================

#define GVIterator(VT,ET,Cmp) _GVIterator<VT,ET,VT&,VT*,Cmp>
#define GConstVIterator(VT,ET,Cmp) _GVIterator<VT,ET,const VT&,const VT*,Cmp>

template <class VT, class ET, class VRef, class VPtr, class Cmp>
class _GVIterator: public _GIteratorBase<VT,ET,VRef,VPtr,Cmp>,
#if _MSC_VER==1200
		   public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t> {
#else
		   public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t,VPtr,VRef> {
#endif

#if _MSC_VER==1200
public:

#else

  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_> friend class _GVIterator;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_> friend class _GDIterator;

#endif

  // shadowing vmi and emi in _GIteratorBase<> but that's ok
  typename _GVMap<VT,ET,Cmp>::iterator vmi;
  typename std::set<_Vertex<VT,ET,Cmp>*>::iterator emi;

protected:

  _GVIterator (Graph<VT,ET,Cmp> *gPtr,
	       typename _GVMap<VT,ET,Cmp>::iterator vmi,
	       typename std::set<_Vertex<VT,ET,Cmp>*>::iterator emi):
    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>(gPtr),vmi(vmi),emi(emi) {
    if (emi!=vmi->second.end()) vPtr=*emi;
  }

public:

  _GVIterator () {}

  // type conversion to a const data iterator
  operator GConstVIterator(VT,ET,Cmp) () const {
    return GConstVIterator(VT,ET,Cmp)(gPtr,vmi,emi);
  }

  // preincrement & postincrement
  _GVIterator &operator ++ () {
    if (++emi==vmi->second.end()) vPtr=NULL;
    else vPtr=*emi;
    return *this;
  }
  _GVIterator operator ++ (int) {
    _GVIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  _GVIterator &operator -- () {
    vPtr=*(--emi);
    return *this;
  }
  _GVIterator operator -- (int) {
    _GVIterator tmp=*this;
    --*this;
    return tmp;
  }
};

// ===========================================================================
// class _GDIterator (internal use ONLY): the data iterator - iterating
//   through the different vertex data
// NOTES: 1. It does NOT inherit from _GIteratorBase<>, hence it's NOT pointing
//           to a vertex - you have to use begin() and end() to get a
//           _GIteratorBase<> iterator for that.
//        2. Use the macros GDIterator(VT,ET,Cmp) and GConstDIterator(VT,ET,Cmp).
// ===========================================================================

#define GDIterator(VT,ET,Cmp) _GDIterator<VT,ET,VT&,VT*,Cmp>
#define GConstDIterator(VT,ET,Cmp) _GDIterator<VT,ET,const VT&,const VT*,Cmp>

template <class VT, class ET, class VRef, class VPtr, class Cmp>
class _GDIterator:
#if _MSC_VER==1200
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t> {
#else
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t,VPtr,VRef> {
#endif

#if _MSC_VER==1200
public:

#else

  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_> friend class _GDIterator;
  template <class VT_, class _ET, class Cmp_> friend class Graph;

#endif

  typename _GVMap<VT,ET,Cmp>::iterator vmi;

protected:
  Graph<VT,ET,Cmp> *gPtr;

  _GDIterator (Graph<VT,ET,Cmp> *gPtr):gPtr(gPtr) {}
  _GDIterator (Graph<VT,ET,Cmp> *gPtr, typename _GVMap<VT,ET,Cmp>::iterator vmi):
    vmi(vmi),gPtr(gPtr) {}

public:

  typedef typename _Vertex<VT,ET,Cmp>::size_type size_type;

  _GDIterator ():gPtr(NULL) {}

  GVIterator(VT,ET,Cmp) begin () {
    assert(gPtr);
    return GVIterator(VT,ET,Cmp)(gPtr,vmi,vmi->second.begin());
  }
  GConstVIterator(VT,ET,Cmp) begin () const {
    assert(gPtr);
    return GConstVIterator(VT,ET,Cmp)(gPtr,vmi,vmi->second.begin());
  }
  GVIterator(VT,ET,Cmp) end () {
    assert(gPtr);
    return GVIterator(VT,ET,Cmp)(gPtr,vmi,vmi->second.end());
  }
  GConstVIterator(VT,ET,Cmp) end () const {
    assert(gPtr);
    return GConstVIterator(VT,ET,Cmp)(gPtr,vmi,vmi->second.end());
  }

  size_type size () const {
    assert(gPtr);
    return vmi->second.size();
  }

  // type conversion to a const data iterator
  operator GConstDIterator(VT,ET,Cmp) () const {
    return GConstDIterator(VT,ET,Cmp)(gPtr,vmi);
  }

  // preincrement & postincrement
  _GDIterator &operator ++ () {
    assert(gPtr);
    ++vmi;
    return *this;
  }
  _GDIterator operator ++ (int) {
    _GDIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  _GDIterator &operator -- () {
    assert(gPtr);
    --vmi;
    return *this;
  }
  _GDIterator operator -- (int) {
    _GDIterator tmp=*this;
    --*this;
    return tmp;
  }

  // note gPtr does not play a role here - but vmi should be unique across
  // different graphs
  bool operator < (const _GDIterator &iter) const { return vmi<iter.vmi; }
  bool operator == (const _GDIterator &iter) const { return vmi==iter.vmi; }
  bool operator != (const _GDIterator &iter) const { return vmi!=iter.vmi; }

  const VT& operator * () const { assert(gPtr && vmi!=gPtr->vMap.end()); return *(vmi->first); }
  const VT* operator -> () const { return &(operator*()); }
};

// ===========================================================================
// class _GNIterator: the neighbor traversal iterator
// NOTE: use the macros GNIterator(VT,ET,Cmp) and GConstNIterator(VT,ET,Cmp)
// ===========================================================================

#define GNIterator(VT,ET,Cmp) _GNIterator<VT,ET,VT&,VT*,Cmp>
#define GConstNIterator(VT,ET,Cmp) _GNIterator<VT,ET,const VT&,const VT*,Cmp>

template <class VT, class ET, class VRef, class VPtr, class Cmp>
class _GNIterator: public _GIteratorBase<VT,ET,VRef,VPtr,Cmp>,
#if _MSC_VER==1200
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t> {
#else
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t,VPtr,VRef> {
#endif

#if _MSC_VER==1200
public:

#else

  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_>
  friend class _GIteratorBase;
  template <class VT_, class ET_, class VRef_, class VPtr_, class Cmp_>
  friend class _GNIterator;

#endif

  _Vertex<VT,ET,Cmp> *src;
  bool dirFlag;   // true if it's for an outward neighbor (default)

  // shadowing vmi and emi in _GIteratorBase<>
  // these two point in the _VMap and _EMap of the src vertex
  typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;

  // WARNING: vmi and emi are not set in the first form
  _GNIterator (Graph<VT,ET,Cmp> *gPtr):
    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>(gPtr),src(NULL),dirFlag(true) {}
  _GNIterator (Graph<VT,ET,Cmp> *gPtr, _Vertex<VT,ET,Cmp> *vPtr,
	       _Vertex<VT,ET,Cmp> *src, 
	       typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi,
	       typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi,
	       bool dirFlag):
    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>(gPtr,vPtr),
    src(src),dirFlag(dirFlag),vmi(vmi),emi(emi)    
  {}

public:

  _GNIterator ():src(NULL),dirFlag(true) {}

  // type conversion to a const iterator
  // we can't reuse the conversion in _GIteratorBase<> because I want to block
  // an assignment from any type of iterator except for _GNIterator<>.
  operator GConstNIterator(VT,ET,Cmp) () const {
    return GConstNIterator(VT,ET,Cmp)(gPtr,vPtr,src,vmi,emi,dirFlag);
  }
  
  // preincrement & postincrement
  _GNIterator &operator ++ () {
    assert(src);

    if ((++emi)!=vmi->second.end()) vPtr=emi->first;
    else if (dirFlag)                    // outward neighbor
      if ((++vmi)!=src->outMap.end()) {
	emi=vmi->second.begin();
	vPtr=emi->first;
      }
      else vPtr=NULL;
    else if ((++vmi)!=src->inMap.end()) {  // inward neighbor
      emi=vmi->second.begin();
      vPtr=emi->first;
    }
    else vPtr=NULL;

    return *this;
  }
  _GNIterator operator ++ (int) {
    _GNIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  _GNIterator &operator -- () {
    assert(src);

    if (vPtr) {
      if (emi!=vmi->second.begin()) --emi;
      else {
	--vmi;
	--(emi=vmi->second.end());
      }
      vPtr=emi->first;
    }
    else {
      // *this currently points to NULL
      if (dirFlag)                       // outward neighbor
	--(vmi=src->outMap.end());
      else                               // inward neighbor
	--(vmi=src->inMap.end());
      --(emi=vmi->second.end());
      vPtr=emi->first;
    }
    
    return *this;
  }
  _GNIterator operator -- (int) {
    _GNIterator tmp=*this;
    --*this;
    return tmp;
  }
};

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline _GNIterator<VT,ET,VRef,VPtr,Cmp> _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::inBegin () const {
  _GNIterator<VT,ET,VRef,VPtr,Cmp> tmpNI(gPtr);  // by default tmpNI.vPtr==NULL

  tmpNI.dirFlag=false;
  tmpNI.src=vPtr;
  if (vPtr && vPtr->inMap.size()) {
    tmpNI.vmi=vPtr->inMap.begin();
    tmpNI.emi=tmpNI.vmi->second.begin();
    tmpNI.vPtr=tmpNI.emi->first;
  }
  return tmpNI;
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline _GNIterator<VT,ET,VRef,VPtr,Cmp> _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::inEnd () const {
  _GNIterator<VT,ET,VRef,VPtr,Cmp> tmpNI(gPtr);  // by default tmpNI.vPtr==NULL

  tmpNI.dirFlag=false;
  tmpNI.src=vPtr;
  return tmpNI;
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline _GNIterator<VT,ET,VRef,VPtr,Cmp> _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::outBegin () const {
  _GNIterator<VT,ET,VRef,VPtr,Cmp> tmpNI(gPtr);  // by default tmpNI.vPtr==NULL

  tmpNI.dirFlag=true;
  tmpNI.src=vPtr;
  if (vPtr && vPtr->outMap.size()) {
    tmpNI.vmi=vPtr->outMap.begin();
    tmpNI.emi=tmpNI.vmi->second.begin();
    tmpNI.vPtr=tmpNI.emi->first;
  }
  return tmpNI;
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline _GNIterator<VT,ET,VRef,VPtr,Cmp> _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::outEnd () const {
  _GNIterator<VT,ET,VRef,VPtr,Cmp> tmpNI(gPtr);  // by default tmpNI.vPtr==NULL

  tmpNI.dirFlag=true;
  tmpNI.src=vPtr;
  return tmpNI;
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline bool _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::inFind (const VT &data,
							 _GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter) {
  if (vPtr && (vmi=vPtr->inMap.find(&data))!=vPtr->inMap.end()) {
    emi=vmi->second.begin();
    nIter.dirFlag=false;
    nIter.src=vPtr;
    nIter.gPtr=gPtr;
    nIter.vPtr=emi->first;
    nIter.vmi=vmi;
    nIter.emi=emi;
    return true;
  }
  else return false;
}

// WARNING: bomb if inFind() is not called first
template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline bool _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::inFindNext (_GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter) {
  if ((++emi)!=vmi->second.end()) {
    nIter.dirFlag=false;
    nIter.src=vPtr;
    nIter.gPtr=gPtr;
    nIter.vPtr=emi->first;
    nIter.vmi=vmi;
    nIter.emi=emi;
    return true;
  }
  else return false;
}

template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline bool _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::outFind (const VT &data,
							  _GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter) {
  if (vPtr && (vmi=vPtr->outMap.find(&data))!=vPtr->outMap.end()) {
    emi=vmi->second.begin();
    nIter.dirFlag=true;
    nIter.src=vPtr;
    nIter.gPtr=gPtr;
    nIter.vPtr=emi->first;
    nIter.vmi=vmi;
    nIter.emi=emi;
    return true;
  }
  else return false;
}

// WARNING: bomb if outFind() is not called first
template <class VT, class ET, class VRef, class VPtr, class Cmp>
inline bool _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::outFindNext (_GNIterator<VT,ET,VRef,VPtr,Cmp> &nIter) {
  if ((++emi)!=vmi->second.end()) {
    nIter.dirFlag=true;
    nIter.src=vPtr;
    nIter.gPtr=gPtr;
    nIter.vPtr=emi->first;
    nIter.vmi=vmi;
    nIter.emi=emi;
    return true;
  }
  else return false;
}
  
// ===========================================================================
// class _GIterator: traverse the entire graph
// NOTE: 1. Use the macros GIterator(VT,ET,Agenda,Cmp) and 
//          GConstIterator(VT,ET,Agenda,Cmp).
//       2. Four 'factory-built' agendas are available: PreorderAgenda,
//          PostorderAgenda, BFSAgenda and FrontAgenda (see graph.hpp). You 
//          can write your own agenda by modeling after them: in particular 
//          you must implement init(), nextComponent(), and visitNext()
//          methods.
//       3. For an agenda the assumption is once an _GIteratorBase<>
//          is output via init() or visitNext() it must be removed from the
//          agenda.
// ===========================================================================

#define GIterator(VT,ET,Agenda,Cmp) _GIterator<VT,ET,VT&,VT*,Agenda<VT,ET,VT&,VT*,Cmp>,Cmp>
#define GConstIterator(VT,ET,Agenda,Cmp) _GIterator<VT,ET,const VT&,const VT*,Agenda<VT,ET,const VT&,const VT*,Cmp>,Cmp>

template <class VT, class ET, class VRef, class VPtr, class Agenda, class Cmp>
class _GIterator: public _GIteratorBase<VT,ET,VRef,VPtr,Cmp>, 
#if _MSC_VER==1200
  public std::iterator<std::forward_iterator_tag,VT,std::ptrdiff_t> {
#else
  public std::iterator<std::forward_iterator_tag,VT,std::ptrdiff_t,VPtr,VRef> {
#endif

#if _MSC_VER==1200
public:

#else

  template <class VT_, class ET_, class VRef_, class VPtr_, class Agenda_, class Cmp_> friend class _GIterator;
  template <class VT_, class ET_, class Cmp_> friend class Graph;

#endif

  Agenda agenda;

public:

  // nullGI is a null iterator used for terminating a traversal
  // we need to override _GIteratorBase<>::nullGI because in the STL algorithms
  // we can't have two iterators of different types (e.g. 
  // copy(g.begin(),g.nullGI,...) won't work )
  static _GIterator nullGI;

  _GIterator () {}
  _GIterator (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) { operator=(gi); }

  // conversion between iterators of different agendas
  template <class Agenda_>
  _GIterator (const _GIterator<VT,ET,VRef,VPtr,Agenda_,Cmp> &gi) { operator=(gi); }

  // start a new traversal from this iterator
  void start () { agenda.init(*this); }

  // WARNING: operator=(const _GIteratorBase<>&) or start() *must* be called before
  //          nextComponent()!
  // if a unqueued and unvisited vertex exists gi will be pointed to it
  // and the agenda is initialized, ready for another traversal within a new
  // connected component, in which case the return value is true; otherwise
  // the return value is false and *this is not modified
  bool nextComponent () {
    return agenda.nextComponent(*this);
  }

  // TO-DO: not finished yet!
  bool isQueued () const { return true; }
  bool isVisited () const { return true; }

  _GIterator &operator = (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::operator=(gi);
    agenda.init(*this);
    return *this;
  }

  // assignment between iterators of different agendas
  template <class Agenda_>
  _GIterator &operator = (const _GIterator<VT,ET,VRef,VPtr,Agenda_,Cmp> &gi) {
    return operator=((const _GIteratorBase<VT,ET,VRef,VPtr,Cmp>&)gi);
  }

  // preincrement & postincrement
  _GIterator &operator ++ () {
    agenda.visitNext(*this);
    return *this;
  }
  _GIterator operator ++ (int) {
    _GIterator tmp=*this;
    ++*this;
    return tmp;
  }
};

// by default gPtr and vPtr are NULL
template <class VT, class ET, class VRef, class VPtr, class Agenda, class Cmp>
_GIterator<VT,ET,VRef,VPtr,Agenda,Cmp> _GIterator<VT,ET,VRef,VPtr,Agenda,Cmp>::nullGI;

// ===========================================================================
// struct Agenda: the mother of all agenda - ALWAYS inherit from this when
//                defining your own agenda!
// note: this collects all methods which require private access to the various
//       other classes.
// ===========================================================================

// the only reason I didn't use stack<> is it doesn't have clear() method
template <class VT, class ET, class VRef, class VPtr, class Cmp>
struct Agenda {
  typedef std::set<_Vertex<VT,ET,Cmp>*> History;

  History history;

  void insertHistory (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    history.insert(gi.vPtr);
  }

  // return true if gi is *not* visited
  bool findHistory (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    return (history.find(gi.vPtr)==history.end());
  }

  // effectively set gi to nullGI but retain gPtr (so _GIterator<>::nextComponent()
  // could work)
  void setNullGI (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) const { assert(gi.gPtr); gi.vPtr=NULL; }

  // do a linear search through all vertices to find the first undiscovered vertex
  bool nextStartVertex (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) const;
};

template <class VT, class ET, class VRef, class VPtr, class Cmp>
bool Agenda<VT,ET,VRef,VPtr,Cmp>::nextStartVertex (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) const {
  assert(gi.gPtr);

  _GVMap<VT,ET,Cmp> &vMap=gi.gPtr->vMap;
  typename _GVMap<VT,ET,Cmp>::iterator vmi;
  typename std::set<_Vertex<VT,ET,Cmp>*>::iterator vsi;

  // a linear search to find the first undiscovered vertex
  for (vmi=vMap.begin();vmi!=vMap.end();++vmi)
    for (vsi=vmi->second.begin();vsi!=vmi->second.end();++vsi)
      if (history.find(*vsi)==history.end()) {
	gi.vPtr=*vsi;
	return true;
      }  
  return false;
}

// ===========================================================================
// struct PreorderAgenda: agenda for preorder traversal
// ===========================================================================

template <class VT, class ET, class VRef, class VPtr, class Cmp>
struct PreorderAgenda: protected Agenda<VT,ET,VRef,VPtr,Cmp>, 
		       public std::list<_GNIterator<VT,ET,VRef,VPtr,Cmp> > {
  _GIteratorBase<VT,ET,VRef,VPtr,Cmp> start;

  void init (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    if (gi==gi.nullGI) return;
    
    history.clear();
    clear();
    start=gi;
    insertHistory(gi);
  }

  // if a undiscovered vertex exists gi will be pointed to it and the agenda is
  // initialized, ready for another traversal within a new connected component
  bool nextComponent (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    if (nextStartVertex(gi)) {
      start=gi;
      insertHistory(gi);
      return true;
    }
    else return false;
  }

  void visitNext (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;

    if ((i=gi.outDegree())!=0) {
      for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
      if (i) {
	push_front(gni);
	insertHistory(gni);
	gi=gni;
	return;
      }
    }
    
    while (size()) {
      gni=*begin();
      erase(begin());
      while (++gni!=gni.nullGI && !findHistory(gni));
      if (gni!=gni.nullGI) {
	push_front(gni);
	insertHistory(gni);
	gi=gni;
	return;
      }
    }

    setNullGI(gi);
  }
};

template <class VT, class ET, class VRef, class VPtr, class Cmp>
bool _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::isConnectedTo (const GConstIteratorBase(VT,ET,Cmp) &iter) const {
  GConstIterator(VT,ET,PreorderAgenda,Cmp) gpi;

  for (gpi=*this;gpi!=gpi.nullGI;++gpi)
    if (gpi==iter) return true;
  return false;
}

// ===========================================================================
// struct PostorderAgenda: agenda for postorder traversal
// ===========================================================================

// the only reason I didn't use stack<> is it doesn't have clear() method
template <class VT, class ET, class VRef, class VPtr, class Cmp>
struct PostorderAgenda: protected Agenda<VT,ET,VRef,VPtr,Cmp>, 
			public std::list<_GNIterator<VT,ET,VRef,VPtr,Cmp> > {
  _GIteratorBase<VT,ET,VRef,VPtr,Cmp> start;

  void init (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    if (gi==gi.nullGI) return;

    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;

    history.clear();
    clear();
    start=gi;

    // push all the first undiscovered neighbors until we find a sink vertex
    insertHistory(gi);
    if ((i=gi.outDegree())!=0) {
      for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
      while (i) {
	push_front(gni);
	insertHistory(gni);
	gi=gni;
	if ((i=gi.outDegree())!=0)
	  for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
	else break;
      }
    }
  }

  // find another untraversed component; gi points to the first vertex
  // found in the component
  bool nextComponent (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;

    if (nextStartVertex(gi)) {
      start=gi;
      
      // push all the first undiscovered neighbors until we find a sink vertex
      insertHistory(gi);
      if ((i=gi.outDegree())!=0) {
	for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
	while (i) {
	  push_front(gni);
	  insertHistory(gni);
	  gi=gni;
	  if ((i=gi.outDegree())!=0)
	    for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
	  else break;
	}
      }
      
      return true;
    }
    else return false;
  }

  void visitNext (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;

    if (size()) {
      gni=*begin();
      erase(begin());
      while (++gni!=gni.nullGI && !findHistory(gni));
      if (gni!=gni.nullGI)
	// push all the first undiscovered neighbors until we find a sink vertex
	do {
	  push_front(gni);
	  insertHistory(gni);
	  gi=gni;
	  if ((i=gi.outDegree())!=0)
	    for (gni=gi.outBegin();i && !findHistory(gni);++gni,--i);
	  else break;
	} while (i);
      else if (size()) gi=*begin();
      else gi=start;
    }
    else setNullGI(gi);
  }
};

// ===========================================================================
// struct BFSAgenda: agenda for Breadth-First-Search traversal
// ===========================================================================

// the only reason I didn't use queue<> is it doesn't have clear() method
template <class VT, class ET, class VRef, class VPtr, class Cmp>
struct BFSAgenda: protected Agenda<VT,ET,VRef,VPtr,Cmp>, 
		  public std::list<_GIteratorBase<VT,ET,VRef,VPtr,Cmp> > {
  void init (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &start) {
    if (start==start.nullGI) return;

    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;  // neighbor iterator
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;  

    history.clear();
    clear();

    // push all neighbors from left to right
    insertHistory(start);
    for (gni=start.outBegin(),i=start.outDegree();i;++gni,--i)
      if (gni!=start) {                  // avoid self-looping edges
	insertHistory(gni);              // the vertex is discovered
	push_back(gni);
      }
  }

  // if a unqueued and unvisited vertex exists gi will be pointed to it
  // and the agenda is initialized, ready for another traversal within a new
  // connected component
  bool nextComponent (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;  // neighbor iterator
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;
    
    if (nextStartVertex(gi)) {
      // push all neighbors from left to right
      insertHistory(gi);
      for (gni=gi.outBegin(),i=gi.outDegree();i;++gni,--i)
	if (findHistory(gni)) {
	  insertHistory(gni);          // the vertex is discovered
	  push_back(gni);
	}
      
      return true;
    }
    else return false;
  }
  
  void visitNext (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    _GNIterator<VT,ET,VRef,VPtr,Cmp> gni;  // neighbor iterator
    typename _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::size_type i;

    if (size()) {
      gi=*begin();
      erase(begin());
      
      // push all neighbors from left to right
      for (gni=gi.outBegin(),i=gi.outDegree();i;++gni,--i)
	if (findHistory(gni)) {
	  insertHistory(gni);          // the vertex is discovered
	  push_back(gni);
	}
    }
    else setNullGI(gi);
  }
};

// ===========================================================================
// struct FrontAgenda: agenda for front traversal
// ===========================================================================

template <class VT, class ET, class VRef, class VPtr, class Cmp>
struct FrontAgenda: public PostorderAgenda<VT,ET,VRef,VPtr,Cmp> {
  void visitNext (_GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    do {
      PostorderAgenda<VT,ET,VRef,VPtr,Cmp>::visitNext(gi);
    } while (gi.outDegree());
  }
};

// ===========================================================================
// class _GPathIterator: iterating along a pre-defined path
// NOTE: 1. use the macros GPathIterator(VT,Cmp) and GConstPathIterator(VT,Cmp)
//       2. the first increment should bring the iterator to a node whose
//          data matches the first element of 'path'
// ===========================================================================

#define GPathIterator(VT,ET,Cmp) _GPathIterator<VT,ET,VT&,VT*,Cmp>
#define GConstPathIterator(VT,ET,Cmp) _GPathIterator<VT,ET,const VT&,const VT*,Cmp>

template <class VT, class ET, class VRef, class VPtr, class Cmp>
class _GPathIterator: public _GIteratorBase<VT,ET,VRef,VPtr,Cmp>, 
#if _MSC_VER==1200
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t> {
#else
  public std::iterator<std::bidirectional_iterator_tag,VT,std::ptrdiff_t,VPtr,VRef> {
#endif
public:

  typedef std::list<VT> Path;

private:

  struct HEntry {  // History entry
    _Vertex<VT,ET,Cmp> *vPtr;
    bool updatePI;    // true iff when adding this entry pi is updated (incremented)
    
    HEntry (_Vertex<VT,ET,Cmp> *vPtr, bool updatePI):vPtr(vPtr),updatePI(updatePI) {}
  };

  typedef std::list<HEntry> History;

  // these shadow the two in _GIteratorBase
  typename _Vertex<VT,ET,Cmp>::_VMap::iterator vmi;
  typename _Vertex<VT,ET,Cmp>::_EMap::iterator emi;

  History history;
  const Path *pathPtr;      // WARNING: we don't keep a local copy of Path!
  typename Path::const_iterator pi;

public:

  // WARNING: the path associated with this iterator must have a longer lifespan
  //          than the iterator!

  _GPathIterator ():pathPtr(NULL) {}
  _GPathIterator (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi):pathPtr(NULL) { operator=(gi); }
  _GPathIterator (const Path &path):pathPtr(&path),pi(path.begin()) {}

  // WARNING: must call _GPathIterator(const Path&) or setPath() first!
  _GPathIterator &operator = (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    assert(pathPtr);

    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::operator=(gi);
    pi=pathPtr->begin();
    history.clear();
    return *this;
  }

  // point to some node without consuming any path element
  // WARNING: must call _GPathIterator(const Path&) or setPath() first!
  void moveTo (const _GIteratorBase<VT,ET,VRef,VPtr,Cmp> &gi) {
    assert(pathPtr);

    history.push_back(HEntry(vPtr,false));
    _GIteratorBase<VT,ET,VRef,VPtr,Cmp>::operator=(gi);
  }

  // NOTE pi always points to the label of the *next* visiting vertex
  const typename Path::const_iterator &readPI () const { return pi; }

  // returns how many path elements are *left* to visit
  // call readPath().size() to get the length of the entire path
  typename Path::size_type pathDist () const { return distance(pi,pathPtr->end()); }

  void setPath (const Path &path) {
    pathPtr=&path; 
    pi=path.begin();
    history.clear();
  }
  const Path &readPath () const { assert(pathPtr); return *pathPtr; }

  // next alternative
  // WARNING: bomb if operator++() is not called first
  bool next () {
    if ((++emi)==vmi->second.end()) return false;
    vPtr=emi->first;
    return true;
  }

  // preincrement & postincrement
  // at most you'll hit a nullGI but you can back to non-nullGI by '--'
  _GPathIterator &operator ++ () {
    assert(pathPtr && vPtr);

    if (pi==pathPtr->end() ||
	(vmi=vPtr->outMap.find(&(*pi)))==vPtr->outMap.end()) {
      history.push_back(HEntry(vPtr,false));
      vPtr=NULL;
    }
    else {
      history.push_back(HEntry(vPtr,true));
      emi=vmi->second.begin();
      vPtr=emi->first;
      ++pi;
    }
    
    return *this;
  }
  _GPathIterator operator ++ (int) {
    _GPathIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  // at most you'll return where you started - it never becomes nullGI!
  _GPathIterator &operator -- () {
    if (history.size()) {
      HEntry &he=history.back();
      
      if (he.updatePI) --pi;
      vPtr=he.vPtr;
      history.pop_back();
    }
    return *this;
  }
  _GPathIterator operator -- (int) {
    _GPathIterator tmp=*this;
    --*this;
    return tmp;
  }  
};

};

#endif
