// ======================================================================
// graphTst.hpp - Testing/demonstrating the use of graph.hpp.
// 
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to suppress
//         warnings under -pedantic -Wall.
// 082901: Benjamin Han <benhdj@cs.cmu.edu> Renamed disconnect() to 
//         removeEdgeTo() in _GIteratorBase<>.
// 081301: Benjamin Han <benhdj@cs.cmu.edu> Added tests for connections/edges
//         between vertices and for embedded preorder traversal.
// 080901: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "graph.hpp"

using namespace Toolbox;
using namespace std;

struct Label {
  char c;

  Label () {}
  Label (char c):c(c) {}

  bool operator < (const Label &other) const { return c<other.c; }
  bool operator == (const Label &other) const { return c==other.c; }
};

ostream &operator << (ostream &os, const Label &l) {
  return os<<l.c;
}

struct Double {
  double d;

  Double () {}
  Double (double d):d(d) {}
};

ostream &operator << (ostream &os, const Double &d) {
  return os<<d.d;
}

int main () {
  Graph<Label,Double> g1,g2;
  map<char,Graph<Label,Double>::Iterator> vMap;

  Graph<Label,Double>::ConstDIterator di1;
  Graph<Label,Double>::ConstVIterator vi1;

  Graph<Label,Double>::ConstNIterator ni1;
  Graph<Label,Double>::NIterator ni2;

  Graph<Label,Double>::Iterator gi;

  Graph<Label,Double>::ConstPreIterator gi1,gi5;
  Graph<Label,Double>::ConstPostIterator gi2;
  Graph<Label,Double>::ConstBFSIterator gi3;
  Graph<Label,Double>::ConstFrontIterator gi4;

  Graph<Label,Double>::ConstPathIterator::Path path;
  Graph<Label,Double>::ConstPathIterator gpi;  

  // ===========================================================================
  // constructing graph g1
  for (char c='a';c<='h';++c) vMap[c]=g1.add(c);
  vMap['a'].addEdgeTo(vMap['b'],1.0);
  vMap['b'].addEdgeTo(vMap['c'],0.33);
  vMap['b'].addEdgeTo(vMap['e'],0.33);
  vMap['b'].addEdgeTo(vMap['f'],0.33);
  vMap['c'].addEdgeTo(vMap['d'],0.5);
  vMap['c'].addEdgeTo(vMap['g'],0.5);
  vMap['d'].addEdgeTo(vMap['c'],0.5);
  vMap['d'].addEdgeTo(vMap['h'],0.5);
  vMap['e'].addEdgeTo(vMap['a'],0.5);
  vMap['e'].addEdgeTo(vMap['f'],0.5);
  vMap['f'].addEdgeTo(vMap['g'],1.0);
  vMap['g'].addEdgeTo(vMap['f'],0.5);
  vMap['g'].addEdgeTo(vMap['h'],0.5);
  vMap['h'].addEdgeTo(vMap['h'],1.0);  // self-looping edge

  // ===========================================================================
  // show the graph
  cout<<"* Graph g1: (# of vertices="<<g1.size()<<"; # of edges="<<g1.eSize()
      <<"; # of vertex data="<<g1.dSize()<<')'<<endl;
  for (char c='a';c<='h';++c) {
    Graph<Label,Double>::Iterator &vi=vMap[c];
    
    cout<<"+ Vertex "<<c<<" ["<<vi.inDegree()<<','
	<<vi.outDegree()<<"]:"<<endl;
    cout<<"  inward edges:";
    for (ni1=vi.inBegin();ni1!=vi.inEnd();++ni1)
      cout<<" ["<<*ni1<<','<<ni1.readEdge(vi)<<']';
    cout<<endl<<"  outward edges:";
    for (ni1=vi.outBegin();ni1!=vi.outEnd();++ni1)
      cout<<" ["<<*ni1<<','<<vi.readEdge(ni1)<<']';
    cout<<endl;
  }

  // ===========================================================================
  // Data iterator test
  cout<<endl<<"* Iterating using data and vertex iterators:"<<endl;
  for (di1=g1.dBegin();di1!=g1.dEnd();++di1) {
    cout<<*di1<<" ["<<di1.size()<<"]: ";
    for (vi1=di1.begin();vi1!=di1.end();++vi1) cout<<*vi1<<' ';
    cout<<endl;
  }
  
  // ===========================================================================
  // connection test
  cout<<endl<<"* Vertex 'a' "<<(g1.find(Label('a')).isConnectedTo(g1.find(Label('g')))?
				"is":"is not")<<" connected to vertex 'g'."<<endl;
  cout<<"* Vertex 'a' "<<(g1.find(Label('a')).hasEdgeTo(g1.find(Label('g')))?
			  "has":"does not have")<<" an edge to vertex 'g'."<<endl;  
  cout<<endl<<"* Vertex 'h' "<<(g1.find(Label('h')).isConnectedTo(g1.find(Label('f')))?
				"is":"is not")<<" connected to vertex 'f'."<<endl;

  // ===========================================================================
  // Preorder traversal
  cout<<endl<<"* Preorder traversal of g1 from 'a':"<<endl; 
  gi1=g1.begin();
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  cout<<endl<<"* Preorder traversal of g1 from 'c':"<<endl; 
  gi1=g1.find(Label('c'));
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Embedding preorder traversals
  cout<<endl<<"* Embedding preorder traversals of g1 (the top-level one is from 'a'):"<<endl;
  gi1=g1.begin();
  do {
    for (;gi1!=gi1.nullGI;++gi1) {
      cout<<*gi1<<endl;
      cout<<"  embedding traversal from this vertex:"<<endl;
      for (gi5=gi1,gi5.start();gi5!=gi5.nullGI;++gi5) cout<<"  "<<*gi5<<endl;
    }
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Postorder traversal
  cout<<endl<<"* Postorder traversal of g1 from 'a':"<<endl; 
  gi2=g1.begin();
  do {
    for (;gi2!=gi2.nullGI;++gi2) cout<<*gi2<<endl;
    cout<<" - next component -"<<endl;
  } while (gi2.nextComponent());
  cout<<endl;

  // ===========================================================================
  // BFS traversal
  cout<<endl<<"* BFS traversal of g1 from 'a':"<<endl; 
  gi3=g1.begin();
  do {
    for (;gi3!=gi3.nullGI;++gi3) cout<<*gi3<<endl;
    cout<<" - next component -"<<endl;
  } while (gi3.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Front traversal
  cout<<endl<<"* Front traversal of g1 from 'a':"<<endl; 
  gi4=g1.begin();
  do {
    for (;gi4!=gi4.nullGI;++gi4) cout<<*gi4<<endl;
    cout<<" - next component -"<<endl;
  } while (gi4.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Path traversal
  path.push_back(Label('b'));
  path.push_back(Label('c'));
  path.push_back(Label('d'));
  path.push_back(Label('h'));
  path.push_back(Label('h'));
  cout<<endl<<"* Path traversal of g1 from 'a':"<<endl;
  gpi.setPath(path);
  gi=g1.begin();
  for (gpi=gi;gpi!=gpi.nullGI;++gpi) cout<<*gpi<<endl;

  // Backward path traversal
  cout<<endl<<"* Backward path traversal from 'h' to 'a' in g1:"<<endl;
  do {
    cout<<*(--gpi)<<endl;
  } while (gpi!=gi);
  
  // ===========================================================================
  // graph assignment
  g2=g1;
  cout<<endl<<"* g1 is assigned to g2: "<<g2.size()<<" vertices."<<endl;

  // ===========================================================================
  // Postorder traversal
  cout<<endl<<"* Postorder traversal of g2 from 'c':"<<endl; 
  gi2=g2.find(Label('c'));
  do {
    for (;gi2!=gi2.nullGI;++gi2) cout<<*gi2<<endl;
    cout<<" - next component -"<<endl;
  } while (gi2.nextComponent());
  cout<<endl;

  // ===========================================================================
  // BFS traversal
  cout<<endl<<"* BFS traversal of g2 from 'c':"<<endl; 
  gi3=g2.find(Label('c'));
  do {
    for (;gi3!=gi3.nullGI;++gi3) cout<<*gi3<<endl;
    cout<<" - next component -"<<endl;
  } while (gi3.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Front traversal
  cout<<endl<<"* Front traversal of g2 from 'c':"<<endl; 
  gi4=g2.find(Label('c'));
  do {
    for (;gi4!=gi4.nullGI;++gi4) cout<<*gi4<<endl;
    cout<<" - next component -"<<endl;
  } while (gi4.nextComponent());
  cout<<endl;

  // ===========================================================================
  // Path traversal
  path.clear();
  path.push_back(Label('d'));
  path.push_back(Label('c'));
  path.push_back(Label('g'));
  cout<<endl<<"* Path traversal of g2 from 'c':"<<endl;
  gpi.setPath(path);
  gi=g2.find(Label('c'));
  for (gpi=gi;gpi!=gpi.nullGI;++gpi) cout<<*gpi<<endl;

  // ===========================================================================
  // Backward path traversal
  cout<<endl<<"* Backward path traversal from 'g' to 'c' in g2:"<<endl;
  do {
    cout<<*(--gpi)<<endl;
  } while (gpi!=gi);
  
  // ===========================================================================
  // disconnecting two nodes
  cout<<endl<<"* Remove edge from 'b' to 'c' in g2 (preorder printout from 'a'):"<<endl;
  g2.find(Label('b')).removeEdgeTo(g2.find(Label('c')));
  gi1=g2.begin();
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // updating edge data
  cout<<endl<<"* Change edge values of ('b','e') and ('b','f') to 0.5 in g2:"<<endl;
  gi1=g2.find(Label('b'));
  gi1.getEdge(g2.find(Label('e')))=0.5;
  gi1.getEdge(g2.find(Label('f')))=0.5;
  cout<<"  edge ('b','e') = "<<gi1.readEdge(g2.find(Label('e')))<<endl;
  cout<<"  edge ('b','f') = "<<gi1.readEdge(g2.find(Label('f')))<<endl;

  // ===========================================================================
  // removing a vertex
  cout<<endl<<"* Prune vertex 'f' away in g2 (preorder printout from 'a'):"<<endl;
  g2.pruneOne(gi=g2.find(Label('f')));
  gi1=g2.begin();
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // duplicating part of the graph
  cout<<endl<<"* Self-duplicating part of g2 from 'a' (preorder printout from 'a'):"<<endl;
  gi1=g2.add(g2.begin());
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // copying part of the graph
  cout<<endl<<"* Copying part of g1 from 'c' to g2 (preorder printout from 'c'):"<<endl;
  gi1=gi=g2.add(g1.find(Label('c')));
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // prune away part of the graph
  cout<<endl<<"* Prune away the just-added subgraph from g2 (preorder printout from 'c'):"<<endl;
  g2.prune(gi);
  gi1=g2.find(Label('c'));
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // erase a neighbor (!)
  cout<<endl<<"* Erase the 1st outward neighbor of 'c' in g2 (preorder printout from 'c'):"<<endl;
  ni2=g2.eraseOne(ni2=g2.find(Label('c')).outBegin());
  cout<<"  the next neighbor is "<<*ni2<<endl;
  gi1=g2.find(Label('c'));
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // updating a vertex
  cout<<endl<<"* Update the vertex label 'c' to 'x' in g2 (preorder printout from 'x'):"<<endl;
  gi1=g2.find(Label('c'));
  gi1.update('x');
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  // ===========================================================================
  // clearing a graph
  cout<<endl<<"* Clear g2 (preorder printout):"<<endl;
  g2.clear();
  gi1=g2.begin();
  do {
    for (;gi1!=gi1.nullGI;++gi1) cout<<*gi1<<endl;
    cout<<" - next component -"<<endl;
  } while (gi1.nextComponent());
  cout<<endl;

  return 0;
}
