// ======================================================================
// random.cpp - random number generation stuff
// 
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added selectMulti() functions.
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <time.h>
#include <assert.h>

#include "random.hpp"

namespace Toolbox {

void Range::computePDist (PDist &accP) const {
  const_iterator ri;
  size_type i;

  accP.resize(size());

  for (i=1,ri=begin(),accP[0]=ri->p,++ri;ri!=end();++i,++ri)
    accP[i]=accP[i-1]+ri->p;
}

// return a random index according to 'accP' - an accumulated
// probability distribution (roulette wheel selection)
int select (const PDist &accP) {
  return lower_bound(accP.begin(),accP.end(),
		     ((double)random()/(double)RAND_MAX)*accP[accP.size()-1])
         -accP.begin();
}

void initRandom () {
  time_t t;

  srandom(time(&t));
}

// same with the above 'select' but guranteed no 'ex'
int select (const PDist &accP, int ex) {
  double exP=(ex==0?accP[0]:accP[ex]-accP[ex-1]); // prob. of the excluded item

  // randD is a random double in [0.0,total P - exP)
  double randD=((double)random()/(double)RAND_MAX)*(accP[accP.size()-1]-exP);
  int pos=lower_bound(accP.begin(),accP.end(),randD)-accP.begin();

  // because accP is monotonically increasing, we can adjust pos if it
  // falls into ex range
  if (pos>=ex)
    pos=lower_bound(accP.begin()+ex+1,accP.end(),randD+exP)-accP.begin();

  return pos;
}

// select n distinguished elements according to the accumulated
// prob. dist. 'accP'
void selectMulti (const PDist &accP, PDist::size_type n, std::set<int> &s) {
  // initializing temp is just to suppress the warning of the use of an 
  // uninitialized var - but in this case even without this temp is guaranteed
  // to become initialized before it's used in the "if (eandD<0.0)..." block

  double randD,temp=0.0,sumP=accP[accP.size()-1];
  PDist::size_type i,pos,startPos,endPos;
  std::set<int>::const_iterator si;

  s.clear();
  if (n>accP.size()) n=accP.size();
  for (i=0;i<n;++i,sumP-=(pos==0?accP[0]:accP[pos]-accP[pos-1])) {
    randD=((double)random()/(double)RAND_MAX)*sumP;
    
    if (s.empty() || 
	(*s.begin()!=0 && randD<accP[*s.begin()-1])) {
      pos=lower_bound(accP.begin(),accP.end(),randD)-accP.begin();
      s.insert(pos);
    }
    else {
      assert(randD>0.0);
      
      for (si=s.begin(),startPos=0,endPos=*si;;) {
	// subtract the sum of prob. of [startPos,endPos)
	if (endPos) {
	  temp=(startPos==0?
		accP[endPos-1]:
		accP[endPos-1]-accP[startPos-1]);
	  
	  randD-=temp;
	}
	
	++si;
	if (si!=s.end() && randD>0.0) {
	  startPos=endPos+1;
	  endPos=*si;
	}
	else break;
      }
      
      if (randD<0.0)
	// search from startPos
	// temp is guaranteed to become initiailzed at this point because
	// the only randD can be less than 0.0 is by entering the
	// "if (endPos)..." block above
	for (pos=startPos,randD+=temp;
	     randD-(accP[pos]-accP[startPos-1])>0.0;++pos);
      else
	// search from endPos+1
	for (pos=endPos+1;
	     randD-(accP[pos]-accP[endPos])>0.0;++pos);
      
      s.insert(pos);
    }
  }
  
  assert(s.size()==n);
}

// select one val from range 'r' according to the prob. dist.
int select (const Range &r) {
  PDist accP;

  r.computePDist(accP);
  return r[select(accP)].val;
}

// same with the above 'select' but guaranteed no 'ex'
int select (const Range &r, int ex) {
  PDist accP;
  Candidate c(ex);

  r.computePDist(accP);
  return r[select(accP,lower_bound(r.begin(),r.end(),c)-r.begin())].val;
}

// select n distinguished vals from range 'r' according to the
// prob. dist.
void selectMulti (const Range &r, PDist::size_type n, std::set<int> &s) {
  PDist accP;
  std::set<int> tempSet;
  std::set<int>::iterator si;

  s.clear();

  r.computePDist(accP);
  selectMulti(accP,n,tempSet);
  for (si=tempSet.begin();si!=tempSet.end();++si) s.insert(r[*si].val);  
}

};
