// ======================================================================
// random.hpp - random number generation stuff
// 
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 052901: Benjamin Han <benhdj@cs.cmu.edu> Minor revision.
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 102500: Benjamin Han <benhdj@cs.cmu.edu> Added selectMulti() functions.
// 101900: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef RANDOM_HPP
#define RANDOM_HPP

#include "tbconsts.hpp"

#include <vector>
#include <set>
#include <algorithm>

#include <stdlib.h>
#include <assert.h>

namespace Toolbox {

typedef std::vector<double> PDist;          // probability distribution

struct Candidate {
  int val;         // a candidate value
  double p;        // probability of this candidate

  Candidate () {}
  Candidate (int val):val(val) {}

  bool operator < (const Candidate &c) const { return val<c.val; }
};

struct Range: public std::vector<Candidate> {
  // compute an accumulated prob. dist.
  void computePDist (PDist &accP) const;
};

typedef std::set<int> RangeSet;             // automatically sorted

// init the random number generator with the current time
void initRandom ();

// function object for random_shuffle()
class RandomIntGenerator {
public:

  ptrdiff_t operator () (ptrdiff_t max) { return random()%max; }
};

// return a random index according to 'accP' - an accumulated
// probability distribution (roulette wheel selection)
int select (const PDist &accP);

// same with the above 'select' but guranteed no 'ex'
int select (const PDist &accP, int ex);

// select n distinguished elements according to the accumulated
// prob. dist. 'accP'
void selectMulti (const PDist &accP, PDist::size_type n, std::set<int> &s);

// select one val from range 'r' according to the prob. dist.
int select (const Range &r);

// same with the above 'select' but guaranteed no 'ex', which is the
// val of a candidate
int select (const Range &r, int ex);

// select n distinguished vals from range 'r' according to the
// prob. dist. 
void selectMulti (const Range &r, PDist::size_type n, std::set<int> &s);

// ============================ inline functions ============================

// return a random int in [0,max)
inline int getRandInt (int max) {
  assert(max>0);
  return random()%max;
}

};

#endif
