// ======================================================================
// regex.cpp - Regular expression matching.
// 
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 102101: Benjamin Han <benhdj@cs.cmu.edu> Added replace().
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 071801: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "regex.hpp"

#include "misc.hpp"

namespace Toolbox {

static Regex replaceRegRegex("([\\]+)([[:digit:]]+)");
//                            ----- --------------
//                       reg 1: \'s  reg 2: reg no.

bool Regex::isEscaped (std::string::size_type i, const std::string &str) const {
  std::string::size_type n;

  if (i==0) return false;
  for (--i,n=0;i && str[i]=='\\';--i,++n);
  if (i==0 && str[i]=='\\') ++n;
  return (n%2==1);
}

bool Regex::isBracketed (std::string::size_type i, const std::string &str) const {
  std::string::size_type j,k;

  if ((j=str.rfind('[',i))!=std::string::npos && !isEscaped(j,str))
    return ((k=str.find(']',j))!=std::string::npos && !isEscaped(k,str) &&
	    i<k);
  else return false;
}

void Regex::set (const std::string &pattern, int cFlags) {
  std::string::size_type i,j,numLP;

  clear();
  Regex::pattern=pattern;
  Regex::cFlags=cFlags;

  // find out how many registers we need
  for (i=numLP=0;(j=pattern.find_first_of("()",i))!=std::string::npos;i=j+1)
    if (pattern[j]=='(') {
      if (!isEscaped(j,pattern) && !isBracketed(j,pattern))
	++numLP;
    }
    else if (!isEscaped(j,pattern) && !isBracketed(j,pattern))
      if (numLP>0) {
	--numLP;
	++numRegs;
      }
      else {
	std::string iStr;
	
	intToString(j,iStr);
	throw RegexErrorLParen(iStr);
      }
  
  if (numLP) {
    std::string iStr;
    
    intToString(j,iStr);
    throw RegexErrorRParen(iStr);
  }
  pmatch=new regmatch_t[numRegs];

  // now compile the regex
  preg=new regex_t;
  if ((i=regcomp(preg,pattern.c_str(),cFlags))!=0) {
    char *buf;
    std::string errMsg;
    
    j=regerror(i,preg,NULL,0);
    buf=new char[j];
    regerror(i,preg,buf,j);
    errMsg=buf;
    delete[] buf;
    
    throw RegexError(errMsg);
  }
}

std::string::size_type Regex::replace (std::string &str, const std::string &replaceStr, std::string::size_type i) {
  if (match(str,i)!=std::string::npos) {
    std::string tmpStr;
    std::string::size_type copyPos,j;
    int regNo;
    
    // set up tmpStr as the register-replaced replacement string
    for (copyPos=j=0;
	 replaceRegRegex.match(replaceStr,j)!=std::string::npos;) {

      // if \ is escaped, or the register # is invalid, don't try to replace it
      // with the content of a register
      if (replaceRegRegex.length(1)%2==0 ||
	  (regNo=atoi(replaceStr.substr(replaceRegRegex.startPos(2),
					replaceRegRegex.length(2)).c_str()))>=numRegs)
	j=replaceRegRegex.endPos();
      else {
	tmpStr+=replaceStr.substr(copyPos,replaceRegRegex.startPos()-copyPos);
	tmpStr+=str.substr(pmatch[regNo].rm_so,pmatch[regNo].rm_eo-pmatch[regNo].rm_so);
	copyPos=j=replaceRegRegex.endPos();
      }
    }
    tmpStr+=replaceStr.substr(copyPos);

    // replace the matched substring with tmpStr
    str.erase(pmatch[0].rm_so,pmatch[0].rm_eo-pmatch[0].rm_so);
    str.insert(pmatch[0].rm_so,tmpStr);

    return pmatch[0].rm_so+tmpStr.length();
  }
  else return std::string::npos;
}

};
