// ======================================================================
// regex.hpp - Regular expression matching.
// 
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 102402: Benjamin Han <benhdj@cs.cmu.edu> Added the missing #include
//         <assert.h>.
// 102101: Benjamin Han <benhdj@cs.cmu.edu> Added remove(), removeAll(),
//         replace(), and replaceAll().
// 101901: Benjamin Han <benhdj@cs.cmu.edu> Fixed a stupid bug in 
//         Regex::match() - when matching starting from i!=0 we need to
//         shift the indices in pmatch as well.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 071801: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef REGEX_HPP
#define REGEX_HPP

#include "tbconsts.hpp"
#include "except.hpp"

#include <regex.h>
#include <assert.h>

#include <string>

namespace Toolbox {

DEFINE_SIMPLE_EXCEPTION(RegexError,"Regex error: %s.");
DEFINE_SIMPLE_EXCEPTION(RegexErrorLParen,"Regex error: Unmatched \'(\' at %s.");
DEFINE_SIMPLE_EXCEPTION(RegexErrorRParen,"Regex error: Unmatched \')\' at %s.");

class Regex {
  std::string pattern;
  regex_t *preg;
  regmatch_t *pmatch;
  int numRegs;
  int cFlags;
  int eFlags;

  bool isEscaped (std::string::size_type i, const std::string &str) const;
  bool isBracketed (std::string::size_type i, const std::string &str) const;

  void clear ();

public:

  // For setting up a Regex, the valid cFlags is 0, or a bitwise-or of one
  // or more of the following (see manpage for regcomp):
  //   REG_EXTENDED: use POSIX extended syntax
  //   REG_ICASE:    case insensitive
  //   REG_NOSUB:    no substring addressing (registers)
  //   REG_NEWLINE:  don't match newline.  

  // For matching a Regex, the valid eFlags is 0, or a bitwise-or of one
  // or more of the following (see manpage for regexec):
  //   REG_NOTBOL:   don't match the beginning of line.
  //   REG_NOTEOL:   don't match the end of line.

  Regex ():preg(NULL),pmatch(NULL),numRegs(1),cFlags(0),eFlags(0) {}
  Regex (const std::string &pattern, int cFlags=REG_EXTENDED);
  Regex (const Regex &src);
  ~Regex () { clear(); }

  // return the # of registers
  int size () const { return numRegs; }

  const std::string &readPattern () const { return pattern; }

  void set (const std::string &pattern, int cFlags=REG_EXTENDED);
  void setEFlags (int eFlags) { Regex::eFlags=eFlags; }

  // start from str[i]; matching behavior affected by setEFlags()
  // return the starting position of the match (or std::string::npos if no match
  // is found)
  std::string::size_type match (const std::string &str, std::string::size_type i=0);

  // start from str[i]; if a match is found remove it
  // return the starting position of the removal (or std::string::npos if no match
  // is found)
  std::string::size_type remove (std::string &str, std::string::size_type i=0);

  // start from str[i]; remove all matches found
  // return the # or removals
  int removeAll (std::string &str, std::string::size_type i=0);

  // start from str[i]; if a match is found replace it with replaceStr
  // you can use \x (x=0 ... # of registers) to replace with the matched part
  // return the ending position of the replacement (or std::string::npos if no 
  // match is found), i.e., the starting position for the next regex operation
  std::string::size_type replace (std::string &str, const std::string &replaceStr, std::string::size_type i=0);

  // start from str[i]; replace all matches with replaceStr
  // you can use \x (x=0 ... # of registers) to replace with the matched part
  // return the # of replacements
  int replaceAll (std::string &str, const std::string &replaceStr, std::string::size_type i=0);

  // the following methods are used to obtain the info about the matches
  bool isMatched (int i=0) const { assert(pmatch); return (pmatch[i].rm_so!=-1); }
  std::string::size_type startPos (int i=0) const { assert(pmatch); return pmatch[i].rm_so; }
  std::string::size_type endPos (int i=0) const { assert(pmatch); return pmatch[i].rm_eo; }
  std::string::size_type length (int i=0) const { assert(pmatch); return pmatch[i].rm_eo-pmatch[i].rm_so; }

  Regex &operator = (const Regex &src);
};

// ============================ inline functions ============================

inline Regex::Regex (const std::string &pattern, int cFlags)
  :preg(NULL),pmatch(NULL),numRegs(1),eFlags(0) {
  set(pattern,cFlags);
}

inline void Regex::clear () {
  pattern.erase();
  if (preg) {
    regfree(preg);
    delete preg;
    preg=NULL;
  }
  if (pmatch) {
    delete[] pmatch;
    pmatch=NULL;
  }
  numRegs=1;
  cFlags=eFlags=0;
}

inline Regex::Regex (const Regex &src)
  :preg(NULL),pmatch(NULL),numRegs(1),eFlags(0) {
  set(src.pattern,src.cFlags);
}

inline std::string::size_type Regex::match (const std::string &str, std::string::size_type i) {
  assert(preg);

  if (!regexec(preg,str.c_str()+i,numRegs,pmatch,eFlags)) {
    for (int j=0;j<numRegs;++j) 
      if (pmatch[j].rm_so!=-1) {
      pmatch[j].rm_so+=i;
      pmatch[j].rm_eo+=i;
    }
    
    return pmatch[0].rm_so;
  }
  else return std::string::npos;
}

inline std::string::size_type Regex::remove (std::string &str, std::string::size_type i) {
  if (match(str,i)!=std::string::npos) {
    str.erase(pmatch[0].rm_so,pmatch[0].rm_eo-pmatch[0].rm_so);
    return pmatch[0].rm_so;
  }
  else return std::string::npos;
}

inline int Regex::removeAll (std::string &str, std::string::size_type i) {
  std::string::size_type num,j;

  for (j=i,num=0;(j=remove(str,j))!=std::string::npos;++num);
  return num;
}

inline int Regex::replaceAll (std::string &str, const std::string &replaceStr, std::string::size_type i) {
  std::string::size_type num,j;

  for (j=i,num=0;(j=replace(str,replaceStr,j))!=std::string::npos;++num);
  return num;
}

inline Regex &Regex::operator = (const Regex &src) {
  set(src.pattern,src.cFlags);
  return *this;
}

};

#endif
