// ======================================================================
// tokDict.hpp - a template of token dictionary.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Cleaned up to remove _MetaDict<> -
//         no longer consider gcc <= 2.96; also now compiles under M$ VC6.
// 110102: Benjamin Han <benhdj@cs.cmu.edu> Revised comment: no longer needed
//         string conversion in class T.
// 102702: Benjamin Han <benhdj@cs.cmu.edu> The map itself now uses T as the
//         key instead of string, and tVec (the linear bookkeeping) stores
//         iterators into the map - this saves space when T=string, and allows
//         fast implementation of TokenDict::rollback(); added setRollback() and
//         rollback() so we can get rid of some temporary tokens.
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 081901: Benjamin Han <benhdj@cs.cmu.edu> The _MetaDict<> workaround: this
//         together with TokenID<>::dict() should fix the undefined init
//         ordering of global/static objects (hopefully it's the last time -
//         although why G++ 2.96 has this behavior is still not clear).
// 050701: Benjamin Han <benhdj@cs.cmu.edu> Added clear().
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 030601: Benjamin Han <benhdj@cs.cmu.edu> Added size().
// 030301: Benjamin Han <benhdj@cs.cmu.edu> Added findToken(); added new
//         forms of addToken() and updateToken() so they returned a bool 
//         value indicating if the addition really occurs.
// 030101: Benjamin Han <benhdj@cs.cmu.edu> Added some assertions to ensure
//         no addition once the map is cleared via clearMap(); rename
//         getToken() into readToken(), and added getToken(); hooked up with
//         vFile.hpp so now we can save and load a TokenDict.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 010701: Benjamin Han <benhdj@cs.cmu.edu> Changed the private members
//         to protected members.
// 010201: Benjamin Han <benhdj@cs.cmu.edu> Added updateToken() method.
// 010101: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TOKDICT_HPP
#define TOKDICT_HPP

#include "tbconsts.hpp"
#include "vFile.hpp"

#include <map>
#include <vector>

namespace Toolbox {

// ===========================================================================
// class TokenDict: A dictionary for tokens (type T)
// ===========================================================================

#define UNKNOWN_TOKEN_ID -1

// IMPORTANT: this must exist: bool operator < (const T &other) const;

template <class T>
class TokenDict: public VFObject {
#ifndef NDEBUG
  bool mapCleared;
#endif

protected:

  typedef std::map<T,int> TMap;
  typedef typename TMap::iterator iterator;
  typedef typename TMap::const_iterator const_iterator;
  typedef std::vector<iterator> TVec;
  
  TMap tMap;   // for adding unique tokens
  TVec tVec;   // for token retrieval based on ID
  int id,rollbackID;

public:

  TokenDict ():
#ifndef NDEBUG
    mapCleared(false),
#endif
    id(0),rollbackID(0)
  {}

  typedef typename TVec::size_type size_type;

  size_type size () const { return tVec.size(); }

  int findToken (const T &key) const {
    assert(!mapCleared);
    const_iterator mi=tMap.find(key);
    return (mi!=tMap.end()?mi->second:UNKNOWN_TOKEN_ID);
  }

  T &getToken (size_type i) {
    assert(i<tVec.size() && i>=0);
    return *(const_cast<T*>(&(tVec[i]->first)));
  }

  const T &readToken (size_type i) const {
    assert(i<tVec.size() && i>=0);
    return tVec[i]->first;
  }

  // only add a token if it's not already in;
  // returns the integer 'handle' of the token
  int addToken (const T &t);

  // if the token is not in, it's same as addToken()
  // CAUTION: do NOT change the string associated with 't' if it's already
  //   in the dictionary (map<> assumes the const-ness of the key element)!
  int updateToken (const T &t);

  // the following addToken() and updateToken() are
  // functionally same as the previous ones, but return
  // true iff a new token is added;
  bool addToken (const T &t, int &idx);
  bool updateToken (const T &t, int &idx);

  // after the last addition to the dictionary, we can 'close' it by
  // removing the map structure, thus shedding some memory allocated
  void clearMap () {
#ifndef NDEBUG
    mapCleared=true;
#endif
    tMap.clear();
  }
  
  void clear () {
#ifndef NDEBUG
    mapCleared=false;
#endif    
    tMap.clear();
    tVec.clear();
    id=rollbackID=0;
  }

  void setRollbackPoint () { rollbackID=id; }
  void rollback ();

  // for VFile IO
  void readFromVFile (VFile &vFile);
  void writeToVFile (VFile &vFile) const;
};

template <class T>
int TokenDict<T>::addToken (const T &t) {
  assert(!mapCleared);
  std::pair<iterator,bool> result=tMap.insert(make_pair(t,id));

  if (result.second) {
    tVec.push_back(result.first);
    return id++;
  }
  else return result.first->second;
}

template <class T> 
int TokenDict<T>::updateToken (const T &t) {
  assert(!mapCleared);
  std::pair<iterator,bool> result=tMap.insert(make_pair(t,id));

  if (result.second) {
    tVec.push_back(result.first);
    return id++;
  }
  else {
    *(const_cast<T*>(&result.first->first))=t;
    return result.first->second;
  }
}

template <class T> 
bool TokenDict<T>::addToken (const T &t, int &idx) {
  assert(!mapCleared);
  std::pair<iterator,bool> result=tMap.insert(make_pair(t,id));

  if (result.second) {
    tVec.push_back(result.first);
    idx=id++;
  }
  else idx=result.first->second;
  return result.second;
}

template <class T> 
bool TokenDict<T>::updateToken (const T &t, int &idx) {
  assert(!mapCleared);
  std::pair<iterator,bool> result=tMap.insert(make_pair(t,id));

  if (result.second) {
    tVec.push_back(result.first);
    idx=id++;
  }
  else {
    *(const_cast<T*>(&result.first->first))=t;
    idx=result.first->second;
  }
  return result.second;
}

template <class T> 
void TokenDict<T>::rollback () {
  typename std::vector<iterator>::iterator pos=tVec.begin()+rollbackID;

  for (int i=rollbackID;i<id;++i)
    tMap.erase(tVec[i]);
  tVec.erase(pos,tVec.end());
  id=rollbackID;
}

template <class T> 
void TokenDict<T>::readFromVFile (VFile &vFile) {
  T t;
  std::pair<iterator,bool> result;
  int i;

  clear();

  vFile>>id;
  for (i=0;i<id;++i) {
    vFile>>t;
    result=tMap.insert(make_pair(t,i));
    tVec.push_back(result.first);
  }
}

// do not save the map part
// ASSUMPTION: no 'hole' in tVec (for i=0..id-1, tVec[i] exists)
template <class T> 
void TokenDict<T>::writeToVFile (VFile &vFile) const {
  int i;

  vFile<<id;
  for (i=0;i<id;++i) vFile<<tVec[i]->first;
}

};

#endif
