// ======================================================================
// tokID.hpp - a token ID which encapsulates the internal table: now you
//            can print a token ID as the content, not just the integer
// 
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 102702: Benjamin Han <benhdj@cs.cmu.edu> Revised some comments; added 
//         static methods setRollbackPoint() and rollback() in TokenID; changed 
//         TokenID::dict() into a static method.
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 081901: Benjamin Han <benhdj@cs.cmu.edu> Chagned the static member dict
//         into a method accessing a local static localDict - this enforced
//         init ordering.
// 062801: Benjamin Han <benhdj@cs.cmu.edu> Renamed clearMap(), clear() and
//         size() to dictClearMap(), dictClear() and dictSize(); added
//         clear().
// 061201: Benjamin Han <benhdj@cs.cmu.edu> Removed the initialization of
//         the static member TokenID::dict - we don't want multiple
//         initializations!
// 050701: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TOKID_HPP
#define TOKID_HPP

#include "tokDict.hpp"

#include <iostream>

namespace Toolbox {

// IMPORTANT: 
//   You only get one internal table (dictionary) per data type (since dict is
//   static).
template <class T>
class TokenID {
protected:

  int id;

  // this ensures the local copy is only initialized once, *before* anything
  // can be done on it (enforcing the init ordering)
  // WARNING: DON'T MAKE THIS INLINE - IT'LL BREAK WITH GCC 3.3!
  static TokenDict<T> &dict ();

#if _MSC_VER!=1200
  friend std::ostream &operator << <> (std::ostream &os, const TokenID &tokID);
#endif

public:

  static void dictClearMap () { dict().clearMap(); }
  static void dictClear () { dict().clear(); }
  static typename TokenDict<T>::size_type dictSize () { return dict().size(); }

  // the following two together allows adding temporary tokens and removing
  // them
  static void setRollbackPoint () { dict().setRollbackPoint(); }
  static void rollback () { dict().rollback(); }

  TokenID ():id(UNKNOWN_TOKEN_ID) {}
  TokenID (const T &token) { id=dict().addToken(token); }

  void clear () { id=UNKNOWN_TOKEN_ID; }
  void set (const T &token) { id=dict().addToken(token); }
  void find (const T &token) { id=dict().findToken(token); }

  T& get () { return dict().getToken(id); }
  const T& read () const { return dict().readToken(id); }

  bool isUndefined () const { return id==UNKNOWN_TOKEN_ID; }

  bool operator < (const TokenID &other) const { return id<other.id; }
  bool operator > (const TokenID &other) const { return id>other.id; }
  bool operator <= (const TokenID &other) const { return id<=other.id; }
  bool operator >= (const TokenID &other) const { return id>=other.id; }
  bool operator == (const TokenID &other) const { return id==other.id; }
  bool operator != (const TokenID &other) const { return id!=other.id; }
};


template <class T>
TokenDict<T> &TokenID<T>::dict () {
  static TokenDict<T> localDict;
  return localDict;
}

template <class T>
std::ostream &operator << (std::ostream &os, const TokenID<T> &tokID) {
  return os<<tokID.dict().readToken(tokID.id);
}

};

#endif
