// ======================================================================
// tokStr.cpp - tokenize a string by white spaces.
// 
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 103101: Benjamin Han <benhdj@cs.cmu.edu> Fixed a bug in 
//         TokenString::subTokStr() when len=0.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 062801: Benjamin Han <benhdj@cs.cmu.edu> Simplified find().
// 052801: Benjamin Han <benhdj@cs.cmu.edu> Added another set() allowing
//         upper case conversion.
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 022301: Benjamin Han <benhdj@cs.cmu.edu> Added join() method; added
//         TokenString::npos; changed subTokStr() into a const metnod;
//         fix bug in transform(): return immediately if the TokenString is
//         empty.
// 022101: Benjamin Han <benhdj@cs.cmu.edu> Changed 'int' into 'size_type'.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 013001: Benjamin Han <benhdj@cs.cmu.edu> Generalize set() so now a
//         string can be tokenized using different delimiters.
// 010701: Benjamin Han <benhdj@cs.cmu.edu> Added 'toLower' arg to
//         set(); squashed a bug in find(): start the string searching
//         from sVec[i]; added subTokStr() and transform().
// 010101: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "tokStr.hpp"

#include <algorithm>

#include <ctype.h>

namespace Toolbox {

const TokenString::size_type &TokenString::npos=std::string::npos;

void TokenString::set (const std::string &str, bool toLower, const char *delim) {
  size_type i,j;

  clear();

  TokenString::str=str;
  if (toLower) std::transform(TokenString::str.begin(),TokenString::str.end(),
			      TokenString::str.begin(),tolower);

  if ((i=str.find_first_not_of(delim))!=std::string::npos) {
    for (;i!=std::string::npos;i=str.find_first_not_of(delim,j+1)) {
      sVec.push_back(i);
      if ((j=str.find_first_of(delim,i+1))!=std::string::npos)
	lVec.push_back(j-i);
      else {
	lVec.push_back(str.length()-i);
	break;
      }
    }
  }
}

void TokenString::set (const std::string &str, Case c, const char *delim) {
  size_type i,j;

  clear();

  TokenString::str=str;
  if (c==lower)
    std::transform(TokenString::str.begin(),TokenString::str.end(),
		   TokenString::str.begin(),tolower);
  else if (c==upper)
    std::transform(TokenString::str.begin(),TokenString::str.end(),
		   TokenString::str.begin(),toupper);

  if ((i=str.find_first_not_of(delim))!=std::string::npos) {
    for (;i!=std::string::npos;i=str.find_first_not_of(delim,j+1)) {
      sVec.push_back(i);
      if ((j=str.find_first_of(delim,i+1))!=std::string::npos)
	lVec.push_back(j-i);
      else {
	lVec.push_back(str.length()-i);
	break;
      }
    }
  }
}

TokenString::size_type TokenString::find (const std::string &target, size_type i) const {
  size_type j;
  std::vector<size_type>::const_iterator vi;

  assert(i>=0 && i<sVec.size());

  j=str.find(target,sVec[i]);

  if (j!=std::string::npos && sVec[i]<=j) {
    vi=lower_bound(sVec.begin()+i,sVec.end(),j);
    return (j!=*vi?vi-sVec.begin()-1:vi-sVec.begin());
  }
  else return npos;
}

void TokenString::subTokStr (size_type start, size_type len, TokenString &tokStr) const {
  assert(start>=0 && start<sVec.size());
  assert(len>=0);

  if (len==0) {
    tokStr.clear();
    return;
  }

  size_type i,s;
  size_type strStartIdx=sVec[start];

  s=start+len-1;
  tokStr.str=str.substr(strStartIdx,sVec[s]+lVec[s]-strStartIdx);

  tokStr.sVec.resize(len);
  tokStr.lVec.resize(len);
  for (i=0;i<len;++i) {
    tokStr.sVec[i]=sVec[i+start]-strStartIdx;
    tokStr.lVec[i]=lVec[i+start];
  }
}

int TokenString::transform (TokenTransformFunc func) {
  size_type i,j,s=sVec.size();
  int result=0;
  std::string tok,newStr;
  std::vector<size_type> sizeDV(s);
  bool sizeChange=false;

  if (s==0) return result;

  for (i=0;i<s && result==0;++i) {
    tok=str.substr(sVec[i],lVec[i]);
    result=func(tok);
    
    if (i) j=sVec[i-1]+lVec[i-1];     // the end of the last token
    else j=0;
    newStr+=str.substr(j,sVec[i]-j);  // reserve the inter-token spaces
    newStr+=tok;
    
    if ((sizeDV[i]=tok.length()-lVec[i])!=0) sizeChange=true;
  }
  
  newStr+=str.substr(sVec[i-1]+lVec[i-1]);
  str=newStr;

  if (sizeChange) {
    // size changed - need to update sVec and lVec!
    for (i=j=0;i<s;++i) {  // j is accumulated shift
      sVec[i]+=j;
      if (sizeDV[i]) {
	j+=sizeDV[i];
	lVec[i]+=sizeDV[i];
      }
    }
  }
  
  return result;
}

void TokenString::join (size_type start, size_type len, std::string &str, const std::string &delim) const {
  assert(start>=0 && start< sVec.size() && len<=sVec.size()-start);

  size_type i,s=start+len;

  str.erase();
  if (s) {
    str+=(*this)[i=start];
    for (++i;i<s;++i) {
      str+=delim;
      str+=(*this)[i];
    }
  }
}

};
