// ======================================================================
// tokStr.hpp - tokenize a string by white spaces.
// 
// 061604: Benjamin Han <benhdj@cs.cmu.edu> added the missing iostream include.
// 122502: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 102402: Benjamin Han <benhdj@cs.cmu.edu> Added the missing #include
//         <assert.h>.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress warnings
//         under -pedantic -Wall.
// 072301: Benjamin Han <benhdj@cs.cmu.edu> Minor revision of the comments.
// 052801: Benjamin Han <benhdj@cs.cmu.edu> Added another set() allowing
//         upper case conversion; minor revision.
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 022301: Benjamin Han <benhdj@cs.cmu.edu> Added join() method; added
//         TokenString::npos; changed subTokStr() into a const metnod.
// 022101: Benjamin Han <benhdj@cs.cmu.edu> Changed 'int' into 'size_type'.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 013001: Benjamin Han <benhdj@cs.cmu.edu> Generalize set() so now a
//         string can be tokenized using different delimiters.
// 010701: Benjamin Han <benhdj@cs.cmu.edu> Added subTokStr(), transform() 
//         and operator <<; also added 'toLower' arg to set().
// 010201: Benjamin Han <benhdj@cs.cmu.edu> WHITESPACE_STR is defined
//         in tbconsts.hpp now (the centralized const definition file).
// 010101: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TOKSTR_HPP
#define TOKSTR_HPP

#include "tbconsts.hpp"

#include <string>
#include <vector>
#include <iostream>

#include <assert.h>

namespace Toolbox {

// ===========================================================================
// class TokenString: Convert a string into a sequence of strings having no
//                    white spaces within
// ===========================================================================

class TokenString {
public:

  // for now 'size_type' and 'npos' are simply replicas of std::string::size_type
  // and std::string::npos - later we might want to change this
  typedef std::string::size_type size_type;
  static const size_type &npos;

private:

  std::string str;
  std::vector<size_type> sVec,lVec;   // vectors for starting positions and lengths

#if _MSC_VER!=1200
  friend std::ostream &operator << (std::ostream &os, const TokenString &tokStr);
#endif

public:

  typedef int (*TokenTransformFunc) (std::string &tok);
  enum Case {lower,upper,none};

  void clear () { sVec.clear(); lVec.clear(); }

  // the 1st form of set() converts the string to all lower-case iff toLower==true;
  // the 2nd form of set() converts the string based on the specified Case;
  // 'delim' contains all the delimiters for tokenization
  void set (const std::string &str, bool toLower=false, const char *delim=WHITESPACE_STR);
  void set (const std::string &str, Case c, const char *delim=WHITESPACE_STR);
  const std::string &read () const { return str; }

  size_type size () const { return sVec.size(); }
  const std::string operator [] (size_type i) const {
    assert(i>=0 && i<sVec.size());
    return str.substr(sVec[i],lVec[i]);
  }

  // find and return the index of the token which starts the target string;
  // the search starts from the i-th token;
  // IMPORTANT: the search is done against the *original* *untokenized* string
  //            (this only matters if 'target' contains delimiters)
  // return npos if no such token exists
  size_type find (const std::string &target, size_type i=0) const;

  // copy a range of tokens into 'tokStr'
  void subTokStr (size_type start, size_type len, TokenString &tokStr) const;

  // apply the transformation to each token
  int transform (TokenTransformFunc func);

  // join 'len' tokens starting from 'start' into 'str', using 'delim' as the delimiter
  void join (size_type start, size_type len, std::string &str, const std::string &delim=" ") const;
};

// ============================ inline functions ============================

inline std::ostream &operator << (std::ostream &os, const TokenString &tokStr) {
  return os<<tokStr.str;
}

};

#endif
