// ======================================================================
// tree.hpp - a tree container with flexible iterators
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 112002: Benjamin Han <benhdj@cs.cmu.edu> nullTI -> nullTI().
// 072102: Benjamin Han <benhdj@cs.cmu.edu> Inserted 'typename' here and there
//         to supress warnings from g++ 3.1.
// 111501: Benjamin Han <benhdj@cs.cmu.edu> Changed Tree<> into _TreeBase<>
//         so we can instantiate unordered/ordered trees from there by
//         plugging in _UTreeNode<> or _TreeNode<>; added in two convenience
//         class UTree<> (unordered) and Tree<> (ordered).
// 110201: Benjamin Han <benhdj@cs.cmu.edu> Added Tree<>::move(); simplify
//         the use of deltaDepth in Tree<>::insert() (the 2nd version).
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress
//         warnings under -pedantic -Wall (mostly adding 'typename').
// 082301: Benjamin Han <benhdj@cs.cmu.edu> Revised
//         Tree<>::erase (_TreeIterator<T,T&,T*,Agenda_,Cmp>) to work with 
//         the revised PostorderAgenda.
// 081001: Benjamin Han <benhdj@cs.cmu.edu> Explicitly including <functional>.
// 080701: Benjamin Han <benhdj@cs.cmu.edu> Renamed dSize() to size() in 
//         _TreeIteratorBase<>; comments revision for iterator types.
// 080201: Benjamin Han <benhdj@cs.cmu.edu> Renamed dFindFirst() and
//         dFindNext() in _TreeIteratorBase<> to find() and findNext(),
//         respectively; renamed dBegin() and dEnd() to begin() and end() in
//         _TreeIteratorBase<>.
// 072001: Benjamin Han <benhdj@cs.cmu.edu> Moved all iterator-related code
//         to treeIter.hpp.
// 071901: Benjamin Han <benhdj@cs.cmu.edu> Simplified the implementation of
//         _TreeNode<>::DMap; add an aditional template argument Cmp to all
//         classes to allow a used-defined binary predicate for data ordering.
// 070401: Benjamin Han <benhdj@cs.cmu.edu> Changed _TreeIteratorBase<>::dBegin()
//         and _TreeIteratorBase<>::dEnd() to inline; fixed *si to **si in
//         _TreeIteratorBase<>::dFindNext(); added 
//         _TreeIteratorBase<>::operator<().
// 062201: Benjamin Han <benhdj@cs.cmu.edu> Revised the inaccurate comments for
//         Tree<>::demote() and Tree<>::promote(); revised the prototype of
//         DMap::operator[](), and fixed the problem of passing wrong pointers.
// 062101: Benjamin Han <benhdj@cs.cmu.edu> Added _TreeIteratorBase<>::isFirstD().
// 042101: Benjamin Han <benhdj@cs.cmu.edu> Changed the license from GPL to
//         LGPL.
// 041501: Benjamin Han <benhdj@cs.cmu.edu> Now dBegin() and dEnd() of 
//         _TreeIteratorBase<> are const methods.
// 041201: Benjamin Han <benhdj@cs.cmu.edu> Fixed the problem of missing history
//         entries in _PathIterator<>::operator++().
// 040901: Benjamin Han <benhdj@cs.cmu.edu> Now _PathIterator is bidirectional.
// 040301: Benjamin Han <benhdj@cs.cmu.edu> Changed the reference arguments
//         of Tree<>'s methods into value arguments; added demote() and 
//         promote() in Tree<>.
// 040201: Benjamin Han <benhdj@cs.cmu.edu> Fixed bug of _TreeNode<>::DMap - 
//         previously it's indexed by ptrs which may point to illegal places
//         after tree pruning; fixed bug in Tree<>::insert(TreeDIterator(T) pos,
//         const TreeConstIteratorBase(T) &ti) - forgot to set the mother
//         pointer for the root node of the copied subtree.
// 032801: Benjamin Han <benhdj@cs.cmu.edu> Added dFindFirst() and dFindNext().
// 031701: Benjamin Han <benhdj@cs.cmu.edu> Added next() for _PathIterator
//         so alternative paths can be traversed.
// 030601: Benjamin Han <benhdj@cs.cmu.edu> Bug fixed: copy constructor of Tree
//         didn't init numNodes to 0.
// 030401: Benjamin Han <benhdj@cs.cmu.edu> Bug fixed: added 
//         _TreeDIterator::depth() and corrected Tree::insert() (tree version).
// 030101: Benjamin Han <benhdj@cs.cmu.edu> Added _PathIterator.
// 022601: Benjamin Han <benhdj@cs.cmu.edu> Added two macros TreeIteratorBase
//         and TreeConstIteratorBase to facilitate the use of 
//         _TreeIteratorBase class.
// 022301: Benjamin Han <benhdj@cs.cmu.edu> Fixed bugs: (a) in Tree::insert()
//         (tree version) when the source tree is empty return immediately;
//         (b) in Tree::Tree(const Tree&) set root to NULL first. (c) moved
//         two typedefs in _TreeIteratorBase from protected to public. (d)
//         static member _Iterator<>::nullTI is now initialized by a default
//         constructor instead of a copy constructor.
// 022101: Benjamin Han <benhdj@cs.cmu.edu> Added _TreeIteratorBase::isRoot().
// 022001: Benjamin Han <benhdj@cs.cmu.edu> Typo: change 
//         _TreeIteratorBase::hSize() to dSize(); change some int
//         variables into variables of the appropriate size_type;
//         add _TreeNode::depth and the some related code.
// 020201: Benjamin Han <benhdj@cs.cmu.edu> namespace Toolbox.
// 012501: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TREE_HPP
#define TREE_HPP

#include "treeIter.hpp"

#include <functional>
#include <stack>

namespace Toolbox {

// ===========================================================================
// class _TreeBase (internal use ONLY): base tree strcture to be partially 
//   instantiated into unordered and ordered trees.
// ===========================================================================

template <class T, class TNode, class Cmp=std::less<T> >
class _TreeBase {
public:

  typedef typename TNode::size_type size_type;

private:

  TNode *root;
  size_type numNodes;             // # of nodes

public:

  // all of the iterator types for easy access
  // you can also directly declare an iterator independent from a tree
  // to do that use the macros defined in treeIter.hpp
  // NOTE: only DIterator/ConstDIterator and PathIterator/ConstPathIterator
  //       are bidirectional iterators (i.e., can be decremented).
  typedef TreeIteratorBase(T,TNode,Cmp) Iterator;
  typedef TreeConstIteratorBase(T,TNode,Cmp) ConstIterator;
  typedef TreeDIterator(T,TNode,Cmp) DIterator;
  typedef TreeConstDIterator(T,TNode,Cmp) ConstDIterator;
  typedef TreeIterator(T,PreorderAgenda,TNode,Cmp) PreIterator;
  typedef TreeConstIterator(T,PreorderAgenda,TNode,Cmp) ConstPreIterator;
  typedef TreeIterator(T,PostorderAgenda,TNode,Cmp) PostIterator;
  typedef TreeConstIterator(T,PostorderAgenda,TNode,Cmp) ConstPostIterator;
  typedef TreeIterator(T,BFSAgenda,TNode,Cmp) BFSIterator;
  typedef TreeConstIterator(T,BFSAgenda,TNode,Cmp) ConstBFSIterator;
  typedef TreeIterator(T,FrontAgenda,TNode,Cmp) FrontIterator;
  typedef TreeConstIterator(T,FrontAgenda,TNode,Cmp) ConstFrontIterator;
  typedef TreePathIterator(T,TNode,Cmp) PathIterator;
  typedef TreeConstPathIterator(T,TNode,Cmp) ConstPathIterator;

  void clear ();

  // both erase() and prune() prune away a subtree rooted in 'ti';
  // but erase() returns the iterator pointing to the next element after
  // the pruning based on the particular agenda of the traversal
  // NOTE: 1. erase() is a very expensive operation; if you don't care 
  //          about the next element, use prune() instead for better 
  //          performance
  //       2. after the pruning all the iterators potentially point to
  //          invalid places (except for the iterator returned by erase())! 
  //          Use begin() to 're-synch'.
  DIterator erase (DIterator pos);

  // for strange reasons MS VC6 compiles only if we put code here
  template <class Agenda_>
  _TreeIterator<T,T&,T*,Agenda_,TNode,Cmp> erase (_TreeIterator<T,T&,T*,Agenda_,TNode,Cmp> pos) {
    assert(pos.tnPtr);

    PostIterator ti(pos);
    TNode *tnPtr;
    _TreeIterator<T,T&,T*,Agenda_,TNode,Cmp> p;
    typename Iterator::size_type i;
    
    if (pos.tnPtr->mother) {
      for (i=0;ti!=pos;++i) { // leave out pos
	pos.agenda.remove(ti);
	tnPtr=ti.tnPtr;
	++ti;
	delete tnPtr;
      }

      // clear daughters and dMap of pos to ensure the operator++ won't wade
      // into the nodes which were already deleted
      pos.tnPtr->clear();
      ++(p=pos);
      
      // update daughters and dMap of the mother (must be after the last
      // statement because we might need these two for advancing the iterator)
      pos.tnPtr->mother->erase(pos.tnPtr);
      
      delete pos.tnPtr;
      ++i;
      
      numNodes-=i;
      return p;
    }
    else {
      clear();
      return pos.nullTI();
    }
  }
  void prune (Iterator pos);

  // NOTE: after the assignment all the iterators potentially point
  //       to invalid places! Use begin() to 're-synch'.
  _TreeBase &operator = (const _TreeBase &t) {
    insert(begin(),t.begin());
    return *this;
  }

  _TreeBase ():root(NULL),numNodes(0) {}
  _TreeBase (const _TreeBase &t):root(NULL),numNodes(0) { operator=(t); }
  ~_TreeBase () { clear(); }

  // NOTE begin() returns an Iterator so you can't expect
  // begin()++ works.
  Iterator begin () { return root; }
  ConstIterator begin () const { return root; }

  size_type size () const { return numNodes; }

  // insert a single node as the left sister of 'pos';
  // return an itertor pointing to the newly inserted node
  Iterator insert (DIterator pos, const T &data);

  // insert a subtree rooted at 'ti' as the left subtree of 'pos'
  // return an itertor pointing to the newly inserted node
  // NOTE: insert() is reentrancy-safe - 'pos' can point to a node under the
  //       node pointed 'ti'
  Iterator insert (DIterator pos, ConstIterator ti);

  // demote the iterator 'pos' to the only daughter of a new node whose data
  // is 'data'; pos still points to the old node (but all data members 
  // updated).
  void demote (DIterator &pos, const T &data);

  // contrary to demote(); all of the other daughters except for 'pos' are pruned;
  // pos still points to the old node (but all data members updated).
  void promote (DIterator &pos);

  // move the subtree rooted at 'src' to the left of 'dest', within the same tree
  // after the call 'src' is updated (so its mother now is changed)
  // ASSUMPTIONS: 1. src and dest are for the same tree
  //              2. dest cannot be a descendant of src
  //              3. both src and dest cannot be the root
  void move (DIterator &src, ConstDIterator dest);
};

template <class T, class TNode, class Cmp>
typename _TreeBase<T,TNode,Cmp>::DIterator _TreeBase<T,TNode,Cmp>::erase (_TreeBase<T,TNode,Cmp>::DIterator pos) {
  assert(pos.tnPtr);
  
  PostIterator ti(pos);
  TNode *tnPtr;
  DIterator p;
  typename Iterator::size_type i;
  
  if (pos.tnPtr->mother) {
    ++(p=pos);
    
    // update daughters and dMap of the mother
    pos.tnPtr->mother->erase(pos.tnPtr);
    
    for (i=0;ti!=ti.nullTI();++i) {
      tnPtr=ti.tnPtr;
      ++ti;
      delete tnPtr;
    }
    numNodes-=i;
  }
  else clear();
  return p;
}

template <class T, class TNode, class Cmp>
void _TreeBase<T,TNode,Cmp>::prune (_TreeBase<T,TNode,Cmp>::Iterator pos) {
  assert(pos.tnPtr);
  
  PostIterator ti(pos);
  TNode *tnPtr;
  typename Iterator::size_type i;
  
  if (pos.tnPtr->mother) {
    pos.tnPtr->mother->erase(pos.tnPtr);
    
    for (i=0;ti!=ti.nullTI();++i) {
      tnPtr=ti.tnPtr;
      ++ti;
      delete tnPtr;
    }
    numNodes-=i;
  }
  else clear();
}

template <class T, class TNode, class Cmp>
void _TreeBase<T,TNode,Cmp>::clear () {
  PostIterator ti;
  TNode *tnPtr;

  for (ti=begin();ti!=ti.nullTI();) {
    tnPtr=ti.tnPtr;
    ++ti;
    delete tnPtr;
  }
  root=NULL;
  numNodes=0;
}

template <class T, class TNode, class Cmp>
typename _TreeBase<T,TNode,Cmp>::Iterator _TreeBase<T,TNode,Cmp>::insert (_TreeBase<T,TNode,Cmp>::DIterator pos, 
									  const T &data) {
  if (pos.mother) {
    TNode *tnPtr=new TNode(data,pos.mother);
    
    ++numNodes;
    pos.mother->insert(pos.tnPtr,tnPtr);    
    return tnPtr;
  }
  else {
    // set up the root node
    if (!root) {
      ++numNodes;
      root=new TNode(data);
    }
    else root->data=data;
    
    return root;
  }
}

template <class T, class TNode, class Cmp>
typename _TreeBase<T,TNode,Cmp>::Iterator _TreeBase<T,TNode,Cmp>::insert (_TreeBase<T,TNode,Cmp>::DIterator pos, 
									  _TreeBase<T,TNode,Cmp>::ConstIterator ti)
{
  if (ti.tnPtr==NULL) return NULL;

  ConstPostIterator ti2(ti);
  std::stack<Iterator> st;
  Iterator daughter;
  TNode *tnPtr=NULL;  // to suppress compilation warning
  typename Iterator::size_type i,j,s,deltaDepth=pos.depth()-ti.depth();

  for (j=0;ti2!=ti2.nullTI();++ti2,++j) {
    tnPtr=new TNode(*ti2);
    
    for (i=0,s=ti2.size();i<s;++i) {
      daughter=st.top();
      st.pop();
      
      daughter.tnPtr->mother=tnPtr;
      tnPtr->push_front(daughter.tnPtr);
    }
    
    tnPtr->depth=ti2.depth()+deltaDepth;
    st.push(tnPtr);
  }

  if (pos.mother) {
    tnPtr=st.top().tnPtr;
    tnPtr->mother=pos.mother;
    pos.mother->insert(pos.tnPtr,tnPtr);
  }
  else {
    if (root) clear();
    root=tnPtr;
  }

  st.pop();
  numNodes+=j;

  return tnPtr;
}

template <class T, class TNode, class Cmp>
void _TreeBase<T,TNode,Cmp>::demote (_TreeBase<T,TNode,Cmp>::DIterator &pos, const T &data) {
  assert(pos.tnPtr);

  TNode *tnPtr=new TNode(data,pos.mother);

  // connecting above
  if (pos.mother) pos.mother->replaceDaughter(pos.tnPtr,tnPtr);
  else root=tnPtr;
  pos.mother=pos.tnPtr->mother=tnPtr;

  // connecting below
  tnPtr->push_front(pos.tnPtr);

  // update depth info
  tnPtr->depth=pos.tnPtr->depth;

  PreIterator ti(pos);
  for (;ti!=ti.nullTI();++ti) ++ti.tnPtr->depth;

  ++numNodes;
}

template <class T, class TNode, class Cmp>
void _TreeBase<T,TNode,Cmp>::promote (_TreeBase<T,TNode,Cmp>::DIterator &pos) {
  if (!pos.mother) return;

  assert(pos.tnPtr);

  Iterator iter=pos;
  DIterator dIter;

  iter.up();
  for (dIter=iter.begin();dIter!=iter.end();)
    if (dIter!=pos) dIter=erase(dIter);
    else ++dIter;

  // connecting above
  if (iter.tnPtr->mother) 
    iter.tnPtr->mother->replaceDaughter(iter.tnPtr,pos.tnPtr);
  else root=pos.tnPtr;
  pos.mother=pos.tnPtr->mother=iter.tnPtr->mother;

  delete iter.tnPtr;

  // update depth info
  PreIterator ti(pos);
  for (;ti!=ti.nullTI();++ti) --ti.tnPtr->depth;

  --numNodes;
}

template <class T, class TNode, class Cmp>
void _TreeBase<T,TNode,Cmp>::move (DIterator &src, ConstDIterator dest) {
  assert(src.tnPtr);
  assert(src.mother);
  assert(dest.mother);

  PreIterator iter;
  typename Iterator::size_type deltaDepth=dest.depth()-src.depth();

  // disconnecting with the original mother
  src.mother->erase(src.tnPtr);

  // connecting with the new mother
  src.mother=src.tnPtr->mother=dest.mother;
  src.mother->insert(dest.tnPtr,src.tnPtr);

  // update the depth info for the subtree rooted at src
  for (iter=src;iter!=iter.nullTI();++iter)
    iter.tnPtr->depth+=deltaDepth;
}

// ===========================================================================
// class Tree: Unordered tree - the ordering between siblings is insignificant
//             (so in each node we store less data - less overhead)
// ===========================================================================

template <class T, class Cmp=std::less<T> >
class UTree: public _TreeBase<T,_UTreeNode<T,Cmp>,Cmp> {};

// ===========================================================================
// class Tree: Ordered tree
// ===========================================================================

template <class T, class Cmp=std::less<T> >
class Tree: public _TreeBase<T,_TreeNode<T,Cmp>,Cmp> {};

};

#endif
