// ======================================================================
// treeAlgorithm.hpp - A bunch of algorithm implementations for a tree.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 112402: Benjamin Han <benhdj@cs.cmu.edu> Moving code around to make it
//         compile under MS VC6 (make method definition *inside* class
//         definition - another MS VC6 quirk).
// 112002: Benjamin Han <benhdj@cs.cmu.edu> nullTI -> nullTI().
// 111501: Benjamin Han <benhdj@cs.cmu.edu> Changed TIsomorphismMixed<>
//         into _TIsomorphismMixedBase<> so we can instantiate tests for
//         unordered/ordered trees from there; added two convenience
//         class UTIsomorphismMixed<> (for unordered trees) and 
//         TIsomorphismMixed<> (for ordered trees); revised some comments
//         to reflect that some tests are good for both unordered/ordered
//         trees.
// 091701: Benjamin Han <benhdj@cs.cmu.edu> Added TIsomorphismMixed<>.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress
//         warnings under -pedantic -Wall; renamed Equivalence*<> to 
//         TIsomorphism*<>, and LexicographicalCompareRelaxed<> to
//         TLexicographicalCompareRelaxed<>.
// 081001: Benjamin Han <benhdj@cs.cmu.edu> Revised comments.
// 080701: Benjamin Han <benhdj@cs.cmu.edu> Renamed dSize() to size() in 
//         _TreeIteratorBase<>; renamed all Homomorphic* identifiers to 
//         Equivalence*.
// 080201: Benjamin Han <benhdj@cs.cmu.edu> Renamed dFindFirst() and
//         dFindNext() in _TreeIteratorBase<> to find() and findNext(),
//         respectively; renamed dBegin() and dEnd() to begin() and end() in
//         _TreeIteratorBase<>.
// 072001: Benjamin Han <benhdj@cs.cmu.edu> Revised using new ways of
//         declaring tree iterators.
// 070401: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TREEALGORITHM_HPP
#define TREEALGORITHM_HPP

#include "tree.hpp"

#include <set>
#include <algorithm>
#include <iostream>

namespace Toolbox {

// ===========================================================================
// struct _DefaultTIsomorphismOp (internal use ONLY): default isomorphism 
//   operator for the strict isomorphism test.
// NOTE: T is unordered/ordered tree.
// ASSUMPTION: operator==() is defined for T::ConstIterator*.
// ===========================================================================

template <class T>
struct _DefaultTIsomorphismOp {
  bool operator () (const typename T::ConstIterator &iter1,
		    const typename T::ConstIterator &iter2) const {
    return (*iter1==*iter2);
  }
};

// ===========================================================================
// struct _DefaultTIsomorphismRelaxedOp (internal use ONLY): default isomorphism 
//   operator for the relaxed isomorphism test (act like a filter, see the 
//   comments with TIsomorphismRelaxed below).
// NOTE: T is unordered/ordered tree.
// ===========================================================================

template <class T>
struct _DefaultTIsomorphismRelaxedOp {
  bool operator () (const typename T::ConstIterator &iter1,
		    const typename T::ConstIterator &iter2) const {
    return true;
  }
};

// ===========================================================================
// struct _DefaultLCRelaxedOp (internal use ONLY): default less-than operator
//   for the relaxed lexicographical test.
// NOTE: T is unordered/ordered tree.
// ASSUMPTION: operator<() is defined for T::ConstIterator*.
// ===========================================================================

template <class T>
struct _DefaultLCRelaxedOp {
  bool operator () (const typename T::ConstIterator &iter1,
		    const typename T::ConstIterator &iter2) const {
    return (*iter1<*iter2);
  }
};

// ===========================================================================
// class TIsomorphism: Function object for a strict isomorphism test.
// NOTE:
//  1. T is unordered/ordered tree.
//  2. Op is a binary predicate testing if two tree iterators point to
//     equivalent data
// ===========================================================================

template <class T, class Op=_DefaultTIsomorphismOp<T> >
class TIsomorphism {
  Op op;
  
public:

  bool operator () (typename T::ConstIterator iter1, typename T::ConstIterator iter2) const {
    typename T::ConstPreIterator i1,i2;

    for (i1=iter1,i2=iter2;i1!=i1.nullTI() && i2!=i2.nullTI();++i1,++i2)
      if (!op(i1,i2) || i1.size()!=i2.size()) return false;

    return true;
  }
};

// ===========================================================================
// class TIsomorphismRelaxed: Function object for a relaxed isomorphism test.
// NOTE: 
//  1. T is unordered/ordered tree.
//  2. Op is a binary predicate testing if two tree iterators point to
//     equivalent data. However Op only affects the tests as a 'filter'
//     (somewhat different from the role of Op in TIsomorphism) in that Cmp (one
//     of the template parameter for Tree<>, see tree.hpp) is used for searching
//     an equivalent tree node implicitly through the use of 
//     _TreeIterator:find*(), and Op only filters out the pair of equivalent nodes
//     passed through Cmp but not equivalent by Op's definition.
// ===========================================================================

template <class T, class Op=_DefaultTIsomorphismRelaxedOp<T> >
class TIsomorphismRelaxed {
  typedef std::set<typename T::ConstIterator> ISet;

  Op op;

public:

  // set first to false iff the two root nodes are assumed equal
  bool operator () (typename T::ConstIterator iter1, typename T::ConstIterator iter2, bool first=true) const {
    if (iter1.size()!=iter2.size()) return false;
    if (first && !(*iter1==*iter2)) return false;

    typename T::ConstDIterator di1,di2;
    typename T::ConstDIterator::size_type s;
    ISet iSet;  // ensure bijection mapping

    for (di1=iter1.begin();di1!=iter1.end();++di1)
      if (iter2.find(*di1,di2)) {
	// loop till we find a new di2 which is relaxed-equivalent to di1
	while (iSet.find(di2)!=iSet.end() || (s=di1.size())!=di2.size() || 
	       !op(di1,di2) || (s && !operator()(di1,di2,false)))
	  if (!iter2.findNext(di2)) return false;
      
	iSet.insert(di2);
      }
      else return false;
    
    return true;    
  }
};

// ===========================================================================
// class _TIsomorphismMixedBase (internal use ONLY): Function object for a 
//   mixed isomorphism test (strict+relaxed) for a tree - used to instantiate 
//   tests for both unordered and ordered trees).
// NOTE:
/// 1. Different from TIsomorphism and TIsomorphismRelaxed, T is *not* a
//     tree, but is a tree node datatype (necessary evil).
//  2. StrictPred is a unary predicate determining if a node needs strict
//     tests (for the entire subtree rooted at that node) instead of relaxed
//     tests (or vice versa).
//  3. Op1 is a binary predicate for strict tests (see NOTE 2 in TIsomorphism<>).
//  4. Op2 is a binary predicate for relaxed tests (see NOTE 2 in 
//     TIsomorphismRelaxed<>).
// ===========================================================================

template <class T, class StrictPred, class TNode, class Cmp=std::less<T>, 
	  class Op1=_DefaultTIsomorphismOp<_TreeBase<T,TNode,Cmp> >, 
	  class Op2=_DefaultTIsomorphismRelaxedOp<_TreeBase<T,TNode,Cmp> > >
class _TIsomorphismMixedBase {

  template <class T_, class Ref_, class Ptr_, class StrictPred_, class TNode_, class Cmp_>
  struct StrictTestAgenda: public PreorderAgenda<T_,Ref_,Ptr_,TNode_,Cmp_> {
    StrictPred_ sPred;

    void visitNext (_TreeIteratorBase<T_,Ref_,Ptr_,TNode_,Cmp_> &ti) {
      if (ti.size() && sPred(ti)) ti=ti.begin();
      else {
	_TreeDIterator<T_,Ref_,Ptr_,TNode_,Cmp_> tdi;
	
	for (tdi=ti;tdi!=start && ++tdi==tdi.nullTI();tdi.up());
	if (tdi!=start) ti=tdi;
	else ti=_TreeIteratorBase<T_,Ref_,Ptr_,TNode_,Cmp_>::nullTI();
      }
    }
  };
  
  StrictPred sPred;
  Op1 op1;
  Op2 op2;

public:

  // for relaxed tests the two root nodes are assumed equal
  // NOTE: set strictFlag=true iff you want a strict test first (but relaxed tests are
  //       still possible depending on sPred)
  bool operator () (TreeConstIteratorBase(T,TNode,Cmp) iter1, 
		    TreeConstIteratorBase(T,TNode,Cmp) iter2,
		    bool strictFlag=true) const {
    if (strictFlag) {
      _TreeIterator<T,const T&,const T*,
	StrictTestAgenda<T,const T&,const T*,StrictPred,TNode,Cmp>,
	TNode,Cmp> si1,si2;
      typename _TreeIterator<T,const T&,const T*,
	StrictTestAgenda<T,const T&,const T*,StrictPred,TNode,Cmp>,
	TNode,Cmp>::size_type s;
    
      // note for leaf nodes we don't recurse
      for (si1=iter1,si2=iter2;si1!=si1.nullTI() && si2!=si2.nullTI();++si1,++si2)
	if (!op1(si1,si2) || (s=si1.size())!=si2.size() ||
	    (s && !sPred(si1) && !operator()(si1,si2,false)))
	  return false;
    
      return true;
    }
    else {
      if (iter1.size()!=iter2.size()) return false;
      
      TreeConstDIterator(T,TNode,Cmp) di1,di2;
      typename TreeConstDIterator(T,TNode,Cmp)::size_type s;
      std::set<TreeConstDIterator(T,TNode,Cmp) > iSet;  // ensure bijection mapping
      
      for (di1=iter1.begin();di1!=iter1.end();++di1)
	if (iter2.find(*di1,di2)) {
	  // loop till we find a new di2 which is relaxed/strict-equivalent to di1
	  // note for leaf nodes we don't recurse
	  while (iSet.find(di2)!=iSet.end() || (s=di1.size())!=di2.size() ||
		 !op2(di1,di2) || (s && !operator()(di1,di2,sPred(di1))))
	    if (!iter2.findNext(di2)) return false;
	  
	  iSet.insert(di2);
	}
	else return false;
    
      return true;
    }
  }
};

// ===========================================================================
// class UTIsomorphismMixed: mixed isomorphism test for unordered trees (see
//   comments of _TIsomorphismMixedBase<>).
// ===========================================================================

template <class T, class StrictPred, class Cmp=std::less<T>, 
	  class Op1=_DefaultTIsomorphismOp<UTree<T,Cmp> >, 
	  class Op2=_DefaultTIsomorphismRelaxedOp<UTree<T,Cmp> > >
class UTIsomorphismMixed: 
  public _TIsomorphismMixedBase<T,StrictPred,_UTreeNode<T,Cmp>,Cmp,Op1,Op2> {};

// ===========================================================================
// class TIsomorphismMixed: mixed isomorphism test for ordered trees (see
//   comments of _TIsomorphismMixedBase<>).
// ===========================================================================

template <class T, class StrictPred, class Cmp=std::less<T>, 
	  class Op1=_DefaultTIsomorphismOp<Tree<T,Cmp> >, 
	  class Op2=_DefaultTIsomorphismRelaxedOp<Tree<T,Cmp> > >
class TIsomorphismMixed: 
  public _TIsomorphismMixedBase<T,StrictPred,_TreeNode<T,Cmp>,Cmp,Op1,Op2> {};

// Relaxed lexicographcal comparison: without considering the siblings ordering
// Op is a binary predicate testing if the data pointed to by the 1st iterator is
// less than that pointed to by the 2nd iterator

// ===========================================================================
// class TLexicographicalCompareRelaxed: Function object for a relaxed
//   lexicographical test.
// NOTE: 
//  1. T is unordered/ordered tree.
//  2. Op is a binary predicate testing if the data pointed to by the 1st 
//     iterator is less than that pointed to by the 2nd iterator.
// ===========================================================================

template <class T, class Op=_DefaultLCRelaxedOp<T> >
class TLexicographicalCompareRelaxed {
  typedef std::set<typename T::ConstDIterator,Op> ISet;

  Op op;

public:

  bool operator () (typename T::ConstIterator iter1, typename T::ConstIterator iter2) const {
    if (op(iter1,iter2)) return true;
    else if (op(iter2,iter1)) return false;
  
    if (iter1.size()<iter2.size()) return true;
    if (iter1.size()>iter2.size()) return false;

    if (iter1.size()==0) return false;

    // now we have:
    // 1. *iter1 is equal to *iter2 (up to the def. of Op)
    // 2. iter1.size()==iter2.size()
    
    typename T::ConstDIterator di1,di2;
    ISet is1,is2;
    
    for (di1=iter1.begin();di1!=iter1.end();++di1) is1.insert(di1);
    for (di2=iter2.begin();di2!=iter2.end();++di2) is2.insert(di2);
    return lexicographical_compare(is1.begin(),is1.end(),is2.begin(),is2.end(),*this);
  }
};

// ===========================================================================
// operator <<: Print out a tree in a bracketed string by traversing it in the
//              preorder fashion.
// NOTE: 
//  1. T is unordered/ordered tree.
//  2. the tree is printed in the following fashion:
//       (root left-child1 left-child2 ...).
// ===========================================================================

template <class T, class TNode, class Cmp>
std::ostream &operator << (std::ostream &os, const _TreeBase<T,TNode,Cmp> &t) {
  typename _TreeBase<T,TNode,Cmp>::ConstPreIterator iter;
  bool nonRoot;
  int prevDepth,depth,count,s;

  if (t.size()==0) return os;

  for (iter=t.begin(),nonRoot=false,count=prevDepth=0;
       iter!=iter.nullTI();++iter) {
    if (nonRoot) {
      depth=iter.depth();
      if ((s=prevDepth-depth)>0) for (;s>0;--s,--count) os<<')';
      if (!iter.size()) os<<' '<<*iter;
      else {
	os<<" (";
	++count;
	os<<*iter;
      }
      prevDepth=depth;
    }
    else {
      if (!iter.size()) os<<*iter;
      else {
	os<<'(';
	++count;
	      os<<*iter;
      }
      nonRoot=true;
    }
  }
  
  while (count--) os<<')';
  
  return os; 
}

};

#endif
