// ======================================================================
// treeIter.hpp - Iterators of a tree - you can extend _TreeIterator<> by
//               implementing your own Agenda.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120602: Benjamin Han <benhdj@cs.cmu.edu> Changed name of treeNode.hpp
//         to _treeNode.hpp.
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Inserted an extra space before
//         '>' in all macros.
// 112002: Benjamin Han <benhdj@cs.cmu.edu> Make template friends go away
//         and use MS versions of iteration implementation if we're using
//         M$ VC6; replace static member nullTI with a static method 
//         nullTI() to avoid duplicate initialization problem in MS VC6.
// 102902: Benjamin Han <benhdj@cs.cmu.edu> Be explicit in publicly
//         subclassing various STL iterators.
// 102402: Benjamin Han <benhdj@cs.cmu.edu> Necessary modifications to
//         reflect the arrangement made in treeNode.hpp (see there).
// 111501: Benjamin Han <benhdj@cs.cmu.edu> Add TNode template parameter
//         to all - so they can be instantiated for both unordered and
//         ordered trees.
// 110501: Benjamin Han <benhdj@cs.cmu.edu> Added front() and back() in 
//         _TreeIteratorBase<>.
// 110301: Benjamin Han <benhdj@cs.cmu.edu> Added 
//         _TreeIteratorBase<>::isLastD(); fixed a tricky bug in type 
//         conversion/assignment between types: when assigning non-const
//         _TreeDIterator<> to const _TreeDIterator<>, 'mother' is set
//         to NULL even if the non-const one is from 
//         _TreeIteratorBase<>::end() (so 'mother' could be non-NULL).
// 103101: Benjamin Han <benhdj@cs.cmu.edu> Worked around a strange g++
//         bug (?) that removes more than one elements when doing 
//         set<>::erase(value) in _TreeNode<>::DMap<>::erase().
// 092401: Benjamin Han <benhdj@cs.cmu.edu> Added 
//         _TreeIteratorBase<>::makeNonConstIterator().
// 090901: Benjamin Han <benhdj@cs.cmu.edu> Fixed bug in
//         _TreePathIterator<>::operator--(): now pi is updated correctly;
//         added moveTo(), readPI() and pathDist() for _TreePathIterator<>.
// 090301: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions to supress
//         warnings under -pedantic -Wall (mostly adding 'typename').
// 082301: Benjamin Han <benhdj@cs.cmu.edu> Revised PreorderAgenda and 
//         PostorderAgenda: they no longer keep the local stack and should
//         run faster and use less memory.
// 082101: Benjamin Han <benhdj@cs.cmu.edu> No longer checking if tnPtr
//         is non-null in the ++ and/or -- operator in _TreeDIterator<>,
//         _TreeIterator<> and _TreePathIterator<> - they should be the
//         programmers' responsibilities.
// 080901: Benjamin Han <benhdj@cs.cmu.edu> fixed a potential bug in 
//         PostorderAgenda<>: pointing before begin() in both init() and 
//         visitNext(); bug: added check if 'root' is nullTI in init() of
//         all agendas.
// 080701: Benjamin Han <benhdj@cs.cmu.edu> Renamed dSize() to size() in 
//         _TreeIteratorBase<>; fixed a potential bug in PreorderAgenda<>:
//         pointing before begin() in both init() and visitNext(); removed
//         the local copy of Path inside _PathIterator - now we only keep a
//         pointer to an external path, so that path *must* have a longer
//         lifespan than the iterator; renamed _PathIterator<> to 
//         _TreePathIterator<> to avoid name clashes with graph path 
//         iterators.
// 080601: Benjamin Han <benhdj@cs.cmu.edu> Minor revisions of all agendas.
// 080301: Benjamin Han <benhdj@cs.cmu.edu> Made
//         _TreeIteratorBase<>::(_TreeNode<>*) a protected method.
// 080201: Benjamin Han <benhdj@cs.cmu.edu> Renamed dFindFirst() and
//         dFindNext() in _TreeIteratorBase<> to find() and findNext(),
//         respectively; removed _TreeIteratorBase<>::hasAlternative and 
//         revised _TreeIteratorBase<>::findNext(); removed
//         _PathIterator<>::hasAlternative and revised the related
//         methods; renamed dBegin() and dEnd() to begin() and end() in
//         _TreeIteratorBase<>.
// 072401: Benjamin Han <benhdj@cs.cmu.edu> Revised _TreeNode<>::DMap - 
//         now the keys to the map are T* instead of T (to save memory and
//         time).
// 072001: Benjamin Han <benhdj@cs.cmu.edu> Created (moved from tree.hpp).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TREEITER_HPP
#define TREEITER_HPP

#include "_treeNode.hpp"

#include <iterator>
#include <list>
#include <algorithm>

namespace Toolbox {

// ===========================================================================
// class _TreeIteratorBase (internal use ONLY): the tree iterator base class
// NOTE: use the macros TreeIteratorBase(T,Cmp) and TreeConstIteratorBase(T,Cmp)
// ===========================================================================

#define TreeIteratorBase(T,TNode,Cmp) _TreeIteratorBase<T,T&,T*,TNode,Cmp >
#define TreeConstIteratorBase(T,TNode,Cmp) _TreeIteratorBase<T,const T&,const T*,TNode,Cmp >

template <class T, class Ref, class Ptr, class TNode, class Cmp> class _TreeDIterator;

template <class T, class Ref, class Ptr, class TNode, class Cmp>
class _TreeIteratorBase {

#if _MSC_VER==1200
public:

#else

  // for type conversion to TreeConstIteratorBase(T,TNode,Cmp)
  template <class T_, class Ref_, class Ptr_, class TNode_, class Cmp_> friend class _TreeIteratorBase;

  // for _TreeDIterator<>::operator=()
  template <class T_, class Ref_, class Ptr_, class TNode_, class Cmp_> friend class _TreeDIterator;

  // for insert(), prune(), etc. in Tree<>
  template <class T_, class TNode_, class Cmp_> friend class _TreeBase;

#endif

  _TreeNodeIndex<T,Cmp,TNode> foundIdx;

protected:

  TNode *tnPtr;

  _TreeIteratorBase (TNode *tnPtr):tnPtr(tnPtr) {}

public:

  typedef typename TNode::size_type size_type;

  // nullTI is a null iterator used for terminating a upward traversal
  static const _TreeIteratorBase &nullTI () {
    static _TreeIteratorBase staticNullTI;

    return staticNullTI;
  }

  _TreeIteratorBase ():tnPtr(NULL) {}

  // yielding daughter iterators
  _TreeDIterator<T,Ref,Ptr,TNode,Cmp> begin () const;
  _TreeDIterator<T,Ref,Ptr,TNode,Cmp> end () const;

  // WARNING: make sure the element exists!
  const T& front () const { assert(tnPtr); return tnPtr->front(); }
  const T& back () const { assert(tnPtr); return tnPtr->back(); }

  bool find (const T &key, _TreeDIterator<T,Ref,Ptr,TNode,Cmp> &iter);
  bool findNext (_TreeDIterator<T,Ref,Ptr,TNode,Cmp> &iter);

  size_type size () const { return (tnPtr?tnPtr->size():0); }
  size_type depth () const { return (tnPtr?tnPtr->depth:0); }
  bool isRoot () const { return (tnPtr?tnPtr->mother==NULL:false); }
  bool isFirstD () const { return (tnPtr?tnPtr->isFirstD():false); }
  bool isLastD () const { return (tnPtr?tnPtr->isLastD():false); }

  _TreeIteratorBase &up () {
    assert(tnPtr); 
    tnPtr=tnPtr->mother; 
    return *this;
  }

  // type conversion to a const iterator
  operator TreeConstIteratorBase(T,TNode,Cmp) () const {
    return TreeConstIteratorBase(T,TNode,Cmp)(tnPtr);
  }

  // type conversion to a daughter iterator
  // NOTE: this is to enable assigning all the other types of iterators
  //       to _TreeDIterator<> (it's probably the only way since end()
  //       returns a TreeDIterator with tnPtr==NULL but mother could be non-NULL
  //       (so we want to distinguish between a pure NULL and a NULL with mother)
  //       this is not a trivial solution - I spent quite some time on this)
  operator _TreeDIterator<T,Ref,Ptr,TNode,Cmp> () const {
    _TreeDIterator<T,Ref,Ptr,TNode,Cmp> dIter;
    
    dIter.tnPtr=tnPtr;
    
    if (tnPtr) dIter.mother=tnPtr->mother;
    else dIter.mother=NULL;
    
    return dIter;
  }

  bool operator < (const _TreeIteratorBase &ti) const { return tnPtr<ti.tnPtr; }
  bool operator == (const _TreeIteratorBase &ti) const { return tnPtr==ti.tnPtr; }
  bool operator != (const _TreeIteratorBase &ti) const { return tnPtr!=ti.tnPtr; }

  const T& operator * () const { assert(tnPtr); return (*tnPtr).data; }
  const T* operator -> () const { return &(operator*()); }
  void update (const T &data) const { assert(tnPtr); tnPtr->update(data); }

  // WARNING: UNLESS YOU ARE SURE THE DATA MODIFICATION WON'T CHANGE THE DATA ORDERING,
  //          YOU SHOULD NOT USE THIS DANGEROUS METHOD!
  T& get () const { assert(tnPtr); return tnPtr->data; }

  // WARNING: THIS YIELDS A NON-CONST ITERATOR OUT OF A CONST ITERATOR! USE WITH CARE.
  void makeNonConstIterator (TreeIteratorBase(T,TNode,Cmp) &iter) const { iter.tnPtr=tnPtr; }

  // This can be skipped but I want to avoid copying foundIdx
  _TreeIteratorBase &operator = (const _TreeIteratorBase &ti) {
    tnPtr=ti.tnPtr;
    return *this;
  }
};

// ===========================================================================
// class _TreeDIterator: the daughter traversal iterator
// NOTE: use the macros TreeDIterator(T,TNode,Cmp) and TreeConstDIterator(T,TNode,Cmp)
// ===========================================================================

#define TreeDIterator(T,TNode,Cmp) _TreeDIterator<T,T&,T*,TNode,Cmp >
#define TreeConstDIterator(T,TNode,Cmp) _TreeDIterator<T,const T&,const T*,TNode,Cmp >

// for traversing daughters
template <class T, class Ref, class Ptr, class TNode, class Cmp>
class _TreeDIterator: public _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>, 
#if _MSC_VER==1200
  public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t> {
#else
  public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t,Ptr,Ref> {
#endif

#if _MSC_VER==1200
public:

#else

  // for type conversion to _TreeDIterator in _TreeIteratorBase<>
  template <class T_, class Ref_, class Ptr_, class TNode_, class Cmp_> friend class _TreeIteratorBase;

  // for operator=(const _TreeDIterator<T,T&,T*,TNode,Cmp> &ti)
  template <class T_, class Ref_, class Ptr_, class TNode_, class Cmp_> friend class _TreeDIterator;

  // for insert(), prune(), etc. in Tree<>
  template <class T_, class TNode_, class Cmp_> friend class _TreeBase;

#endif


  TNode *mother;

  _TreeDIterator (TNode *tnPtr):_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>(tnPtr) {
    if (tnPtr) mother=tnPtr->mother;
    else mother=NULL;
  }

  // can't reuse _TreeIteratorBase<>::operator=() since that'll introduce
  // ambiguous calling
  _TreeDIterator &operator = (TNode *tnPtr) {
    _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::operator=(tnPtr);
    if (tnPtr) mother=tnPtr->mother;
    else mother=NULL;
    return *this;
  }

public:

  typedef typename TreeIteratorBase(T,TNode,Cmp)::size_type size_type;

  _TreeDIterator ():mother(NULL) {}
  _TreeDIterator (const _TreeDIterator<T,T&,T*,TNode,Cmp> &ti) { operator=(ti); }

  // override _TreeIteratorBase::depth() since end() has a depth although
  // its tnPtr is null
  size_type depth () const { return (mother?mother->depth+1:0); }

  // overrides _TreeIteratorBase<>::up() because we can go up even if
  // a _TreeDIterator<> points to NULL, because it records its own mother
  _TreeDIterator &up () {
    tnPtr=mother;
    if (tnPtr) mother=tnPtr->mother;
    else mother=NULL;
    return *this;
  }

  // So you ask: how does _TreeDIterator<>=_TreeIteratorBase<> work? The
  // _TreeIteratorBase<> is type converted to _TreeDIterator, which is then
  // assigned to the left-hand-side _TreeDIterator<>; in case the right-hand
  // side is non-const, the following operator=() takes care of that.
  // Similar things happen for _TreeDIterator<>(_TreeIteratorBase<>).

  _TreeDIterator &operator = (const _TreeDIterator<T,T&,T*,TNode,Cmp> &ti) {
    _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::operator=(ti.tnPtr);
    mother=ti.mother;
    return *this;
  }

  // preincrement & postincrement
  // WARNING: watch out for the root node!
  _TreeDIterator &operator ++ () {
    assert(tnPtr);
    tnPtr=tnPtr->next();
    return *this;
  }
  _TreeDIterator operator ++ (int) {
    _TreeDIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  // WARNING: watch out for the root node!
  _TreeDIterator &operator -- () {    
    if (tnPtr) tnPtr=tnPtr->previous();
    else {
      assert(mother);
      tnPtr=mother->backPtr();
    }

    return *this;
  }
  _TreeDIterator operator -- (int) {
    _TreeDIterator tmp=*this;
    --*this;
    return tmp;
  }  
};

template <class T, class Ref, class Ptr, class TNode, class Cmp>
inline _TreeDIterator<T,Ref,Ptr,TNode,Cmp> _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::begin () const {
  _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tmpDI;  // by default tmpDI.tnPtr==NULL

  tmpDI.mother=tnPtr;
  if (tnPtr)
    if (tnPtr->size()) tmpDI.tnPtr=tnPtr->frontPtr();
  return tmpDI;
}

template <class T, class Ref, class Ptr, class TNode, class Cmp>
inline _TreeDIterator<T,Ref,Ptr,TNode,Cmp> _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::end () const {
  _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tmpDI;  // by default tmpDI.tnPtr==NULL
  
  tmpDI.mother=tnPtr;
  return tmpDI;
}

template <class T, class Ref, class Ptr, class TNode, class Cmp>
bool _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::find (const T &key, _TreeDIterator<T,Ref,Ptr,TNode,Cmp> &iter) {
  assert(tnPtr);
  
  if (tnPtr->find(&key,foundIdx)) {
    iter=foundIdx.readPtr();
    return true;
  }
  else return false;
}

template <class T, class Ref, class Ptr, class TNode, class Cmp>
bool _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::findNext (_TreeDIterator<T,Ref,Ptr,TNode,Cmp> &iter) {
  if (foundIdx.next()) {
    iter=foundIdx.readPtr();
    return true;
  }
  else return false;
}

// ===========================================================================
// class _TreeIterator: traverse the entire tree
// NOTE: 1. Use the macros TreeIterator(T,Agenda,TNode,Cmp) and 
//          TreeConstIterator(T,Agenda,TNode,Cmp).
//       2. Four 'factory-built' agendas are available: PreorderAgenda,
//          PostorderAgenda, BFSAgenda and FrontAgenda. You can write your own 
//          agenda by modeling after them: in particular you must implement init(),
//          visitNext() and remove() methods.
//       3. For an agenda the assumption is once an _TreeIteratorBase<>
//          is output via init() or visitNext() it must be removed from the
//          agenda.
// ===========================================================================

#define TreeIterator(T,Agenda,TNode,Cmp) _TreeIterator<T,T&,T*,Agenda<T,T&,T*,TNode,Cmp >,TNode,Cmp >
#define TreeConstIterator(T,Agenda,TNode,Cmp) _TreeIterator<T,const T&,const T*,Agenda<T,const T&,const T*,TNode,Cmp >,TNode,Cmp >

template <class T, class Ref, class Ptr, class Agenda, class TNode, class Cmp>
class _TreeIterator: public _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>, 
#if _MSC_VER==1200
  public std::iterator<std::forward_iterator_tag,T,std::ptrdiff_t> {
#else
  public std::iterator<std::forward_iterator_tag,T,std::ptrdiff_t,Ptr,Ref> {
#endif

#if _MSC_VER==1200
public:

#else

  template <class T_, class Ref_, class Ptr_, class Agenda_, class TNode_, class Cmp_> friend class _TreeIterator;
  template <class T_, class TNode_, class Cmp_> friend class _TreeBase;

#endif

  Agenda agenda;

public:

  // nullTI is a null iterator used for terminating a traversal
  // we need to override _TreeIteratorBase<>::nullTI because in the STL algorithms
  // we can't have two iterators of different types (e.g. 
  // copy(tree.begin(),tree.nullTI,...) won't work )
  static const _TreeIterator &nullTI () {
    static _TreeIterator staticNullTI;

    return staticNullTI;
  }

  _TreeIterator () {}
  _TreeIterator (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) { operator=(ti); }

  // conversion between iterators of different agendas
  template <class Agenda_>
  _TreeIterator (const _TreeIterator<T,Ref,Ptr,Agenda_,TNode,Cmp> &ti) { operator=(ti); }

  // start a new traversal from this iterator - effectively making this node 
  // as the root of a traversal
  void start () { agenda.init(*this); }

  _TreeIterator &operator = (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::operator=(ti);
    agenda.init(*this);
    return *this;
  }

  // assignment between iterators of different agendas
  template <class Agenda_>
  _TreeIterator &operator = (const _TreeIterator<T,Ref,Ptr,Agenda_,TNode,Cmp> &ti) {
    return operator=((const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>&)ti);
  }

  // preincrement & postincrement
  _TreeIterator &operator ++ () {
    agenda.visitNext(*this);
    return *this;
  }
  _TreeIterator operator ++ (int) {
    _TreeIterator tmp=*this;
    ++*this;
    return tmp;
  }
};

// ===========================================================================
// struct PreorderAgenda: agenda for preorder traversal
// ===========================================================================

template <class T, class Ref, class Ptr, class TNode, class Cmp>
struct PreorderAgenda
{
  _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> start;

  void init (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    start=ti;
  }

  void visitNext (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    if (ti.size()) ti=ti.begin();
    else {
      _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tdi;
      
      for (tdi=ti;tdi!=start && ++tdi==tdi.nullTI();tdi.up());
      if (tdi!=start) ti=tdi;
      else ti=_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::nullTI();
    }
  }

  void remove (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {}
};

// ===========================================================================
// struct PostorderAgenda: agenda for postorder traversal
// ===========================================================================

template <class T, class Ref, class Ptr, class TNode, class Cmp>
struct PostorderAgenda {
  _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> start;

  void init (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    if (ti==ti.nullTI()) return;

    // traverse to the leftmost leafnode
    for (start=ti;ti.size();ti=ti.begin());
  }

  void visitNext (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    if (ti!=start) {
      _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tdi=ti;
      
      if (++tdi!=tdi.nullTI())
	// traverse to the leftmost leafnode
	for (ti=tdi;ti.size();ti=ti.begin());
      else ti.up();
    }
    else ti=_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::nullTI();
  }

  void remove (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {}
};

// ===========================================================================
// struct BFSAgenda: agenda for Breadth-First-Search traversal
// ===========================================================================

// the only reason I didn't use queue<> is it doesn't have clear() method
template <class T, class Ref, class Ptr, class TNode, class Cmp>
struct BFSAgenda: public std::list<_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> > {
  void init (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    if (ti==ti.nullTI()) return;

    _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tdi;  // daughter iterator
    typename _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::size_type i;  

    clear();

    // push all daughters from left to right
    for (tdi=ti.begin(),i=ti.size();i>0;++tdi,--i) push_back(tdi);
  }

  void visitNext (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    _TreeDIterator<T,Ref,Ptr,TNode,Cmp> tdi;  // daughter iterator
    typename _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::size_type i;

    if (size()) {
      ti=*begin();
      erase(begin());
      
      // push all daughters from left to right
      for (tdi=ti.begin(),i=ti.size();i>0;++tdi,--i) push_back(tdi);
    }
    else ti=_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::nullTI();
  }

  void remove (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    typename std::list<_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> >::iterator li;

    li=find(begin(),end(),ti);
    if (li!=end()) erase(li);
  }
};

// ===========================================================================
// struct FrontAgenda: agenda for front traversal
// ===========================================================================

template <class T, class Ref, class Ptr, class TNode, class Cmp>
struct FrontAgenda: public PostorderAgenda<T,Ref,Ptr,TNode,Cmp> {
  void visitNext (_TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    do {
      PostorderAgenda<T,Ref,Ptr,TNode,Cmp>::visitNext(ti);
    } while (ti.size());
  }
};

// ===========================================================================
// class _TreePathIterator: iterating along a pre-defined path
// NOTE: 1. use the macros TreePathIterator(T,TNode,Cmp) and TreeConstPathIterator(T,TNode,Cmp)
//       2. the first increment should bring the iterator to a node whose
//          data matches the first element of 'path'
// ===========================================================================

#define TreePathIterator(T,TNode,Cmp) _TreePathIterator<T,T&,T*,TNode,Cmp >
#define TreeConstPathIterator(T,TNode,Cmp) _TreePathIterator<T,const T&,const T*,TNode,Cmp >

template <class T, class Ref, class Ptr, class TNode, class Cmp>
class _TreePathIterator: public _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>, 
#if _MSC_VER==1200
  public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t> {
#else
  public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t,Ptr,Ref> {
#endif
public:

  typedef std::list<T> Path;

private:

  struct HEntry { // History entry
    TNode *tnPtr;
    bool updatePI;    // true iff when adding this entry pi is updated (incremented)
    
    HEntry (TNode *tnPtr, bool updatePI):tnPtr(tnPtr),updatePI(updatePI) {}
  };

  typedef std::list<HEntry> History;

  // this shadows the one in _TreeIteratorBase
  _TreeNodeIndex<T,Cmp,TNode> foundIdx;

  History history;
  const Path *pathPtr;      // WARNING: we don't keep a local copy of Path!
  typename Path::const_iterator pi;

public:

  // WARNING: the path associated with this iterator must have a longer lifespan
  //          than the iterator!

  _TreePathIterator ():pathPtr(NULL) {}
  _TreePathIterator (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti):pathPtr(NULL) { operator=(ti); }
  _TreePathIterator (const Path &path):pathPtr(&path),pi(path.begin()) {}

  // WARNING: must call _TreePathIterator(const Path&) or setPath() first!
  _TreePathIterator &operator = (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    assert(pathPtr);

    _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::operator=(ti);
    pi=pathPtr->begin();
    history.clear();
    return *this;
  }

  // point to some node without consuming any path element
  // WARNING: must call _TreePathIterator(const Path&) or setPath() first!
  void moveTo (const _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp> &ti) {
    assert(pathPtr);

    history.push_back(HEntry(tnPtr,false));
    _TreeIteratorBase<T,Ref,Ptr,TNode,Cmp>::operator=(ti);
  }

  // NOTE pi always points to the data of the *next* visiting node
  const typename Path::const_iterator &readPI () const { return pi; }

  // returns how many path elements are *left* to visit
  // call readPath().size() to get the length of the entire path
  typename Path::size_type pathDist () const { return distance(pi,pathPtr->end()); }

  void setPath (const Path &path) {
    pathPtr=&path; 
    pi=path.begin();
    history.clear();
  }
  const Path &readPath () const { assert(pathPtr); return *pathPtr; }

  // next alternative
  // WARNING: bomb if operator++() is not called first
  bool next () {
    if (!foundIdx.next()) return false;
    tnPtr=foundIdx.readPtr();
    return true;
  }

  // preincrement & postincrement
  // at most you'll hit a nullTI but you can back to non-nullTI by '--'
  _TreePathIterator &operator ++ () {
    assert(pathPtr && tnPtr);

    if (pi==pathPtr->end() || !tnPtr->find(&(*pi),foundIdx)) {
      history.push_back(HEntry(tnPtr,false));
      tnPtr=NULL;
    }
    else {
      history.push_back(HEntry(tnPtr,true));
      tnPtr=foundIdx.readPtr();
      ++pi;
    }
	
    return *this;
  }
  _TreePathIterator operator ++ (int) {
    _TreePathIterator tmp=*this;
    ++*this;
    return tmp;
  }

  // predecrement & postdecrement
  // at most you'll return where you started - it never becomes nullTI!
  _TreePathIterator &operator -- () {
    if (history.size()) {
      HEntry &he=history.back();
      
      if (he.updatePI) --pi;
      tnPtr=he.tnPtr;
      history.pop_back();
    }
    return *this;
  }
  _TreePathIterator operator -- (int) {
    _TreePathIterator tmp=*this;
    --*this;
    return tmp;
  }  
};

};

#endif
