// ======================================================================
// trie.hpp - an unordered trie (prefix tree) structure
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 112002: Benjamin Han <benhdj@cs.cmu.edu> nullTI -> nullTI().
// 120601: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef TRIE_HPP
#define TRIE_HPP

#include "tree.hpp"

#include <list>

namespace Toolbox {

// ===========================================================================
// INTERNAL USE
// struct _TrieNode: payload of each trie node
// ===========================================================================

template <class Label, class Data>
struct _TrieNode {
  Label label;
  Data data;

  bool operator < (const _TrieNode &other) const { return label<other.label; }
};

// ===========================================================================
// class Trie: unordered trie
// ===========================================================================

template <class Label, class Data>
class Trie: public UTree<_TrieNode<Label,Data> > {
  
  _TrieNode<Label,Data> key;

public:

  typedef std::list<Label> Prefix;

  // TO-DO: I'm not sure why I need to declare the following 2 typedef's;
  //        to illustrate in cfg.hpp, CFG<>::find(const RHS &...) I can use 
  //        RHSTrie::PreIterator with no problem, but if I didn't give the
  //        typedef's here, g++ 3.01 would've reported syntax error on
  //        Iterator addPrefix(...), and others.
  typedef typename UTree<_TrieNode<Label,Data> >::Iterator Iterator;
  typedef typename UTree<_TrieNode<Label,Data> >::DIterator DIterator;

  // insert a virtual node as the root
  Trie () { insert(begin(),key); }

  Iterator addPrefix (const Prefix &prefix);
  Iterator addPrefix (const Prefix &prefix, const Data &defaultData);
  Iterator findPrefix (const Prefix &prefix);

  // return true iff 'iter' has a daughter node with 'label'
  bool findNext (Iterator &iter, const Label &label);
};

template <class Label, class Data>
typename Trie<Label,Data>::Iterator
Trie<Label,Data>::addPrefix (const Prefix &prefix) {
  typename Prefix::const_iterator pi;
  Iterator iter;
  DIterator dIter;

  for (pi=prefix.begin(),iter=begin();pi!=prefix.end();++pi,iter=dIter) {
    key.label=*pi;
    if (!iter.find(key,dIter))
      // prefix partially exists - need to augment the trie
      do {
	iter=insert(iter.end(),key);
	if (++pi!=prefix.end()) key.label=*pi;
	  else return iter;
      } while (true);
  }
  return iter;
}

template <class Label, class Data>
typename Trie<Label,Data>::Iterator
Trie<Label,Data>::addPrefix (const Prefix &prefix, const Data &defaultData) {
  typename Prefix::const_iterator pi;
  Iterator iter;
  DIterator dIter;

  for (pi=prefix.begin(),iter=begin();pi!=prefix.end();++pi,iter=dIter) {
    key.label=*pi;
    if (!iter.find(key,dIter))
      // prefix partially exists - need to augment the trie
      do {
	iter=insert(iter.end(),key);
	iter.get().data=defaultData;
	if (++pi!=prefix.end()) key.label=*pi;
	else return iter;
      } while (true);
  }
  return iter;
}

template <class Label, class Data>
typename Trie<Label,Data>::Iterator
Trie<Label,Data>::findPrefix (const Prefix &prefix) {
  typename Prefix::const_iterator pi;
  Iterator iter;
  DIterator dIter;

  for (pi=prefix.begin(),iter=begin();pi!=prefix.end();++pi,iter=dIter) {
    key.label=*pi;
    if (!iter.find(key,dIter)) return iter.nullTI();
  }
  return iter;
}

template <class Label, class Data>
bool Trie<Label,Data>::findNext (Iterator &iter, const Label &label) {
  DIterator dIter;

  key.label=label;
  if (iter.find(key,dIter)) {
    iter=dIter;
    return true;
  }
  else return false;
}

template <class Label, class Data>
std::ostream &operator << (std::ostream &os, const _TrieNode<Label,Data> &n) {
  return os<<n.label<<':'<<n.data;
}

};

#endif
