// ======================================================================
// vFile.hpp - Versatile files: kitchen sink for data storage.
// 
// 121602: Benjamin Han <benhdj@cs.cmu.edu> removed "using namespace std;" 
//         and inserted std:: prefix where needed.
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Created (total rewrite).
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef VFILE_HPP
#define VFILE_HPP

#include <assert.h>
#include <fstream>

#include "except.hpp"
#include "_vfHelper.hpp"

namespace Toolbox {

#define VFILE_CLOSED -1
#define VFILE_READ    0
#define VFILE_WRITE   1

DEFINE_SIMPLE_EXCEPTION(VFileOpenError,"Cannot open the file %s.");

// IMPORTANT: for user-defined types they have to inherit from VFObject (
//            see _vfHelper.hpp) and they must provide implementations
//            for VFObjec::writeToVFile(VFile&) and 
//            VFObject::readFromVFile(VFile&)!

// NOTE for M$ VC6: VFObject doesn't work if it's not the only super-class
//   and one of the others is an STL container with a predefined _VFHelper -
//   Rule of thumb: no multiple inheritance in M$ VC6!

class VFile {
#if _MSC_VER==1200
public:
#else
  template <class T> friend struct _VFHelper;
#endif

  int mode;               // VFILE_READ or VFILE_WRITE
  std::fstream *fsPtr;

public:

  typedef VFILE_SIZETYPE size_type;

  void open (const std::string &fileName, int mode);
  void close ();

  VFile ():mode(VFILE_CLOSED),fsPtr(NULL) {}
  ~VFile () { close(); }

  template <class T>
  void write (const T &t) {
    assert(mode==VFILE_WRITE);
    _VFHelper<T>::write(t,*this);
  }
  
  // for array
  template <class T>
  void write (const T *t, size_type size) {
    assert(mode==VFILE_WRITE);
    // const since we only have write() defined for _VFHelper<const char*>
    _VFHelper<const T*>::write(t,size,*this);
  }
  
  template <class T>
  void read (T &t) {
    assert(mode==VFILE_READ);
    _VFHelper<T>::read(t,*this);
  }
  
  // for array
  template <class T>
  void read (T *&t, size_type &size) {
    assert(mode==VFILE_READ);
    _VFHelper<T*>::read(t,size,*this);
  }
  
  // ====== Syntactic sugar ======
  // NOTE: only for non-array versions of write() and read()
  template <class T>
  VFile &operator << (const T &t) {
    write(t);
    return *this;
  }
  
    template <class T>
    VFile &operator >> (T &t) {
      read(t);
      return *this;
    }
};
  
#define _DEFINE_GENERIC_VFHELPER(TYPE)					\
  template <>								\
  struct _VFHelper<TYPE> {						\
    static void write (const TYPE i, VFile &vf) { vf.fsPtr->write((const char*)(&i),sizeof(i)); } \
    static void read (TYPE &i, VFile &vf) { vf.fsPtr->read((char*)(&i),sizeof(i)); } \
  }
  
// a plethora of helpers for basic datatypes
_DEFINE_GENERIC_VFHELPER(bool);
_DEFINE_GENERIC_VFHELPER(unsigned char);
_DEFINE_GENERIC_VFHELPER(signed char);
_DEFINE_GENERIC_VFHELPER(char);
_DEFINE_GENERIC_VFHELPER(unsigned short int);
_DEFINE_GENERIC_VFHELPER(signed short int);
_DEFINE_GENERIC_VFHELPER(unsigned int);
_DEFINE_GENERIC_VFHELPER(signed int);
_DEFINE_GENERIC_VFHELPER(unsigned long int);
_DEFINE_GENERIC_VFHELPER(signed long int);
_DEFINE_GENERIC_VFHELPER(float);
_DEFINE_GENERIC_VFHELPER(double);

template <>
struct _VFHelper<const char*> {
  typedef VFILE_SIZETYPE size_type;
  static void write (const char cPtr[], VFile &vf) {
    size_type len=strlen(cPtr);
    vf.fsPtr->write((const char*)(&len),sizeof(len));
    vf.fsPtr->write(cPtr,len*sizeof(char));
  }
  static void write (const char cPtr[], size_type size, VFile &vf) {
    vf.fsPtr->write((const char*)(&size),sizeof(size));
    vf.fsPtr->write(cPtr,size*sizeof(char));
  }
};

template <>
struct _VFHelper<char*> {
  typedef VFILE_SIZETYPE size_type;
  static void read (char *&cPtr, VFile &vf) {
    size_type len;
    vf.fsPtr->read((char*)(&len),sizeof(len));
    if (len) {
      cPtr=new char[len+1];
      vf.fsPtr->read(cPtr,len*sizeof(char));
      cPtr[len]=0;
    }
    else cPtr=NULL;
  }
  static void read (char *&cPtr, size_type &size, VFile &vf) {
    vf.fsPtr->read((char*)(&size),sizeof(size));
    if (size) {
      cPtr=new char[size];
      vf.fsPtr->read(cPtr,size*sizeof(char));
    }
    else cPtr=NULL;
  }
};

template <>
struct _VFHelper<std::string> {
  static void write (const std::string &str, VFile &vf) {
    std::string::size_type len=str.length();
    vf.fsPtr->write((const char*)(&len),sizeof(len));
    vf.fsPtr->write((const char*)(str.c_str()),len*sizeof(const char));
  }
  static void read (std::string &str, VFile &vf) {
    char *buf;
    vf.read(buf);
    if (buf) {
      str=buf;
      delete[] buf;
    }
    else str.erase();
  }
};

};

#endif
