// ======================================================================
// vfTst.cpp - code for testing and demonstrating the use of vFile.?pp.
// 
// 120202: Benjamin Han <benhdj@cs.cmu.edu> Created.
// ======================================================================

//    Copyright (C) 2000-2004 Benjamin Han <benhdj@cs.cmu.edu>
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "vFile.hpp"

#include <iostream>

using namespace std;
using namespace Toolbox;

// some user-defined type: must inherit from VFObject and implement
// writeToVFile(VFile&) and readFromVFile(VFile&)!
struct Bar: public VFObject {
  int i;
  
  Bar ():i(0) {}
  Bar (int i):i(i) {}
  
  void writeToVFile (VFile &vf) const { vf<<i; }
  void readFromVFile (VFile &vf) { vf>>i; }
  
  bool operator == (const Bar &other) const { return i==other.i; }
};

// test file IO for basic types
template <class T>
void test (const T &out, const char *s) {
  VFile vf;
  T in;
  
  vf.open("vfTst.dat",VFILE_WRITE);
  vf<<out;
  vf.close();
  vf.open("vfTst.dat",VFILE_READ);
  vf>>in;
  vf.close();
  
  if (in==out) cout<<s<<" succeeded."<<endl;
  else cout<<s<<" failed."<<endl;
}

// test file IO for arrays
template <class T>
void testArray (const T out[], VFile::size_type outSize, const char *s) {
  VFile vf;
  VFile::size_type i,inSize;
  T *in;
  
  vf.open("vfTst.dat",VFILE_WRITE);
  vf.write(out,outSize);
  vf.close();
  vf.open("vfTst.dat",VFILE_READ);
  vf.read(in,inSize);
  vf.close();
  
  if (inSize==outSize) {
    for (i=0;i<inSize && in[i]==out[i];++i);
    if (i==inSize) cout<<s<<" succeeded."<<endl;
    else cout<<s<<" failed."<<endl;
  }
  else cout<<s<<" failed."<<endl;
  
  delete[] in;
}

// test file IO for C-style strings (null-terminated char*)
void testCPtr (char *&in, const char *out, const char *s) {
  VFile vf;
  
  vf.open("vfTst.dat",VFILE_WRITE);
  vf<<out;
  vf.close();
  vf.open("vfTst.dat",VFILE_READ);
  vf>>in;
  vf.close();
  
  if (strcmp(in,out)==0) cout<<s<<" succeeded."<<endl;
  else cout<<s<<" failed."<<endl;
  
  delete[] in;
}

// test file IO for an STL sequence
template <class T>
void testSequence (const T &out, const char *s) {
  VFile vf;
  T in;
  
  vf.open("vfTst.dat",VFILE_WRITE);
  vf<<out;
  vf.close();
  vf.open("vfTst.dat",VFILE_READ);
  vf>>in;
  vf.close();
  
  if (in.size()==out.size()) {
    typename T::const_iterator i,j;
    
    for (i=in.begin(),j=out.begin();i!=in.end() && *i==*j;++i,++j);
    
    if (i==in.end()) cout<<s<<" succeeded."<<endl;
    else cout<<s<<" failed."<<endl;
  }
  else cout<<s<<" failed."<<endl;
}

// test file IO for an STL maps
template <class T>
void testMap (const T &out, const char *s) {
  VFile vf;
  T in;
  
  vf.open("vfTst.dat",VFILE_WRITE);
  vf<<out;
  vf.close();
  vf.open("vfTst.dat",VFILE_READ);
  vf>>in;
  vf.close();
  
  if (in.size()==out.size()) {
    typename T::const_iterator i,j;
    
    for (i=in.begin(),j=out.begin();
	 i!=in.end() && i->first==j->first && i->second==j->second;
	 ++i,++j);
    
    if (i==in.end()) cout<<s<<" succeeded."<<endl;
    else cout<<s<<" failed."<<endl;
  }
  else cout<<s<<" failed."<<endl;
}

int main () {
  VFile vf;
  
  bool b1=false;
  const bool b2=true;
  test(b1,"bool");
  test(b2,"const bool");

  unsigned char uc1='a';
  const unsigned char uc2='b';
  test(uc1,"unsigned char");
  test(uc2,"const unsigned char");

  // below for basic datatypes I didn't test const but they should work

  signed char sc='a';
  test(sc,"signed char");
  
  char c='a';
  test(c,"char");
  
  unsigned short us=10;
  test(us,"unsigned short");
  
  signed short ss=-10;
  test(ss,"signed short");
  
  short s=10;
  test(s,"short");
  
  unsigned int ui=100000;
  test(ui,"unsigned int");
  
  signed int si=-100000;
  test(si,"signed int");
  
  int i=10;
  test(i,"int");

  unsigned long ul=100000;
  test(ul,"unsigned long");
  
  signed long sl=-100000;
  test(sl,"signed long");
  
  long l=10;
  test(l,"long");
  
  float f=0.001f;
  test(f,"float");
  
  double d=0.001;
  test(d,"double");
  
  Bar bar(10);
  test(bar,"user-defined type");
  
  char *cPtr1,*cPtr2="hello";
  testCPtr(cPtr1,cPtr2,"const char* (string)");
  
  char *cPtr3=new char[2];
  cPtr3[0]='a'; cPtr3[1]=0;
  testCPtr(cPtr1,cPtr3,"char* (string)");
  delete[] cPtr3;
  
  int *ip1=new int[3];
  for (i=0;i<2;++i) ip1[i]=i;
  testArray(ip1,3,"int array");
  delete[] ip1;
  
  int ip2[3]={0,1,2};
  testArray(ip2,3,"const int array");
  
  char *cp1=new char[3];
  for (c=0;c<2;++c) cp1[c]=c;
  testArray(cp1,3,"char* (array)");
  delete[] cp1;
  
  char cp2[3]={'a','b','c'};
  testArray(cp2,3,"const char* (array)");

  string str="hello";
  test(str,"string");
  
  vector<int> v;
  v.push_back(1);
  v.push_back(2);
  testSequence(v,"vector<int>");
  
  list<char> lc;
  lc.push_back('A');
  lc.push_back('B');
  testSequence(lc,"list<char>");
  
  set<char> setC;
  setC.insert('A');
  setC.insert('B');
  testSequence(setC,"set<char>");
  
  multiset<char> ms;
  ms.insert('A');
  ms.insert('B');
  ms.insert('B');
  testSequence(ms,"multiset<char>");
  
  map<int,int> m;
  m.insert(make_pair(0,0));
  m.insert(make_pair(1,1));
  testMap(m,"map<int,int>");
  
  map<int,int> mm;
  mm.insert(make_pair(0,0));
  mm.insert(make_pair(1,1));
  mm.insert(make_pair(1,2));
  testMap(mm,"multimap<int,int>");

  return 0;
}
